/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "<T>.defs"

struct <T>AssocArrNode
{
    char* name;			/* name */
    <T>   value;		/* value */
    struct <T>AssocArrNode* left;
    struct <T>AssocArrNode* right;
};

#include "<T>AssocArr.h"


static unsigned long hash(const char* s)
{
    unsigned long x;
    unsigned char c;

    x = 0;
    while (*s)
    {
      c = *s++;
      if(c >= 0140)
	  c -= 40;
      x = ((x <<3 ) + (x >> 28) + c);
    }

    return x;
}


static struct <T>AssocArrNode* alloc<T>AssocArrNode(const char* name, <T> value)
{
    struct <T>AssocArrNode* p;

    p = (struct <T>AssocArrNode *)xmalloc(sizeof(struct <T>AssocArrNode));

    if(p == NULL)
    {
	fprintf(stderr, "<T>AssocArr: Can't allocate a memory.\n");
	exit(1);
    }

    if((p->name = (char *)strdup(name)) == NULL)
    {
	fprintf(stderr, "<T>AssocArr: Can't allocate a memory.\n");
	exit(1);
    }

    p->value = value;
    p->left = p->right = NULL;

    return p;
}

void make<T>AssocArr(<T>AssocArr* arr, int size)
{
    arr->size = size;
    arr->assoctab = (struct <T>AssocArrNode **)xcalloc(size, sizeof(struct <T>AssocArrNode *));

    if(arr->assoctab == NULL)
    {
	fprintf(stderr, "<T>AssocArr: Can't allocate a memory.\n");
	exit(1);
    }
}

static void delete<T>AssocArrNode(struct <T>AssocArrNode* node)
{
    if(node->left)
	delete<T>AssocArrNode(node->left);
    if(node->right)
	delete<T>AssocArrNode(node->right);
    xfree(node->name);
    xfree(node);
}    

void delete<T>AssocArr(<T>AssocArr* arr)
{
    int i;

    for(i = 0; i < arr->size; i++)
	if(arr->assoctab[i])
	    delete<T>AssocArrNode(arr->assoctab[i]);
    xfree(arr->assoctab);
    arr->size = 0;
    arr->assoctab = NULL;
}

void install<T>AssocArr(<T>AssocArr* arr, const char* name, <T> value)
{
    struct <T>AssocArrNode* node;
    struct <T>AssocArrNode* prev;
    long addr;
    int is_left;

    if(arr->size == 0)
	make<T>AssocArr(arr, DEFAULT_<T>ASSCO_TABLE_SIZE);

    addr = hash(name) % arr->size;
    node = arr->assoctab[addr];
    if(node == NULL)
    {
	arr->assoctab[addr] = alloc<T>AssocArrNode(name, value);
	return;
    }

    while(node)
    {
	prev = node;
	if(strcmp(node->name, name) > 0)
	{
	    is_left = 1;
	    node = node->left;
	}
	else if(strcmp(node->name, name) < 0)
	{
	    is_left = 0;
	    node = node->right;
	}
	else
	{
	    node->value = value;
	    return;
	}
    }

    if(is_left)
	prev->left = alloc<T>AssocArrNode(name, value);
    else
	prev->right = alloc<T>AssocArrNode(name, value);
}

<T>* search<T>AssocArr(const <T>AssocArr* arr, const char* name)
{
    struct <T>AssocArrNode* node;

    if(arr->size == 0)
	return NULL;

    node = arr->assoctab[hash(name) % arr->size];

    while(node)
    {
	if(strcmp(node->name, name) > 0)
	    node = node->left;
	else if(strcmp(node->name, name) < 0)
	    node = node->right;
	else
	    return &node->value;
    }

    return NULL;
}

static void map<T>AssocArrNode(struct <T>AssocArrNode* node, void (* mapfn)(const char* name, <T>* value))
{
    mapfn(node->name, &node->value);
    if(node->left)
	map<T>AssocArrNode(node->left, mapfn);
    if(node->right)
	map<T>AssocArrNode(node->right, mapfn);
}

void map<T>AssocArr(<T>AssocArr* arr, void (* mapfn)(const char* name, <T>* value))
{
    int i;

    for(i = 0; i < arr->size; i++)
	if(arr->assoctab[i])
	    map<T>AssocArrNode(arr->assoctab[i], mapfn);
}
