/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include "QHFunc.defs"
#include "QHFuncStack.h"


void makeQHFuncStack(QHFuncStack* st, int size)
{
    if(size <= 0)
    {
	fprintf(stderr, "QHFuncStack: Invalid stack size:%d\n", size);
	exit(1);
    }

    if((st->elem = (QHFunc *)xcalloc(size, sizeof(QHFunc))) == NULL)
    {
	fprintf(stderr, "QHFuncStack: Can't allocate a memory.\n");
	exit(1);
    }
    st->ptr = 0;
    st->size = size;
}

void deleteQHFuncStack(QHFuncStack* st)
{
    xfree(st->elem);
}

void resizeQHFuncStack(QHFuncStack* st, int size)
{
    if(size <= 0)
    {
	fprintf(stderr, "QHFuncStack: Invalid stack resize:%d\n", size);
	exit(1);
    }

    if((st->elem = (QHFunc *)xrealloc(st->elem, sizeof(QHFunc) * size)) == NULL)
    {
	fprintf(stderr, "QHFuncStack: Can't reallocate a memory.\n");
	exit(1);
    }

    st->size = size;
}

void emptyQHFuncStack(QHFuncStack* st)
{
    st->ptr = 0;
}

void pushQHFuncStack(QHFuncStack* st, QHFunc value)
{
    if(st->ptr == st->size)
	resizeQHFuncStack(st, st->size * 2);
    st->elem[st->ptr++] = value;
}

QHFunc topQHFuncStack(QHFuncStack* st)
{
    if(isemptyQHFuncStack(st))
    {
	fprintf(stderr, "QHFuncStack: Attempted topQHFuncStack() for empty stack.\n");
	exit(1);
    }

    return st->elem[st->ptr - 1];
}

QHFunc popQHFuncStack(QHFuncStack* st)
{
    if(isemptyQHFuncStack(st))
    {
	fprintf(stderr, "QHFuncStack: Attempted popQHFuncStack() for empty stack.\n");
	exit(1);
    }

    return st->elem[--st->ptr];
}
