/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "QHObj.defs"

struct QHObjAssocArrNode
{
    char* name;			/* name */
    QHObj   value;		/* value */
    struct QHObjAssocArrNode* left;
    struct QHObjAssocArrNode* right;
};

#include "QHObjAssocArr.h"


static unsigned long hash(const char* s)
{
    unsigned long x;
    unsigned char c;

    x = 0;
    while (*s)
    {
      c = *s++;
      if(c >= 0140)
	  c -= 40;
      x = ((x <<3 ) + (x >> 28) + c);
    }

    return x;
}


static struct QHObjAssocArrNode* allocQHObjAssocArrNode(const char* name, QHObj value)
{
    struct QHObjAssocArrNode* p;

    p = (struct QHObjAssocArrNode *)xmalloc(sizeof(struct QHObjAssocArrNode));

    if(p == NULL)
    {
	fprintf(stderr, "QHObjAssocArr: Can't allocate a memory.\n");
	exit(1);
    }

    if((p->name = (char *)strdup(name)) == NULL)
    {
	fprintf(stderr, "QHObjAssocArr: Can't allocate a memory.\n");
	exit(1);
    }

    p->value = value;
    p->left = p->right = NULL;

    return p;
}

void makeQHObjAssocArr(QHObjAssocArr* arr, int size)
{
    arr->size = size;
    arr->assoctab = (struct QHObjAssocArrNode **)xcalloc(size, sizeof(struct QHObjAssocArrNode *));

    if(arr->assoctab == NULL)
    {
	fprintf(stderr, "QHObjAssocArr: Can't allocate a memory.\n");
	exit(1);
    }
}

static void deleteQHObjAssocArrNode(struct QHObjAssocArrNode* node)
{
    if(node->left)
	deleteQHObjAssocArrNode(node->left);
    if(node->right)
	deleteQHObjAssocArrNode(node->right);
    xfree(node->name);
    xfree(node);
}    

void deleteQHObjAssocArr(QHObjAssocArr* arr)
{
    int i;

    for(i = 0; i < arr->size; i++)
	if(arr->assoctab[i])
	    deleteQHObjAssocArrNode(arr->assoctab[i]);
    xfree(arr->assoctab);
    arr->size = 0;
    arr->assoctab = NULL;
}

void installQHObjAssocArr(QHObjAssocArr* arr, const char* name, QHObj value)
{
    struct QHObjAssocArrNode* node;
    struct QHObjAssocArrNode* prev;
    long addr;
    int is_left;

    if(arr->size == 0)
	makeQHObjAssocArr(arr, DEFAULT_QHObjASSCO_TABLE_SIZE);

    addr = hash(name) % arr->size;
    node = arr->assoctab[addr];
    if(node == NULL)
    {
	arr->assoctab[addr] = allocQHObjAssocArrNode(name, value);
	return;
    }

    while(node)
    {
	prev = node;
	if(strcmp(node->name, name) > 0)
	{
	    is_left = 1;
	    node = node->left;
	}
	else if(strcmp(node->name, name) < 0)
	{
	    is_left = 0;
	    node = node->right;
	}
	else
	{
	    node->value = value;
	    return;
	}
    }

    if(is_left)
	prev->left = allocQHObjAssocArrNode(name, value);
    else
	prev->right = allocQHObjAssocArrNode(name, value);
}

QHObj* searchQHObjAssocArr(const QHObjAssocArr* arr, const char* name)
{
    struct QHObjAssocArrNode* node;

    if(arr->size == 0)
	return NULL;

    node = arr->assoctab[hash(name) % arr->size];

    while(node)
    {
	if(strcmp(node->name, name) > 0)
	    node = node->left;
	else if(strcmp(node->name, name) < 0)
	    node = node->right;
	else
	    return &node->value;
    }

    return NULL;
}

static void mapQHObjAssocArrNode(struct QHObjAssocArrNode* node, void (* mapfn)(const char* name, QHObj* value))
{
    mapfn(node->name, &node->value);
    if(node->left)
	mapQHObjAssocArrNode(node->left, mapfn);
    if(node->right)
	mapQHObjAssocArrNode(node->right, mapfn);
}

void mapQHObjAssocArr(QHObjAssocArr* arr, void (* mapfn)(const char* name, QHObj* value))
{
    int i;

    for(i = 0; i < arr->size; i++)
	if(arr->assoctab[i])
	    mapQHObjAssocArrNode(arr->assoctab[i], mapfn);
}
