/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include "SWRBuff.defs"
#include "SWRBuffStack.h"


void makeSWRBuffStack(SWRBuffStack* st, int size)
{
    if(size <= 0)
    {
	fprintf(stderr, "SWRBuffStack: Invalid stack size:%d\n", size);
	exit(1);
    }

    if((st->elem = (SWRBuff *)xcalloc(size, sizeof(SWRBuff))) == NULL)
    {
	fprintf(stderr, "SWRBuffStack: Can't allocate a memory.\n");
	exit(1);
    }
    st->ptr = 0;
    st->size = size;
}

void deleteSWRBuffStack(SWRBuffStack* st)
{
    xfree(st->elem);
}

void resizeSWRBuffStack(SWRBuffStack* st, int size)
{
    if(size <= 0)
    {
	fprintf(stderr, "SWRBuffStack: Invalid stack resize:%d\n", size);
	exit(1);
    }

    if((st->elem = (SWRBuff *)xrealloc(st->elem, sizeof(SWRBuff) * size)) == NULL)
    {
	fprintf(stderr, "SWRBuffStack: Can't reallocate a memory.\n");
	exit(1);
    }

    st->size = size;
}

void emptySWRBuffStack(SWRBuffStack* st)
{
    st->ptr = 0;
}

void pushSWRBuffStack(SWRBuffStack* st, SWRBuff value)
{
    if(st->ptr == st->size)
	resizeSWRBuffStack(st, st->size * 2);
    st->elem[st->ptr++] = value;
}

SWRBuff topSWRBuffStack(SWRBuffStack* st)
{
    if(isemptySWRBuffStack(st))
    {
	fprintf(stderr, "SWRBuffStack: Attempted topSWRBuffStack() for empty stack.\n");
	exit(1);
    }

    return st->elem[st->ptr - 1];
}

SWRBuff popSWRBuffStack(SWRBuffStack* st)
{
    if(isemptySWRBuffStack(st))
    {
	fprintf(stderr, "SWRBuffStack: Attempted popSWRBuffStack() for empty stack.\n");
	exit(1);
    }

    return st->elem[--st->ptr];
}
