/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include "SwitchStmt.defs"
#include "SwitchStmtStack.h"


void makeSwitchStmtStack(SwitchStmtStack* st, int size)
{
    if(size <= 0)
    {
	fprintf(stderr, "SwitchStmtStack: Invalid stack size:%d\n", size);
	exit(1);
    }

    if((st->elem = (SwitchStmt *)xcalloc(size, sizeof(SwitchStmt))) == NULL)
    {
	fprintf(stderr, "SwitchStmtStack: Can't allocate a memory.\n");
	exit(1);
    }
    st->ptr = 0;
    st->size = size;
}

void deleteSwitchStmtStack(SwitchStmtStack* st)
{
    xfree(st->elem);
}

void resizeSwitchStmtStack(SwitchStmtStack* st, int size)
{
    if(size <= 0)
    {
	fprintf(stderr, "SwitchStmtStack: Invalid stack resize:%d\n", size);
	exit(1);
    }

    if((st->elem = (SwitchStmt *)xrealloc(st->elem, sizeof(SwitchStmt) * size)) == NULL)
    {
	fprintf(stderr, "SwitchStmtStack: Can't reallocate a memory.\n");
	exit(1);
    }

    st->size = size;
}

void emptySwitchStmtStack(SwitchStmtStack* st)
{
    st->ptr = 0;
}

void pushSwitchStmtStack(SwitchStmtStack* st, SwitchStmt value)
{
    if(st->ptr == st->size)
	resizeSwitchStmtStack(st, st->size * 2);
    st->elem[st->ptr++] = value;
}

SwitchStmt topSwitchStmtStack(SwitchStmtStack* st)
{
    if(isemptySwitchStmtStack(st))
    {
	fprintf(stderr, "SwitchStmtStack: Attempted topSwitchStmtStack() for empty stack.\n");
	exit(1);
    }

    return st->elem[st->ptr - 1];
}

SwitchStmt popSwitchStmtStack(SwitchStmtStack* st)
{
    if(isemptySwitchStmtStack(st))
    {
	fprintf(stderr, "SwitchStmtStack: Attempted popSwitchStmtStack() for empty stack.\n");
	exit(1);
    }

    return st->elem[--st->ptr];
}
