/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "common.h"
#include "interp.h"
#if defined(sgi)
#include <strings.h>
#endif

Array array_temp;

union qhobj_template
{
    void*          p;		/* void pointer */
    int*           i;		/* int pointer */
    double*        r;		/* double pointer */
    XComplex*      c;		/* complex pointer */
    QHObj*         x;		/* any-obj pointer */
};

static int calc_array_entry(Array* a)
{
    int i, d, t;
    int entry;
    int* size = a->size;
    int dim   = a->dim;

    entry = d = popintStack(&istack);
    if(d < 0 ||
       d >= size[0])
	execute_error("Out of array region");

    if(dim == 1)
	return d;

    for(i = 1; i < dim; i++)
    {
	d = popintStack(&istack);
	t = size[i];
	if(d < 0 ||
	   d >= t)
	    execute_error("Out of array region");
	entry = entry * t + d;
    }
    return entry;
}


static inline void reference_array(const Array* a, int align, QHObj* x)
{
    union qhobj_template u;

    u.p = a->elem;
    if(a->type == QH_Int)
    {
	x->type   = QH_Int;
	x->data.i = u.i[align];
    }
    else if(a->type == QH_Real)
    {
	x->type   = QH_Real;
	x->data.r = u.r[align];
    }
    else if(a->type == QH_Complex)
    {
	x->type   = QH_Complex;
	x->data.c = u.c[align];
    }
    else if(a->type == QH_Any)
    {
	*x = u.x[align];
    }
    else
    {
	fprintf(stderr, "Systematic error: Unknown array data type:%d\n",
		a->type);
    }
}

static void assign_array(Array* a, int align, const QHObj* x)
{
    union qhobj_template u;

    u.p = a->elem;
    switch(a->type)
    {
      case QH_Int:
	u.i[align] = cast_to_int(x);
	break;
      case QH_Real:
	u.r[align] = cast_to_real(x);
	break;
      case QH_Complex:
	u.c[align] = cast_to_complex(x);
	break;
      case QH_Any:
	u.x[align] = *x;
	break;
      default:
	fprintf(stderr, "Systematic error: Unknown array data type:%d\n", a->type);
	break;
    }
}

Array* alloc_array(void)
{
    Array* a;

    a = (Array *)alloc_cell();
    a->header.type = None;	/* protection for garbage collection */
    a->type = None;
    a->dim = 0;
    a->nelem = 0;
    a->size = NULL;
    a->elem = NULL;
    a->dbp = NULL;

    return a;
}

void qh_new(QHProg* p)
{
    QHObj* res = memstack + MEMADDR(p->res);
    Array* a;
    int i, size;
    int type = p->a1;
    long dim = p->a2;

    if(dim > 0xff)
	execute_error("Too large dimension");

    a = res->data.a = alloc_array();
    a->type = type;
    a->dim  = dim;
    a->size = (int *)xmalloc(sizeof(int) * dim);
    res->type = QH_Array;

    size = 1;
    for(i = 0; i < dim; i++)
    {
	a->size[i] = popintStack(&istack);
	if(a->size[i] <= 0)
	{
	    sprintf(temp_buff, "new: zero or negative array dimention");
	    execute_error(temp_buff);
	}
	size *= a->size[i];
    }

    a->nelem = size;
    a->elem = xmalloc(sizeof_qhtype(type) * size);
    bzero((char *)a->elem, sizeof_qhtype(type) * size);
    a->header.type = QH_Array;
}

void qh_aref(QHProg* p)
{
    QHObj* x;
    QHObj* res;
    int align;
    Array* a;

    x   = memstack + MEMADDR(p->a1);
    res = memstack + MEMADDR(p->res);

    if(x->type == QH_Array)
    {
	a = x->data.a;

	if(a->dim != p->a2)
	    execute_error("Mismatch array dimension");

	align = calc_array_entry(a);

	reference_array(a, align, res);
    }
    else if(x->type == QH_String)
    {
	QHString* s = x->data.s;

	if(p->a2 != 1)
	    execute_error("Invalid string index specifyed");
	align = popintStack(&istack);

	if(align < 0 || align >= s->maxlen)
	    res->data.i = 0;
	else
	    res->data.i = s->str[align];
    }
    else
	execute_error("Atempted none-array reference");
}


#define STRING_OPERATION(X)						\
    do {								\
	QHString* s = res->data.s;					\
	if(p->a2 != 1)							\
	    execute_error("Invalid string index specifyed");		\
	if(s->isconst)							\
	    execute_error("Attempted constant string assignment");	\
	align = popintStack(&istack);					\
	touch_string(s, align);						\
	X;								\
	return;								\
    } while(0)


/* a[] = x */
void qh_aset(QHProg* p)
{
    QHObj* res = memstack + MEMADDR(p->res);
    QHObj* x   = memstack + MEMADDR(p->a1);
    int align;
    Array* a;

    if(res->type == QH_Array)
    {
	a = res->data.a;
	if(a->dim != p->a2)
	    execute_error("Mismatch array dimension");
	align = calc_array_entry(a);
	assign_array(a, align, x);
    }
    else if(res->type == QH_String)
    {
	STRING_OPERATION(s->str[align] = cast_to_int(x));
    }
    else
	execute_error("Atempted none-array assignment");
}

/* a[] += x */
void qh_aaset(QHProg* p)
{
    QHObj* res;
    QHObj* x;
    QHObj y;
    Array* a;
    int align;
    XComplex c;

    x   = memstack + MEMADDR(p->a1);
    res = memstack + MEMADDR(p->res);

    if(res->type == QH_Array)
    {
	a = res->data.a;

	if(a->dim != p->a2)
	    execute_error("Mismatch array dimension");

	align = calc_array_entry(a);
	reference_array(a, align, &y);
	switch(a->type)
	{
	  case QH_Int:
	    y.data.i += cast_to_int(x);
	    break;
	  case QH_Real:
	    y.data.r += cast_to_real(x);
	    break;
	  case QH_Complex:
	    c = cast_to_complex(x);
	    y.data.c.re += c.re;
	    y.data.c.im += c.im;
	    break;
	  case QH_Any:
	    if(y.type  >= QH_String ||
	       x->type >= QH_String)
		invalid_operation(p->op);
	    switch(SWITCH_DATA_TYPE(&y, x))
	    {
	      case QHDATATYPE_II:
		y.data.i += x->data.i;
		break;

	      case QHDATATYPE_IR:
		y.type = QH_Real;
		y.data.r = (double)y.data.i + x->data.r;
		break;

	      case QHDATATYPE_IC:
		y.type = QH_Complex;
		y.data.c.re = (double)y.data.i + x->data.c.re;
		y.data.c.im = x->data.c.im;
		break;

	      case QHDATATYPE_RI:
		y.data.r += (double)x->data.i;
		break;

	      case QHDATATYPE_RR:
		y.data.r += x->data.r;
		break;

	      case QHDATATYPE_RC:
		y.type = QH_Complex;
		y.data.c.re = y.data.r + x->data.c.re;
		y.data.c.im = x->data.c.im;
		break;

	      case QHDATATYPE_CI:
		y.data.c.re += (double)x->data.i;
		break;

	      case QHDATATYPE_CR:
		y.data.c.re += x->data.r;
		break;

	      case QHDATATYPE_CC:
		y.data.c.re += x->data.c.re;
		y.data.c.im += x->data.c.im;
		break;
	    }
	    break;
	}
	assign_array(a, align, &y);
    }
    else if(res->type == QH_String)
	STRING_OPERATION(s->str[align] += cast_to_int(x));
    else
	execute_error("Atempted none-array operation");
}

/* a[] -= x */
void qh_asset(QHProg* p)
{
    QHObj* res;
    QHObj* x;
    QHObj y;
    Array* a;
    int align;
    XComplex c;

    x   = memstack + MEMADDR(p->a1);
    res = memstack + MEMADDR(p->res);

    if(res->type == QH_Array)
    {
	a = res->data.a;

	if(a->dim != p->a2)
	    execute_error("Mismatch array dimension");

	align = calc_array_entry(a);
	reference_array(a, align, &y);
	switch(a->type)
	{
	  case QH_Int:
	    y.data.i -= cast_to_int(x);
	    break;
	  case QH_Real:
	    y.data.r -= cast_to_real(x);
	    break;
	  case QH_Complex:
	    c = cast_to_complex(x);
	    y.data.c.re -= c.re;
	    y.data.c.im -= c.im;
	    break;
	  case QH_Any:
	    if(y.type  >= QH_String ||
	       x->type >= QH_String)
		invalid_operation(p->op);
	    switch(SWITCH_DATA_TYPE(&y, x))
	    {
	      case QHDATATYPE_II:
		y.data.i -= x->data.i;
		break;

	      case QHDATATYPE_IR:
		y.type = QH_Real;
		y.data.r = (double)y.data.i - x->data.r;
		break;

	      case QHDATATYPE_IC:
		y.type = QH_Complex;
		y.data.c.re = (double)y.data.i - x->data.c.re;
		y.data.c.im = -x->data.c.im;
		break;

	      case QHDATATYPE_RI:
		y.data.r -= (double)x->data.i;
		break;

	      case QHDATATYPE_RR:
		y.data.r -= x->data.r;
		break;

	      case QHDATATYPE_RC:
		y.type = QH_Complex;
		y.data.c.re = y.data.r - x->data.c.re;
		y.data.c.im = -x->data.c.im;
		break;

	      case QHDATATYPE_CI:
		y.data.c.re -= (double)x->data.i;
		break;

	      case QHDATATYPE_CR:
		y.data.c.re -= x->data.r;
		break;

	      case QHDATATYPE_CC:
		y.data.c.re -= x->data.c.re;
		y.data.c.im -= x->data.c.im;
		break;
	    }
	    break;
	}
	assign_array(a, align, &y);
    }
    else if(res->type == QH_String)
	STRING_OPERATION(s->str[align] -= cast_to_int(x));
    else
	execute_error("Atempted none-array operation");
}

/* a[] *= x */
void qh_amset(QHProg* p)
{
    QHObj* res;
    QHObj* x;
    QHObj y;
    Array* a;
    int align;
    XComplex c1, c2;
    double r;

    x   = memstack + MEMADDR(p->a1);
    res = memstack + MEMADDR(p->res);

    if(res->type == QH_Array)
    {
	a = res->data.a;

	if(a->dim != p->a2)
	    execute_error("Mismatch array dimension");

	align = calc_array_entry(a);
	reference_array(a, align, &y);
	switch(a->type)
	{
	  case QH_Int:
	    y.data.i *= cast_to_int(x);
	    break;
	  case QH_Real:
	    y.data.r *= cast_to_real(x);
	    break;
	  case QH_Complex:
	    c1 = y.data.c;
	    c2 = cast_to_complex(x);
	    y.data.c.re = c1.re * c2.re - c1.im * c2.im;
	    y.data.c.im = c1.re * c2.im + c2.re * c1.im;
	    break;
	  case QH_Any:
	    if(y.type  >= QH_String ||
	       x->type >= QH_String)
		invalid_operation(p->op);
	    switch(SWITCH_DATA_TYPE(&y, x))
	    {
	      case QHDATATYPE_II:
		y.data.i *= x->data.i;
		break;

	      case QHDATATYPE_IR:
		y.type = QH_Real;
		y.data.r = (double)y.data.i * x->data.r;
		break;

	      case QHDATATYPE_IC:
		r = (double)y.data.i;
		y.type = QH_Complex;
		y.data.c.re = r * x->data.c.re;
		y.data.c.im = r * x->data.c.im;
		break;

	      case QHDATATYPE_RI:
		y.data.r *= (double)x->data.i;
		break;

	      case QHDATATYPE_RR:
		y.data.r *= x->data.r;
		break;

	      case QHDATATYPE_RC:
		r = y.data.r;
		y.type = QH_Complex;
		y.data.c.re = r * x->data.c.re;
		y.data.c.im = r * x->data.c.im;
		break;

	      case QHDATATYPE_CI:
		y.data.c.re *= (double)x->data.i;
		y.data.c.im *= (double)x->data.i;
		break;

	      case QHDATATYPE_CR:
		y.data.c.re *= x->data.r;
		y.data.c.im *= x->data.r;
		break;

	      case QHDATATYPE_CC:
		c1 = y.data.c;
		c2 = x->data.c;
		y.data.c.re = c1.re * c2.re - c1.im * c2.im;
		y.data.c.im = c1.re * c2.im + c2.re * c1.im;
		break;
	    }
	    break;
	}
	assign_array(a, align, &y);
    }
    else if(res->type == QH_String)
	STRING_OPERATION(s->str[align] *= cast_to_int(x));
    else
	execute_error("Atempted none-array operation");
}


/* a[] /= x */
void qh_adset(QHProg* p)
{
    QHObj* res;
    QHObj* x;
    QHObj y;
    Array* a;
    int align;
    XComplex c1, c2;
    double r, rr;

    x   = memstack + MEMADDR(p->a1);
    res = memstack + MEMADDR(p->res);

    if(res->type == QH_Array)
    {
	a = res->data.a;

	if(a->dim != p->a2)
	    execute_error("Mismatch array dimension");

	align = calc_array_entry(a);
	reference_array(a, align, &y);
	switch(a->type)
	{
	  case QH_Int:
	    y.data.i /= cast_to_int(x);
	    break;
	  case QH_Real:
	    y.data.r /= cast_to_real(x);
	    break;
	  case QH_Complex:
	    c1 = y.data.c;
	    c2 = cast_to_complex(x);
	    r  = c2.re * c2.re + c2.im * c2.im;
	    y.data.c.re = (c1.re * c2.re + c1.im * c2.im) / r;
	    y.data.c.im = (c2.re * c1.im - c1.re * c2.im) / r;
	    break;
      case QH_Any:
	    if(y.type  >= QH_String ||
	       x->type >= QH_String)
		invalid_operation(p->op);
	    switch(SWITCH_DATA_TYPE(&y, x))
	    {
	      case QHDATATYPE_II:
		y.data.i /= x->data.i;
		break;

	      case QHDATATYPE_IR:
		y.type = QH_Real;
		y.data.r = (double)y.data.i / x->data.r;
		break;

	      case QHDATATYPE_IC:
		r = x->data.c.re * x->data.c.re + x->data.c.im * x->data.c.im;
		rr = (double)y.data.i;
		y.type = QH_Complex;
		y.data.c.re =  rr * x->data.c.re / r;
		y.data.c.im = -rr * x->data.c.im / r;
		break;

	      case QHDATATYPE_RI:
		y.data.r /= (double)x->data.i;
		break;

	      case QHDATATYPE_RR:
		y.data.r /= x->data.r;
		break;

	      case QHDATATYPE_RC:
		r = x->data.c.re * x->data.c.re + x->data.c.im * x->data.c.im;
		rr = y.data.r;
		y.type = QH_Complex;
		y.data.c.re =  rr * x->data.c.re / r;
		y.data.c.im = -rr * x->data.c.im / r;
		break;

	      case QHDATATYPE_CI:
		r = (double)x->data.i;
		y.data.c.re /= r;
		y.data.c.im /= r;
		break;

	      case QHDATATYPE_CR:
		r = x->data.r;
		y.data.c.re /= r;
		y.data.c.im /= r;
		break;

	      case QHDATATYPE_CC:
		r = x->data.c.re * x->data.c.re + x->data.c.im * x->data.c.im;
		c1.re = (y.data.c.re * x->data.c.re
			 + y.data.c.im * x->data.c.im) / r;
		c1.im = (x->data.c.re * y.data.c.im
			 - y.data.c.re * x->data.c.im) / r;
		y.data.c = c1;
		break;
	    }
	    break;
	}
	assign_array(a, align, &y);
    }
    else if(res->type == QH_String)
	STRING_OPERATION(s->str[align] /= cast_to_int(x));
    else
	execute_error("Atempted none-array operation");
}

/* a[] %= x */
void qh_arset(QHProg* p)
{
    QHObj* res;
    QHObj* x;
    QHObj y;
    Array* a;
    int align;
    XComplex c1, c2;

    x   = memstack + MEMADDR(p->a1);
    res = memstack + MEMADDR(p->res);

    if(res->type == QH_Array)
    {
	a = res->data.a;

	if(a->dim != p->a2)
	    execute_error("Mismatch array dimension");

	align = calc_array_entry(a);
	reference_array(a, align, &y);
	switch(a->type)
	{
	  case QH_Int:
	    y.data.i %= cast_to_int(x);
	    break;
	  case QH_Real:
	    y.data.r = fmod(y.data.r, cast_to_real(x));
	    break;
	  case QH_Complex:
	    c1 = y.data.c;
	    c2 = cast_to_complex(x);
	    y.data.c.re = fmod(c1.re, c2.re);
	    y.data.c.im = fmod(c1.im, c2.im);
	    break;
	  case QH_Any:
	    if(y.type  >= QH_String ||
	       x->type >= QH_String)
		invalid_operation(p->op);
	    switch(SWITCH_DATA_TYPE(&y, x))
	    {
	      case QHDATATYPE_II:
		y.data.i %= x->data.i;
		break;

	      case QHDATATYPE_IR:
		y.type = QH_Real;
		y.data.r = fmod((double)y.data.i, x->data.r);
		break;

	      case QHDATATYPE_IC:
		y.type = QH_Complex;
		y.data.c.re = fmod((double)y.data.i, x->data.c.re);
		y.data.c.im = 0.0;
		break;

	      case QHDATATYPE_RI:
		y.data.r = fmod(y.data.r, (double)x->data.i);
		break;

	      case QHDATATYPE_RR:
		y.data.r = fmod(y.data.r, x->data.r);
		break;

	      case QHDATATYPE_RC:
		y.type = QH_Complex;
		y.data.c.re = fmod(y.data.c.re, x->data.c.re);
		y.data.c.im = 0.0;
		break;

	      case QHDATATYPE_CI:
		y.data.c.re = fmod(y.data.c.re, (double)x->data.i);
		y.data.c.im = fmod(y.data.c.re, (double)x->data.i);
		break;

	      case QHDATATYPE_CR:
		y.data.c.re = fmod(y.data.c.re, x->data.r);
		y.data.c.im = fmod(y.data.c.re, x->data.r);
		break;

	      case QHDATATYPE_CC:
		y.data.c.re = fmod(y.data.c.re, x->data.c.re);
		y.data.c.im = fmod(y.data.c.im, x->data.c.im);
		break;
	    }
	    break;
	}
	assign_array(a, align, &y);
    }
    else if(res->type == QH_String)
	STRING_OPERATION(s->str[align] %= cast_to_int(x));
    else
	execute_error("Atempted none-array operation");
}

/* a[]++ */
void qh_axinc(QHProg* p)
{
    Array* a;
    QHObj* x;
    QHObj* res;
    int align;
    QHObj y;
    int type;

    x   = memstack + MEMADDR(p->a1);
    res = memstack + MEMADDR(p->res);

    if(x->type == QH_Array)
    {
	res->type = QH_Array;

	a = res->data.a = x->data.a;

	if(a->dim != p->a2)
	    execute_error("Mismatch array dimension");

	align = calc_array_entry(a);

	reference_array(a, align, &y);
	*res = y;

	if(a->type == QH_Any)
	{
	    type = y.type;
	    if(type >= QH_String)
		invalid_operation(p->op);
	}
	else
	    type = a->type;

	switch(type)
	{
	  case QH_Int:
	    y.data.i++;
	    break;
	  case QH_Real:
	    y.data.r++;
	    break;
	  case QH_Complex:
	    y.data.c.re++;
	    break;
	}
	assign_array(a, align, &y);
    }
    else if(x->type == QH_String)
	STRING_OPERATION(s->str[align]++);
    else
	execute_error("Atempted none-array increment");
}

/* ++a[] */
void qh_aincx(QHProg* p)
{
    Array* a;
    QHObj* x;
    QHObj* res;
    int align;
    QHObj y;
    int type;

    x   = memstack + MEMADDR(p->a1);
    res = memstack + MEMADDR(p->res);

    if(x->type == QH_Array)
    {
	a = x->data.a;

	if(a->dim != p->a2)
	    execute_error("Mismatch array dimension");

	align = calc_array_entry(a);

	reference_array(a, align, &y);

	if(a->type == QH_Any)
	{
	    type = y.type;
	    if(type >= QH_String)
		invalid_operation(p->op);
	}
	else
	    type = a->type;

	switch(type)
	{
	  case QH_Int:
	    y.data.i++;
	    break;
	  case QH_Real:
	    y.data.r++;
	    break;
	  case QH_Complex:
	    y.data.c.re++;
	    break;
	}
	*res = y;
	assign_array(a, align, &y);
    }
    else if(x->type == QH_String)
	STRING_OPERATION(++s->str[align]);
    else
	execute_error("Atempted none-array increment");
}

/* a[]-- */
void qh_axdec(QHProg* p)
{
    Array* a;
    QHObj* x;
    QHObj* res;
    int align;
    QHObj y;
    int type;

    x   = memstack + MEMADDR(p->a1);
    res = memstack + MEMADDR(p->res);

    if(x->type == QH_Array)
    {
	a = x->data.a;

	if(a->dim != p->a2)
	    execute_error("Mismatch array dimension");

	align = calc_array_entry(a);

	reference_array(a, align, &y);
	*res = y;

	if(a->type == QH_Any)
	{
	    type = y.type;
	    if(type >= QH_String)
		invalid_operation(p->op);
	}
	else
	    type = a->type;

	switch(type)
	{
	  case QH_Int:
	    y.data.i--;
	    break;
	  case QH_Real:
	    y.data.r--;
	    break;
	  case QH_Complex:
	    y.data.c.re--;
	    break;
	}
	assign_array(a, align, &y);
    }
    else if(x->type == QH_String)
	STRING_OPERATION(s->str[align]--);
    else
	execute_error("Atempted none-array decrement");
}

/* --a[] */
void qh_adecx(QHProg* p)
{
    Array* a;
    QHObj* x;
    QHObj* res;
    int align;
    QHObj y;
    int type;

    x   = memstack + MEMADDR(p->a1);
    res = memstack + MEMADDR(p->res);

    if(x->type == QH_Array)
    {
	a = x->data.a;

	if(a->dim != p->a2)
	    execute_error("Mismatch array dimension");

	align = calc_array_entry(a);

	reference_array(a, align, &y);

	if(a->type == QH_Any)
	{
	    type = y.type;
	    if(type >= QH_String)
		invalid_operation(p->op);
	}
	else
	    type = a->type;

	switch(type)
	{
	  case QH_Int:
	    y.data.i--;
	    break;
	  case QH_Real:
	    y.data.r--;
	    break;
	  case QH_Complex:
	    y.data.c.re--;
	    break;
	}
	*res = y;
	assign_array(a, align, &y);
    }
    else if(x->type == QH_String)
	STRING_OPERATION(--s->str[align]);
    else
	execute_error("Atempted none-array decrement");
}

void qh_aref1(QHProg* p)
{
    long addr1, addr2, addr3;
    QHObj* x;
    QHObj* ind;
    QHObj* res;
    int align;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    ind   = memstack + addr2;
    res   = memstack + addr3;

    if(x->type == QH_Array)
    {
	Array* a = x->data.a;

	if(a->dim != 1)
	    execute_error("Mismatch array dimension");

	align = cast_to_int(ind);

	if(align >= a->nelem ||
	   align < 0)
	    execute_error("Out of array region");

	reference_array(a, align, res);
    }
    else if(x->type == QH_String)
    {
	QHString* s = x->data.s;
	res->type = QH_Int;

	if(s == NULL)
	    res->data.i = 0;
	else
	{
	    align = cast_to_int(ind);
	    if(align < 0 || align >= s->maxlen)
		res->data.i = 0;
	    else
		res->data.i = s->str[align];
	}
    }
    else
	execute_error("Atempted none-array reference");
}

Array* clone_array(const Array* a)
{
    Array* x;
    int i;
    int data_size;

    x = alloc_array();
    x->type = a->type;
    x->dim = a->dim;
    x->nelem = a->nelem;
    x->size = (int *)xmalloc(x->dim * sizeof(int));
    data_size = x->nelem * sizeof_qhtype(x->type);
    x->elem = xmalloc(data_size);

    for(i = 0; i < a->dim; i++)
	x->size[i] = a->size[i];

    if(x->type != QH_Any)
	bcopy(a->elem, x->elem, data_size);
    else
	for(i = 0; i < a->nelem; i++)
	    qh_copy_clone_object((QHObj *)x->elem + i, (QHObj *)a->elem + i);

    x->header.type = QH_Array;
    return x;
}

Bool equal_array(const Array* a1, const Array* a2)
{
    int i, n;

    if(a1->type  != a2->type ||
       a1->dim   != a2->dim  ||
       a1->nelem != a2->nelem)
	return False;

    for(i = 0; i < a1->dim; i++)
	if(a1->size[i] != a2->size[i])
	    return False;

    n = a1->nelem * sizeof_qhtype(a1->type);

    if(a1->type != QH_Any)
    {
	if(bcmp(a1->elem, a2->elem, n) != 0)
	    return False;
    }
    else
    {
	for(i = 0; i < a1->nelem; i++)
	    if(!qh_equal_object((QHObj *)a1->elem + i, (QHObj *)a2->elem + i))
		return False;
    }

    return True;
}
