#include <stdio.h>
#include "common.h"
#include "interp.h"
#include "QHObjAssocArr.h"

void qh_ascnew(QHProg* q)
{
    QHObj* res = memstack + MEMADDR(q->res);
    AssocArray* a;

    a = (AssocArray *)alloc_cell();
    a->tab = (QHObjAssocArr *)xmalloc(sizeof(QHObjAssocArr));

    makeQHObjAssocArr(a->tab, ASSOC_TABLE_SIZE);

    res->type = a->header.type = QH_AssocArray;
    res->data.assoc = a;
}

void qh_ascref(QHProg* p)
{
    QHObj* res = memstack + MEMADDR(p->res);
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* key = memstack + MEMADDR(p->a2);
    QHObj* t;
    char* s;

    if(x->type != QH_AssocArray)
	invalid_argument(x);

    if(!IS_QHSTRING(key))
	invalid_argument(key);

    s = key->data.s->str;
    if(s == NULL)
	s = "";
    t = searchQHObjAssocArr(x->data.assoc->tab, s);

    if(t != NULL)
	*res = *t;
    else
    {
	res->type = QH_Int;
	res->data.i = 0;
    }
}

void qh_ascset(QHProg* p)
{
    QHObj* res = memstack + MEMADDR(p->res);
    QHObj* key = memstack + MEMADDR(p->a1);
    QHObj* x   = memstack + MEMADDR(p->a2);
    char* s;

    if(res->type != QH_AssocArray)
	invalid_argument(res);

    if(!IS_QHSTRING(key))
	invalid_argument(key);

    s = key->data.s->str;
    if(s == NULL)
	s = "";
    installQHObjAssocArr(res->data.assoc->tab, s, *x);
}

#if defined(__GNUC__)
AssocArray* clone_assoc(AssocArray* a)
{
    AssocArray* clone_temp;

    /* Don't move `clone_temp' to global.
     * The follow function can be recursively called from
     * qh_copy_clone_object, then `clone_temp' must be different storage.
     */
    void clone_assoc_to_temp(const char* name, QHObj* value)
    {
	QHObj clone;

	qh_copy_clone_object(&clone, value);
	installQHObjAssocArr(clone_temp->tab, name, clone);
    }

    clone_temp = (AssocArray *)alloc_cell();
    clone_temp->tab = (QHObjAssocArr *)xmalloc(sizeof(QHObjAssocArr));
    makeQHObjAssocArr(a->tab, ASSOC_TABLE_SIZE);

    mapQHObjAssocArr(a->tab, clone_assoc_to_temp);

    clone_temp->header.type = QH_AssocArray;
    return clone_temp;
}

Bool equal_assoc(const AssocArray* a1, const AssocArray* a2)
{
    Bool equal_flag;
    const AssocArray* assoc_temp;

    void compare_assoc(const char* name, QHObj* value)
    {
	QHObj* obj;

	if(equal_flag == False)
	    return;

	obj = searchQHObjAssocArr(assoc_temp->tab, name);
	if(obj == NULL)
	{
	    equal_flag = False;
	    return;
	}

	equal_flag = qh_equal_object(obj, value);
    }

    equal_flag = True;
    assoc_temp = a1;
    mapQHObjAssocArr(a2->tab, compare_assoc);
    if(equal_flag == False)
	return False;
    assoc_temp = a2;
    mapQHObjAssocArr(a1->tab, compare_assoc);

    return equal_flag;
}
#else
struct stack_link
{
    Bool flag;
    AssocArray* ptr;
    struct stack_link* link;
};

static struct stack_link* stack_link = NULL;
static void clone_assoc_to_temp(const char* name, QHObj* value)
{
    QHObj clone;

    qh_copy_clone_object(&clone, value);
    installQHObjAssocArr(stack_link->ptr->tab, name, clone);
}

AssocArray* clone_assoc(AssocArray* a)
{
    struct stack_link stack;
    AssocArray* ret;

    ret = stack.ptr = (AssocArray *)alloc_cell();
    stack.link = stack_link;
    stack_link = &stack;

    ret->tab = (QHObjAssocArr *)xmalloc(sizeof(QHObjAssocArr));
    makeQHObjAssocArr(a->tab, ASSOC_TABLE_SIZE);
    mapQHObjAssocArr(a->tab, clone_assoc_to_temp);
    ret->header.type = QH_AssocArray;
    stack_link = stack_link->link;
    return ret;
}

static void compare_assoc(const char* name, QHObj* value)
{
    QHObj* obj;

    if(stack_link->flag == False)
	return;

    obj = searchQHObjAssocArr(stack_link->ptr->tab, name);
    if(obj == NULL)
    {
	stack_link->flag = False;
	return;
    }

    stack_link->flag = qh_equal_object(obj, value);
}

Bool equal_assoc(const AssocArray* a1, const AssocArray* a2)
{
    struct stack_link stack;
    Bool ret;

    stack.flag = True;
    stack.ptr = (AssocArray *)a1;
    stack.link = stack_link;
    stack_link = &stack;

    mapQHObjAssocArr(a2->tab, compare_assoc);
    if(stack.flag == False)
	ret = False;
    else
    {
	stack.ptr = (AssocArray *)a2;
	mapQHObjAssocArr(a1->tab, compare_assoc);
	ret = stack.flag;
    }

    stack_link = stack_link->link;
    return ret;
}
#endif
