/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <ctype.h>
#include "common.h"
#include "functab.h"
#include "interp.h"
#include "QHObjAssocArr.h"
#include "timer.h"
#include "fdrand48.h"

#if defined(sgi)
#include <strings.h>
#endif

int Q_FMT_INT, Q_FMT_REAL, Q_FMT_COMPLEX, Q_FMT_STRING;
int Q_ARGC;
int Q_NULL;
int Q_GEVENT_TYPE, Q_GEVENT_VALUE, Q_GPOINTERX, Q_GPOINTERY;
int Q_TIME;
int Q_GWINDOW_WIDTH, Q_GWINDOW_HEIGHT, Q_GWINDOW_PWIDTH, Q_GWINDOW_PHEIGHT;
int Q_GCURRENT_SCREEN;
int Q_GVISUAL_SCREEN;
int min_data_addr;
int qh_argstart;



void extern_unload_function(int argc, QHObj* argv, QHObj* ret)
{
    /* Do nothing. */
}

static const char* type_string(int type)
{
    switch(type)
    {
      case None:	return "void";
      case QH_Int:	return "int";
      case QH_Real:	return "real";
      case QH_Complex:	return "complex";
      case QH_String:	return "string";
      case QH_Any:	return "any";
      case QH_Array:	return "array";
      case QH_AssocArray:	return "assoc-array";
      default: return "undefined";
    }
}

void invalid_argument(QHObj* x)
{
    sprintf(temp_buff, "Invalid argument: %s\n", type_string(x->type));
    execute_error(temp_buff);
}

void qh_restore(void)
{
    QHFunc x;
    int i, endsp;

    x = popQHFuncStack(&fstk);

    endsp = qh_sp;
    for(i = x.sp; i < endsp; i++)
    {
	memstack[i].type   = QH_Int;
	memstack[i].data.i = 0;
    }

    qh_top = qh_sp = x.sp;
    qh_pc = x.pc;
}

char* obj_string(char* buff, QHObj* x)
{
    QHObj* fmt;
    buff[0] = '\0';

    switch(x->type)
    {
      case QH_Int:
	fmt = memstack + Q_FMT_INT;
	if(fmt->type != QH_String || fmt->data.s->str == NULL)
	    execute_error("Wrong integer format: FMT_INT");
	sprintf(buff, fmt->data.s->str, x->data.i);
	break;
      case QH_Real:
	fmt = memstack + Q_FMT_REAL;
	if(fmt->type != QH_String || fmt->data.s->str == NULL)
	    execute_error("Wrong integer format: FMT_REAL");
	sprintf(buff, fmt->data.s->str, x->data.r);
	break;
      case QH_Complex:
	fmt = memstack + Q_FMT_COMPLEX;
	if(fmt->type != QH_String || fmt->data.s->str == NULL)
	    execute_error("Wrong integer format: FMT_COMPLEX");
	sprintf(buff, fmt->data.s->str, x->data.c.re, x->data.c.im);
	break;
      case QH_String:
	fmt = memstack + Q_FMT_STRING;
	if(fmt->type != QH_String || fmt->data.s->str == NULL)
	    execute_error("Wrong integer format: FMT_STRING");
	if(fmt->data.s->str[0] == '%' &&
	   fmt->data.s->str[1] == 's' &&
	   fmt->data.s->str[2] == '\0')
	    strcpy(buff, x->data.s->str);
	else
	    sprintf(buff, fmt->data.s->str, x->data.s->str);
	break;
      case QH_Array:
	switch(x->data.a->type)
	{
	  case QH_Int:
	    sprintf(buff, "<int>");
	    break;
	  case QH_Real:
	    sprintf(buff, "<real>");
	    break;
	  case QH_Complex:
	    sprintf(buff, "<complex>");
	    break;
	}
	break;
      case QH_Keyword:
	if(x->data.i == QH_Key_Null)
	{
	    sprintf(buff, "(null)");
	    break;
	}
	/* through */
      default:
	fprintf(stderr, "## Warning: Can't write the object: %d\n", x->type);
	break;
    }

    return buff;
}

void qh_write(QHProg* prog)
{
    char* s;

    s = obj_string(temp_buff, memstack + MEMADDR(prog->a1));
    write(1, s, strlen(s));
}

void qh_typeof(QHObj* ret)
{
    ret->type = QH_Int;
    ret->data.i = STACK(1)->type;
}

void qh_atypeof(QHObj* ret)
{
    QHObj* x = STACK(1);

    if(x->type != QH_Array)
	execute_error("Array expected");

    ret->type = QH_Int;
    ret->data.i = x->data.a->type;
}

void qh_exit(QHObj* ret)
{
    qhexit(cast_to_int(STACK(1)));
}

void qh_dimension(QHObj* ret)
{
    QHObj* x = STACK(1);

    ret->type = QH_Int;

    if(x->type != QH_Array)
	ret->data.i = 0;
    else
	ret->data.i = x->data.a->dim;
}

void qh_arraysize(QHObj* ret)
{
    QHObj* x = STACK(1);

    ret->type = QH_Int;
    if(x->type != QH_Array)
	ret->data.i = 0;
    else
	ret->data.i = x->data.a->nelem;
}

void qh_argv(QHObj* ret)
{
    int i;

    i = cast_to_int(STACK(1));
    if(i < 0 || i >= prog_argc)
	ret->data.s = alloc_constant_string("");
    else
	ret->data.s = alloc_constant_string(prog_argv[i]);
    ret->type = QH_String;
}

void qh_system(QHObj* ret)
{
    ret->type = QH_Int;
    ret->data.i = system(cast_to_string(STACK(1)));
}

void qh_sleep(QHObj* ret)
{
    usleep((unsigned long)(cast_to_real(STACK(1)) * 1000000));
}

int cast_to_int(const QHObj* x)
{
    if(x->type == QH_Int)
	return x->data.i;
    if(x->type == QH_Real)
    {
	if(x->data.r < INT_MIN)
	    return INT_MIN;
	if(x->data.r > INT_MAX)
	    return INT_MAX;
	return (int)x->data.r;
    }
    if(x->type == QH_Complex)
	return (int)x->data.c.re;
    if(x->type == QH_String)
    {
	if(x->data.s->str == NULL)
	    return 0;
	return atoi(x->data.s->str);
    }
    if(x->type == QH_Keyword && x->data.i == QH_Key_Null)
	return 0;

    execute_error("Attempted converting no value object to int");
    return 0;
}


double cast_to_real(const QHObj* x)
{
    if(x->type == QH_Int)
	return (double)x->data.i;
    if(x->type == QH_Real)
	return x->data.r;
    if(x->type == QH_Complex)
	return x->data.c.re;
    if(x->type == QH_String)
    {
	if(x->data.s->str == NULL)
	    return 0.0;
	return atof(x->data.s->str);
    }
    if(x->type == QH_Keyword && x->data.i == QH_Key_Null)
	return 0.0;

    execute_error("Attempted converting no value object to real");
    return 0;
}

XComplex cast_to_complex(const QHObj* x)
{
    XComplex y;

    if(x->type == QH_Int)
    {
	y.re = (double)x->data.i;
	y.im = 0.0;
	return y;
    }

    if(x->type == QH_Real)
    {
	y.re = (double)x->data.r;
	y.im = 0.0;
	return y;
    }

    if(x->type == QH_Complex)
	return x->data.c;

    if(x->type == QH_String)
    {
	char* s = x->data.s->str;
	y.re = y.im = 0.0;

	if(s == NULL)
	    return y;

	while(isspace(*s))
	    s++;
	if(*s == '(')
	    s++;
	y.re = atof(s);
	while(*s && !isspace(*s))
	    ;
	if(*s == 0)
	    return y;
	y.im = atof(s + 1);
	return y;
    }

    execute_error("Attempted converting no value object to complex");

    return y;			/* dmy */
}

char* cast_to_string(const QHObj* x)
{
    if(!IS_QHSTRING(x))
	execute_error("String expected");
    return x->data.s->str;
}

void qh_rand(QHObj* ret)
{
    double min = cast_to_real(STACK(1));
    double max = cast_to_real(STACK(2));

    ret->type = QH_Real;
    ret->data.r = min + fdrand48() * (max - min);
}

void qh_srand(QHObj* ret)
{
    unsigned int seed = (unsigned int)cast_to_int(STACK(1));
    fsrand48(seed);
}

void qh_time(QHObj* ret)
{
    double t;

    t = get_current_calender_time();
    if(ret != NULL)		/* NULL for initialization of system clock */
    {
	ret->type = QH_Real;
	ret->data.r = t;
    }

    memstack[Q_TIME].type   = QH_Real;
    memstack[Q_TIME].data.r = t;
}

void qh_clock(QHObj* ret)
{
    if(ret == NULL)		/* NULL for initialization */
	set_laptime();
    else
    {
	ret->type = QH_Real;
	ret->data.r = get_laptime();
    }
}

void write_fname_and_lineno(QHProgTraceInfo* info)
{
    if(info->fname_entry != 0)
	fprintf(stderr, ":%s", text_file_entry_to_string(info->fname_entry));
    if(info->lineno != 0)
	fprintf(stderr, ", near line %d", info->lineno);
}

void qh_expand_cell_space(QHObj* res)
{
    expand_CELL_space(cast_to_int(STACK(1)));
}

#if defined(nec_ews) || defined(NO_USLEEP)
#include <sys/time.h>
/*
 * sleep microseconds
 */
static int usleep(unsigned long usec)
{
    struct timeval tv;

    tv.tv_sec  = usec / 1000000;
    tv.tv_usec = usec % 1000000;
    return select(0, NULL, NULL, NULL, &tv);
}

#endif /* nec_ews */

static void clear_qhstruct(QHObj* x)
{
    int i, n;
    QHObj* p;

    n = x->data.st->n;
    p = x->data.st->member;

    for(i = 0; i < n; i++)
    {
	p[i].type = QH_Int;
	p[i].data.i = 0;
    }
}

void qh_clear(QHObj* res)
{
    QHObj* x = STACK(1);

    switch(x->type)
    {
      case QH_Array:
	bzero((char *)x->data.a->elem,
	      x->data.a->nelem * sizeof_qhtype(x->data.a->type));
	break;
      case QH_AssocArray:
	deleteQHObjAssocArr(x->data.assoc->tab);
	makeQHObjAssocArr(x->data.assoc->tab, ASSOC_TABLE_SIZE);
	break;
      case QH_String:
	if(x->data.s->str == NULL)
	    break;
	if(x->data.s->isconst)
	    execute_error("Attempted modification of constant string");
	bzero(x->data.s->str, x->data.s->maxlen);
	break;
      case QH_Struct:
	clear_qhstruct(x);
	break;
/*      case QH_List:
	x->type = QH_Keyword;
	x->data.i = QH_Key_Null;
	break; */
      case QH_Int:
	x->data.i = 0;
	break;
      case QH_Real:
	x->data.r = 0.0;
	break;
      case QH_Complex:
	x->data.c.re = x->data.c.im = 0.0;
	break;
      default:
	invalid_argument(x);
    }
}

void qh_copy_clone_object(QHObj* target, const QHObj* src)
{
    if(src->type <= QH_Complex)
    {
	*target = *src;
	return;
    }

    switch(src->type)
    {
      case QH_Array:
	target->data.a = clone_array(src->data.a);
	target->type = QH_Array;
	break;
      case QH_AssocArray:
	target->data.assoc = clone_assoc(src->data.assoc);
	target->type = QH_AssocArray;
	break;
      case QH_String:
	target->data.s = clone_string(src->data.s);
	target->type = QH_String;
	break;
      case QH_Struct:
	target->data.st = clone_struct(src->data.st);
	target->type = QH_Struct;
	break;
      case QH_List:
	target->data.list = clone_list(src->data.list);
	target->type = QH_List;
	break;
      case QH_Keyword:
	*target = *src;
	break;
      default:
	fprintf(stderr, "Systematic error: Unknown data type %d, in qh_copy_clone_object().\n", src->type);
	exit(1);
    }
}

Bool qh_equal_object(const QHObj* obj1, const QHObj* obj2)
{
    if(obj1->type != obj2->type)
	return False;

    switch(obj1->type)
    {
      case QH_Int:
	return obj1->data.i == obj2->data.i;
      case QH_Real:
	return obj1->data.r == obj2->data.r;
      case QH_Complex:
	return (obj1->data.c.re == obj2->data.c.re) &&
	       (obj1->data.c.im == obj2->data.c.im);
      case QH_Array:
	return equal_array(obj1->data.a, obj2->data.a);
      case QH_AssocArray:
	return equal_assoc(obj1->data.assoc, obj2->data.assoc);
      case QH_String:
	return equal_string(obj1->data.s, obj2->data.s);
      case QH_Struct:
	return equal_struct(obj1->data.st, obj2->data.st);
      case QH_List:
	return equal_list(obj1->data.list, obj2->data.list);
      case QH_Keyword:
	return obj1->data.i == obj2->data.i;
    }

    fprintf(stderr, "Systematic error: Unknown data type %d, in qh_equal_object().\n", obj1->type);
    exit(1);
}


void qh_clone(QHObj* res)
{
    qh_copy_clone_object(res, STACK(1));
}

void qh_equal(QHObj* res)
{
    res->data.i = qh_equal_object(STACK(1), STACK(2));
    res->type = QH_Int;
}

void qh_call_extern_function(MemFunction* f, int argc, QHObj* argv, QHObj* ret)
{
#ifdef USE_DLD_LIBRARY
    if(f->extern_function == extern_unload_function)
    {
	sprintf(temp_buff, "qhf_%s", f->name);
	f->extern_function = (void (*)(int, QHObj*))dld_get_func(temp_buff);
	if(f->extern_function == NULL)
	{
	    fprintf(stderr,
		    "Error: undefined extern function: %s\n", temp_buff);
	    qhexit(1);
	}
    }
    f->extern_function(argc, argv, ret);
#endif
}

void qh_vla_bind_proc(long *narg_buff)
{
    QHObj* argv;
    QHObj* argc;
    Array* a;
    int narg;

    narg = *narg_buff;
    argc = memstack + qh_argstart + narg;
    argv = argc + 1;

    qh_sp += narg;
    qh_top = qh_sp;
    EXPAND_MEMORY_IF_OVERFLOW(QHObj, memstack, qh_top, memstack_size);

    argc->type = QH_Int;
    argc->data.i = narg;

    argv->type = None;	/* protect this object from G.C. */
    argv->data.a = a = alloc_array();
    argv->type = QH_Array;
    a->header.type = QH_Array;
    a->header.shared_data = True;
    a->type = QH_Any;
    a->dim = 1;
    a->nelem = narg;
    a->size = (int *)narg_buff;
    a->elem = (void *)(memstack + qh_argstart);
}

int sizeof_qhtype(int type)
{
    if(type == QH_Array ||
       type < 0 ||
       type >= QH_LastType)
    {
	fprintf(stderr, "Systemaitc error: Invalid sizeof_qhtype argument: %d\n", type);
	exit(1);
    }

    return object_size_list[type];
}
