/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "xmalloc.h"

#if defined(sgi) || defined(__osf__)
#define inline __inline
#elif !defined(__GNUC__)
#define inline
#endif

long RGB(Display *disp, int r, int g, int b);
/* r = [0, 255]
 * g = [0, 255]
 * b = [0, 255]
 */

long MyColor(Display* disp, const char* color);


static Bool parse_rgb(Display* disp, const char* s, int* r, int* g, int* b)
{
    XColor c;

    if(!XParseColor(disp, DefaultColormap(disp, 0), s, &c))
	return False;
    *r = c.red   / 257;
    *g = c.green / 257;
    *b = c.blue  / 257;
    return True;
}

static inline double calc_color_diff(int r1, int g1, int b1,
				     int r2, int g2, int b2)
{
    double rd, gd, bd;

    rd = r2 - r1;
    gd = g2 - g1;
    bd = b2 - b1;

    return rd * rd + gd * gd + bd * bd;
}


static long search_near_color(Display* disp, int r, int g, int b)
{
    Visual* vis;
    int vis_class;
    int depth;
    double d, mind;
    int scr;
    static XColor* xc;
    static int xc_size;
    long i, k;

    scr = DefaultScreen(disp);
    vis = DefaultVisual(disp, scr);
    depth = DefaultDepth(disp, scr);

#if defined(__cplusplus) || defined(c_plusplus)
    vis_class = vis->c_class;	/* C++ class of screen (monochrome, etc.) */
#else
    vis_class = vis->class;	/* class of screen (monochrome, etc.) */
#endif

    if(depth == 1)		/* black or white */
    {
	d = (double)r * r
	  + (double)g * g
	  + (double)b * b;

	if(d < 127.5 * 127.5 * 127.5)
	    return WhitePixel(disp, scr);
	return BlackPixel(disp, scr);
    }

    if(vis_class == TrueColor || vis_class == DirectColor)
    {
	return 0;		/* Can't search near color becose maybe too
				 * many colors are allocated.
				 */
    }

    if(xc_size == 0)
    {
	xc_size = DisplayCells(disp, scr);
	if((xc = (XColor *)xmalloc(sizeof(XColor) * xc_size)) == NULL)
	{
	    fprintf(stderr, "Warning: Can't allocate memory for color cells.\n");
	    xc_size = 0;
	    return 0;
	}
	for(i = 0; i < xc_size; i++)
	    xc[i].pixel = i;
    }

    XQueryColors(disp, DefaultColormap(disp, scr), xc, xc_size);

    r *= 257;
    g *= 257;
    b *= 257;
    
    mind = calc_color_diff(r, g, b, xc[0].red, xc[0].green, xc[0].blue);
    k = 0;
    for(i = 1; i < xc_size; i++)
    {
	d = calc_color_diff(r, g, b, xc[i].red, xc[i].green, xc[i].blue);
	if(d < mind)
	{
	    mind = d;
	    k = i;
	}
    }

    return k;
}

long RGB(Display *disp, int r, int g, int b)
{
    XColor c;

    c.red   = r * 257;
    c.green = g * 257;
    c.blue  = b * 257;
    if(!XAllocColor(disp, DefaultColormap(disp, 0), &c))
    {
	fprintf(stderr, "Warning: Can't allocate color #%02x%02x%02x.\n",
		r, g, b);
	return search_near_color(disp, r, g, b);
    }

    return c.pixel;
}

long MyColor(Display* disp, const char* color)
{
    int r, g, b;

    while(*color == ' ' ||
	  *color == '\t')
	color++;

    if(!parse_rgb(disp, color, &r, &g, &b))
    {
	fprintf(stderr, "Warning: Wrong color name `%s'.\n", color);
	return 0;
    }

    return RGB(disp, r, g, b);
}
