/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "interp.h"

static int align_offset = 1;
static char compile_function_name[BUFSIZ];
static intStack stk;
extern int current_module;

int new_align(void)
{
    return align_offset++;
}

static int symbol_label_address(Symbol* sym)
{
    if(sym->data.label_ref == NULL)
    {
	if(compile_function_name[0] == '\0')
	    sprintf(temp_buff, "Undefined label: %s", module_name_string(sym->module, sym->name));
	else
	    sprintf(temp_buff, "In function %s: Undefined lable: %s ", compile_function_name, module_name_string(sym->module, sym->name));
	yyerror3(temp_buff);
	return 0;
    }
    return sym->data.label_ref->pos;
}

static Quad2 quad1_to_quad2(Quad1* q1)
{
    Quad2 q;

    q.op = q1->op;
    if(IS_RELGOTOOP(q.op))
    {
	q.res = symbol_label_address(q1->res);
	q.a1 = q1->a1->data.align;
	q.a2 = q1->a2->data.align;
	return q;
    }

    if(IS_GOTOOP(q.op))
    {
	q.res = symbol_label_address(q1->res);
	if(q.op == GOTO)
	    q.a1 = 0;
	else
	    q.a1 = q1->a1->data.align;
	q.a2 = 0;
	return q;
    }

    if(q.op == CALL)
    {
	q.res = q1->res->data.align;
	q.a1  = q1->a1->data.func;
	q.a2  = q1->a2->name;
	return q;
    }

    if(q.op == CALLEXT)
    {
	q.res = q1->res->data.align;
	q.a1  = q1->a1->data.func;
	q.a2  = q1->a2->name;
	return q;
    }

    if(q.op == ASCNEW)
    {
	q.res = q1->res->data.align;
	q.a1 = q.a2 = 0;
	return q;
    }

    if(q.op == NEW)
    {
	q.res = q1->res->data.align;
	q.a2  = q1->a2->name;
	switch(q1->a1->name)
	{
	  case INT_OP:		q.a1 = QH_Int;     break;
	  case REAL_OP:		q.a1 = QH_Real;    break;
	  case COMPLEX_OP:	q.a1 = QH_Complex; break;
	  case STRING_OP:	q.a1 = QH_String;  break;
	  case ANY_SPEC:	q.a1 = QH_Any;     break;
	  default:
	    fprintf(stderr, "Systematic error: Unknown type(%ld) specified in new.\n",q1->a1->name);
	    exit(1);
	}

	return q;
    }

    q.res = (q1->res->type != SYM_KEYWORD) ? (q1->res->data.align) : (q1->res->name);
    q.a1  = (q1->a1 ->type != SYM_KEYWORD) ? (q1->a1 ->data.align) : (q1->a1 ->name);
    q.a2  = (q1->a2 ->type != SYM_KEYWORD) ? (q1->a2 ->data.align) : (q1->a2 ->name);

    return q;
}

static void quad1_to_quad2_list(QuadList* q)
{
    QuadNode* node;

    for(node = q->first; node != NULL; node = node->next)
    {
	node->u.q2 = quad1_to_quad2(&node->u.q1);
	node->type = Q_DATA2;
    }
}

static void give_number_quad(QuadList* q)
{
    QuadNode* node;
    int i;

    i = 1;
    for(node = q->first; node != NULL; node = node->next)
    {
	node->pos = i;
	i += (node->u.op != LABEL);
    }
}

static QuadList erase_label_node(QuadList* q)
{
    QuadNode* node;

    if(ISEMPTY_QUADLIST(q))
	return *q;

    quad_goto_first(q);
    do
    {
	if(cur_quad_node(q)->u.op != LABEL)
	{
	    quad_next(q);
	}
	else
	{
	    *q = delete_quad_list(q);
	}
    } while(!ISLAST_QUADLIST(q));

    node = cur_quad_node(q);
    if(node->u.op == LABEL)
    {
	node->u.op = NOP;
	node->u.q1.res = None;
	node->u.q1.a1 = None;
	node->u.q1.a1 = None;
    }

    return *q;
}

static void load_static_object(QHObj* m, Symbol* sym)
{
    switch(sym->type)
    {
      case SYM_VAR:
	*m = null_qhobj;
	break;
      case SYM_INT:
	m->type    = QH_Int;
	m->data.i  = atoi(STRING(sym->name));
	break;
      case SYM_REAL:
	m->type    = QH_Real;
	m->data.r  = atof(STRING(sym->name));
	break;
      case SYM_STR:
	m->type = QH_String;
	m->data.s = alloc_constant_string(STRING(sym->name));
	break;
      case SYM_KEYOBJ:
	m->type = QH_Keyword;
	m->data.i = sym->name;
	break;
      default:
	fprintf(stderr, "Systematic error: Bad static object type: %d\n", sym->type);
	qhexit(1);
	break;
    }
}

QuadList compile_global_code(QuadList* q)
{
    int i;
    Symbol* sym;
    FILE* sort_cmd = NULL;

    if(print_code_mode)
    {
	puts("**** static data ****");
	if((sort_cmd = popen("sort +1 -n", "w")) == NULL)
	{
	    perror("sort");
	    sort_cmd = stdout;
	}
    }

    compile_function_name[0] = '\0';

    min_data_addr = 0;
    for(i = 0; i < SYMTABSIZ; i++)
    {
	for(sym = constant_symbol[i]; sym != NULL; sym = sym->next)
	{
	    if(sort_cmd)
		fprintf(sort_cmd, "M[ %d ] %s\n", sym->data.align, symbol_name(sym));
	    load_static_object(memstack + sym->data.align, sym);
	    min_data_addr++;
	}
    }

    for(i = 0; i < SYMTABSIZ; i++)
    {
	for(sym = global_symbol[i]; sym != NULL; sym = sym->next)
	{
	    if(sort_cmd)
		fprintf(sort_cmd, "M[ %d ] %s\n", sym->data.align, symbol_name(sym));
	    load_static_object(memstack + sym->data.align, sym);
	}
    }

    if(sort_cmd != NULL && sort_cmd != stdout)
	pclose(sort_cmd);

    memstack[0].type = QH_Int;
    memstack[0].data.i = 0;

    qh_sp = align_offset;

    *q = add_quad_node_last(q, gen_quad_node1(END,
					      NULL,
					      NULL,
					      NULL));

    give_number_quad(q);
    quad1_to_quad2_list(q);
    *q = erase_label_node(q);

    return *q;
}

static void set_local_variable_align(QuadList* q, int* x)
{
    QuadNode* np;
    int align;
    int i;
    Symbol* sym;

    align = -1;
    for(np = q->first; np != NULL; np = np->next)
    {
	/* res */
	if(np->u.q1.res->type == SYM_VAR &&
	   np->u.q1.res->data.align == 0)
	{
	    np->u.q1.res->data.align = align;
	    align--;
	}

	/* a1 */
	if(np->u.q1.a1->type == SYM_VAR &&
	   np->u.q1.a1->data.align == 0)
	{
	    np->u.q1.a1->data.align = align;
	    align--;
	}

	/* a2 */
	if(np->u.q1.a2->type == SYM_VAR &&
	   np->u.q1.a2->data.align == 0)
	{
	    np->u.q1.a2->data.align = align;
	    align--;
	}
    }

    *x = -1 - align;

    for(i = 0; i < SYMTABSIZ; i++)
	for(sym = local_symbol[i]; sym != NULL; sym = sym->next)
	    if(sym->data.align > 0) /* function argument align */
		sym->data.align = -sym->data.align - *x; /* decreasing number of local variables */
}


void combine_vla_special_arguments(long module)
{
    Symbol* sym;

    sym = install_local_symbol(install_string("ARGV"), module);
    sym->data.align = 1;
    sym->initialized_flag = True;

    sym = install_local_symbol(install_string("ARGC"), module);
    sym->data.align = 2;
    sym->initialized_flag = True;
}

void compile_function(long name, long module, int narg, QuadList* q)
{
    Symbol* sym;
    Function* f;
    int nlocal;

    strcpy(compile_function_name, module_name_string(module, name));
    sym = install_function(name, module, narg);
    f = func_table + sym->data.func;

    if(f->state == F_Defined)
    {
	sprintf(temp_buff, "Redefined function `%s'",
		module_name_string(module, name));
	yyerror(temp_buff);
	reuse_quad_list(q);
	return;
    }

    f->state = F_Defined;

    set_local_variable_align(q, &nlocal);
    write_unused_local_variable();

    if(ISEMPTY_QUADLIST(q) ||
       q->last->u.op != RETURN)
	*q = add_quad_node_last(q, gen_quad_node1(RETURN,
						  NULL,
						  NULL,
						  NULL));

    give_number_quad(q);
    quad1_to_quad2_list(q);
    *q = erase_label_node(q);

    if(f->narg >= 0)
	f->offsp = nlocal + f->narg;
    else
	f->offsp = nlocal + 2;	/* +2 for argc and argv  */
    f->body = *q;
}

/* return next base point */
int load_function1(QuadNode* node, int base)
{
    int i, f;
    Quad2 t;
    QHTextPos tp;

    i = base;
    while(node)
    {
	t.op = node->u.op;
	tp = node->tp;
	switch(t.op)
	{
	  case GOTO:
	    t.res = base + node->u.q2.res;
	    t.a1 = 0;
	    t.a2 = 0;
	    break;
	  case IFGOTO:
	  case IFNGOTO:
	    t.res = base + node->u.q2.res;
	    t.a1 = node->u.q2.a1;
	    t.a2 = 0;
	    break;
	  case IFEQGOTO:
	  case IFNQGOTO:
	  case IFLEGOTO:
	  case IFLTGOTO:
	  case IFGEGOTO:
	  case IFGTGOTO:
	    t.res = base + node->u.q2.res;
	    t.a1 = node->u.q2.a1;
	    t.a2 = node->u.q2.a2;
	    break;
	  case CALL:
	  case CALLEXT:
	    t.res = node->u.q2.res;
	    t.a1 = node->u.q2.a1;
	    t.a2 = node->u.q2.a2;

	    f = node->u.q2.a1;
	    if(FUNCTAB(f).load_flag == False)
	    {
		FUNCTAB(f).load_flag = True;
		FUNCTAB(f).align     = nmemfunc++;
		pushintStack(&stk, f);
	    }
	    break;
	  default:
	    t = node->u.q2;
	    break;
	}
	i++;
	node = node->next;
	next_qhprog(&t, &tp);
    }

    return i;
}

static void load_memfunc(void)
{
    int qhqsiz;
    int i;
    int f, mf;

    qhqsiz = get_current_qhprog();

    memfunctab = (MemFunction *)xmalloc(sizeof(MemFunction) * nmemfunc);

    for(i = 0; i < nmemfunc; i++)
	memfunctab[i].addr = 0;

    for(i = 0; i < qhqsiz; i++)
    {
	if(qhprog[i].op == CALL ||
	   qhprog[i].op == CALLEXT)
	{
	    f = qhprog[i].a1;
	    mf = FUNCTAB(qhprog[i].a1).align;
	    qhprog[i].a1 = mf;
	    if(memfunctab[mf].addr == 0)
	    {
		memfunctab[mf].addr  = FUNCTAB(f).addr;
		memfunctab[mf].name  = strdup(STRING(FUNCTAB(f).name));

		if(FUNCTAB(f).module)
		    memfunctab[mf].module_name =
			strdup(STRING(FUNCTAB(f).module));
		else
		    memfunctab[mf].module_name = NULL;

		if(FUNCTAB(f).extern_function)
		    memfunctab[mf].extern_function = extern_unload_function;
		else
		    memfunctab[mf].extern_function = NULL;

		memfunctab[mf].narg  = FUNCTAB(f).narg;
		memfunctab[mf].offsp = FUNCTAB(f).offsp;
	    }
	}
    }

    for(i = 0; i < ndef_func; i++)
    {
	mf = FUNCTAB(i).align;
	if(FUNCTAB(i).loadfn == True &&
	   memfunctab[mf].addr == 0)
	{
	    memfunctab[mf].addr  = FUNCTAB(i).addr;
	    memfunctab[mf].name  = strdup(STRING(FUNCTAB(i).name));

	    if(FUNCTAB(i).module)
		memfunctab[mf].module_name = strdup(STRING(FUNCTAB(i).module));
	    else
		memfunctab[mf].module_name = NULL;

	    if(FUNCTAB(i).extern_function)
		memfunctab[mf].extern_function = extern_unload_function;
	    else
		memfunctab[mf].extern_function = NULL;
	    memfunctab[mf].narg  = FUNCTAB(i).narg;
	    memfunctab[mf].offsp = FUNCTAB(i).offsp;
	}
    }
}

void load_qhprog(QuadList* q)
{
    int base;
    Quad2 t;
    int f;
    int i;

    base = 0;
    nmemfunc  = 0;
    t.op = NOP;
    t.res = 0;
    t.a1 = 0;
    t.a2 = 0;
    next_qhprog(&t, NULL);
    makeintStack(&stk, 64);

    base = load_function1(q->first, base);

    for(i = 0; i < ndef_func; i++)
    {
	if(FUNCTAB(i).loadfn == True &&
	   FUNCTAB(i).load_flag == False)
	{
	    pushintStack(&stk, i);
	    FUNCTAB(i).load_flag = True;
	    FUNCTAB(i).align = nmemfunc++;
	}
    }

    while(!isemptyintStack(&stk))
    {
	f = popintStack(&stk);

	if(FUNCTAB(f).addr < 0)	/* built in function */
	    continue;

	if(FUNCTAB(f).extern_function) /* external function */
	    continue;

	if(FUNCTAB(f).body.first == NULL)
	{
	    int vlaf;

	    vlaf = install_function(FUNCTAB(f).name, FUNCTAB(f).module, -1)->data.func;
	    if(FUNCTAB(vlaf).body.first == NULL)
	    {
		sprintf(temp_buff, "Undefined function `%s/%d'",
			module_name_string(FUNCTAB(f).module, FUNCTAB(f).name),
			FUNCTAB(f).narg);
		yyerror3(temp_buff);
	    }
	    else
	    {
		if(FUNCTAB(vlaf).load_flag == False)
		{
		    FUNCTAB(vlaf).load_flag = True;
		    FUNCTAB(vlaf).loadfn    = True;
		    FUNCTAB(vlaf).align = nmemfunc++;
		    FUNCTAB(vlaf).addr = base + 1;
		    base = load_function1(FUNCTAB(vlaf).body.first, base);
		}
		FUNCTAB(f).addr = FUNCTAB(vlaf).addr;
		FUNCTAB(f).offsp = FUNCTAB(vlaf).offsp;
		FUNCTAB(f).narg = -1;
	    }
	}
	else
	{
	    FUNCTAB(f).addr = base + 1;
	    base = load_function1(FUNCTAB(f).body.first, base);
	}
    }

    if(nmemfunc > 0)
	load_memfunc();

    deleteintStack(&stk);
}

QuadList start_up_routine(void)
{
    QuadList q;
    Symbol* sym;

    init_quad_list(&q);

    sym = install_global_symbol(install_string("FMT_INT"), None);
    Q_FMT_INT = sym->data.align;
    sym->initialized_flag = True;
    sym->declared_flag = True;
    q = add_quadlistnode1_last(&q,
			       ASSIGN,
			       sym,
			       install_constant_symbol(install_string("%d"),
						       SYM_STR),
			       NULL);

    sym = install_global_symbol(install_string("FMT_REAL"), None);
    Q_FMT_REAL = sym->data.align;
    sym->initialized_flag = True;
    sym->declared_flag = True;
    q = add_quadlistnode1_last(&q,
			       ASSIGN, 
			       sym,
			       install_constant_symbol(install_string("%g"),
						       SYM_STR),
			       NULL);

    sym = install_global_symbol(install_string("FMT_COMPLEX"), None);
    Q_FMT_COMPLEX = sym->data.align;
    sym->initialized_flag = True;
    sym->declared_flag = True;
    q = add_quadlistnode1_last(&q,
			       ASSIGN, 
			       sym,
			       install_constant_symbol(install_string("(%g %g)"),
						       SYM_STR),
			       NULL);

    sym = install_global_symbol(install_string("FMT_STRING"), None);
    Q_FMT_STRING = sym->data.align;
    sym->initialized_flag = True;
    sym->declared_flag = True;
    q = add_quadlistnode1_last(&q,
			       ASSIGN, 
			       sym,
			       install_constant_symbol(install_string("%s"),
						       SYM_STR),
			       NULL);

    sym = install_global_symbol(install_string("ARGC"), None);
    Q_ARGC = sym->data.align;
    sym->initialized_flag = True;
    sym->declared_flag = True;
    q = add_quadlistnode1_last(&q,
			       ASSIGN, 
			       sym,
			       install_integer(prog_argc),
			       NULL);

    sym = install_constant_symbol(QH_Key_Null, SYM_KEYOBJ);
    Q_NULL = sym->data.align;

    sym = install_global_symbol(install_string("GEVENT_TYPE"), None);
    sym->initialized_flag = True;
    sym->declared_flag = True;
    Q_GEVENT_TYPE = sym->data.align;

    sym = install_global_symbol(install_string("GEVENT_VALUE"), None);
    sym->initialized_flag = True;
    sym->declared_flag = True;
    Q_GEVENT_VALUE = sym->data.align;

    sym = install_global_symbol(install_string("GPOINTERX"), None);
    sym->initialized_flag = True;
    sym->declared_flag = True;
    Q_GPOINTERX = sym->data.align;

    sym = install_global_symbol(install_string("GPOINTERY"), None);
    sym->initialized_flag = True;
    sym->declared_flag = True;
    Q_GPOINTERY = sym->data.align;

    sym = install_global_symbol(install_string("TIME"), None);
    sym->initialized_flag = True;
    sym->declared_flag = True;
    Q_TIME = sym->data.align;

    sym = install_global_symbol(install_string("GWINDOW_WIDTH"), None);
    sym->initialized_flag = True;
    sym->declared_flag = True;
    Q_GWINDOW_WIDTH = sym->data.align;

    sym = install_global_symbol(install_string("GWINDOW_HEIGHT"), None);
    sym->initialized_flag = True;
    sym->declared_flag = True;
    Q_GWINDOW_HEIGHT = sym->data.align;

    sym = install_global_symbol(install_string("GWINDOW_PWIDTH"), None);
    sym->initialized_flag = True;
    sym->declared_flag = True;
    Q_GWINDOW_PWIDTH = sym->data.align;

    sym = install_global_symbol(install_string("GWINDOW_PHEIGHT"), None);
    sym->initialized_flag = True;
    sym->declared_flag = True;
    Q_GWINDOW_PHEIGHT = sym->data.align;

    sym = install_global_symbol(install_string("GCURRENT_SCREEN"), None);
    sym->initialized_flag = True;
    sym->declared_flag = True;
    Q_GCURRENT_SCREEN = sym->data.align;

    sym = install_global_symbol(install_string("GVISUAL_SCREEN"), None);
    sym->initialized_flag = True;
    sym->declared_flag = True;
    Q_GVISUAL_SCREEN = sym->data.align;

    return q;
}
