/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "common.h"

#define SYSTEM_CPP "/lib/cpp"
static int cpp_pid;
extern char_ptrStack include_file_list;

static Bool check_file_executable(const char* path)
{
    struct stat buff;

    if(stat(path, &buff) < 0)
    {
	perror(path);
	return False;
    }
    if(buff.st_uid == getuid())
	return (buff.st_mode & S_IXUSR) == S_IXUSR;
    if(buff.st_gid == getgid())
	return (buff.st_mode & S_IXGRP) == S_IXGRP;
    return (buff.st_mode & S_IXOTH) == S_IXOTH;
}

FILE* cpp_popen(const char* filename, char* path, char** argv)
{
    int fd1[2];			/* parent -> child */
    int fd2[2];			/* parent <- child */
    FILE* fp;
    char buff[BUFSIZ];
    int i;

    if(pipe(fd1) < 0 || pipe(fd2) < 0)
    {
	perror("pipe");
	exit(1);
    }

    if(!check_file_executable(path))
    {
	fprintf(stderr, "Warning: %s can not execute. So use %s\n",
		path, SYSTEM_CPP);
	path    = SYSTEM_CPP;
	argv[0] = SYSTEM_CPP;
    }

    if((cpp_pid = fork()) < 0)
    {
	perror("fork");
	exit(1);
    }

    if(cpp_pid == 0)
    {
	if(dup2(fd2[1], 1) < 0)
	{
	    perror("dup2");
	    exit(1);
	}
	if(dup2(fd1[0], 0) < 0)
	{
	    perror("dup2");
	    exit(1);
	}
	close(fd1[0]);
	close(fd1[1]);
	close(fd2[0]);
	close(fd2[1]);

	execv(path, argv);
	perror(path);
	exit(1);
    }

    close(fd1[0]);
    close(fd2[1]);

    for(i = 0; i < include_file_list.ptr; i++)
    {
	sprintf(buff, "#include \"%s\"\n", include_file_list.elem[i]);
	write(fd1[1], buff, strlen(buff));
    }

    if(filename != NULL)
    {
	sprintf(buff, "#include \"%s\"\n", filename);
	write(fd1[1], buff, strlen(buff));
    }
    else
    {
	while((i = read(0, buff, BUFSIZ)) > 0)
	    write(fd1[1], buff, i);
    }
    close(fd1[1]);

    if((fp = fdopen(fd2[0], "r")) == NULL)
    {
	perror("fdopen");
	exit(1);
    }

    return fp;
}

void cpp_pclose(FILE* fp)
{
    int pid;
    int status;

    pid = fileno(fp);

    while(cpp_pid != wait(&status))
	;

    fclose(fp);
}
