#include <stdio.h>
#include "common.h"
#include "interp.h"
#include "defstruct.h"

/*

                      T_STRUCT(number of members)
                     /                          \
                    /                            \
                   /                              \
                  /                                \
                 /                                  \
        T_PARAM(number of declared)                  T_PRRAM
             /     \                                /       \
            /       \                              /         \
           /         \                            /           \
          /           \ [T_STRUCT or T_VAR or T_VAR = expr ]   T_PRRAM
         /             \                                        \
T_VAR(tag) or NULL      T_PARAM                                  ...
                       /   \
                      /     ... 
             IDENTIFIER

*/

#define FIRST_TABLE_SIZE 32

QHStructTagTable* struct_tag_table = NULL;
Tree** struct_tree = NULL;
static int struct_tag_table_size = FIRST_TABLE_SIZE;
static int next_table_entry_index;

static int next_entry(void)
{
    int i;

    if(struct_tag_table == NULL)
    {
	struct_tag_table
	    = (QHStructTagTable *)xmalloc(struct_tag_table_size
					  * sizeof(QHStructTagTable));
	struct_tree
	    = (Tree **)xmalloc(struct_tag_table_size * sizeof(Tree *));
	for(i = 0; i < struct_tag_table_size; i++)
	    struct_tree[i] = NULL;
    }
    else if(struct_tag_table_size == next_table_entry_index)
    {
	i = struct_tag_table_size;
	struct_tag_table_size *= 2;
	struct_tag_table
	    = (QHStructTagTable *)xrealloc(struct_tag_table,
					   struct_tag_table_size
					       * sizeof(QHStructTagTable));
	struct_tree
	    = (Tree **)xrealloc(struct_tree,
				struct_tag_table_size * sizeof(Tree *));
	for(; i < struct_tag_table_size; i++)
	    struct_tree[i] = NULL;
    }

    return next_table_entry_index++;
}

int current_struct_entry(void)
{
    return next_table_entry_index;
}

int install_new_struct_entry(int nmember, long* member_id)
{
    int entry;

    entry = next_entry();
    struct_tag_table[entry].nmember = nmember;
    struct_tag_table[entry].member_id = member_id;

    return entry;
}

void print_struct_table(void)
{
    int i, j;

    for(i = 0; i < next_table_entry_index; i++)
    {
	printf("S[%d]\t", i);
	for(j = 0; j < struct_tag_table[i].nmember; j++)
	{
	    printf("%d", struct_tag_table[i].member_id[j]);
	    if(struct_tag_table[i].next[j] != NULL)
		printf("(%d)", struct_tag_table[i].next[j] - struct_tag_table);
	    putchar('\t');
	}
	putchar('\n');
    }
}

static inline Bool is_defined_struct(Tree* node)
{
    return search_struct_symbol(node, False)->type != None;
}

static Bool check_new_struct(Tree* expr)
{
    if(expr == NULL)
	return True;
    if(expr->type != T_STTNEW)
	return check_new_struct(expr->left) && check_new_struct(expr->right);

    return is_defined_struct(expr->left);
}

static Bool check_struct_tree(Tree* t)
{
    Tree* p;

    if(t->val < 0)
    {
/*	if(!is_defined_struct(t->left->left))
	{
	    p = t->left->left;
	    if(p->tp.fname > 0)
		sprintf(temp_buff, "%s:Undefined struct `%s', near line %d",
			STRING(p->tp.fname),
			module_name_string(p->module, p->val),
			p->tp.lineno);
	    else
		sprintf(temp_buff, "Undefined struct `%s', near line %d",
			module_name_string(p->module, p->val),
			p->tp.lineno);
	    yyerror3(temp_buff);

	    return False;
	} */
	return True;
    }

    for(p = t->right; p; p = p->right)
    {
	if(p->left->type == T_STRUCT)
	    if(check_struct_tree(p->left) == False)
		return False;

	if(p->left->type == T_OP)
	    if(check_new_struct(p->left->right) == False)
		return False;
    }

    return True;
}

static void check_redefined_struct_member(long* member, int i, Tree* t)
{
    int j;

    for(j = 0; j < i; j++)
	if(member[j] == t->val)
	{
	    if(t->tp.fname > 0)
		sprintf(temp_buff, "%s:Redefined struct member `%s', near line %d",
			STRING(t->tp.fname),
			module_name_string(t->module, t->val),
			t->tp.lineno);
	    else
		sprintf(temp_buff, "Redefined struct member `%s', near line %d",
			module_name_string(t->module, t->val),
			t->tp.lineno);
	    yyerror3(temp_buff);
	}
}

static int make_new_struct_entry(Tree* t)
{
    long* member;
    QHStructTagTable** next;
    int i, n;
    Tree* p;
    int entry;

    if(t->val < 0)
    {
	Symbol* sym;

	sym = search_struct_symbol(t->left->left, False);

	if(sym->type == None)
	{
	    p = t->left->left;
	    if(p->tp.fname > 0)
		sprintf(temp_buff, "%s:Undefined struct `%s', near line %d",
			STRING(p->tp.fname),
			module_name_string(p->module, p->val),
			p->tp.lineno);
	    else
		sprintf(temp_buff, "Undefined struct `%s', near line %d",
			module_name_string(p->module, p->val),
			p->tp.lineno);
	    yyerror3(temp_buff);
	    return 0;
	}
	return sym->data.struct_entry;
    }

    n = t->val;
    member = (long *)xmalloc(n * sizeof(long));
    next = (QHStructTagTable **)xmalloc(n * sizeof(QHStructTagTable *));
    for(i = 0; i < n; i++)
	next[i] = NULL;


    i = 0;
    for(p = t->right; p; p = p->right)
    {
	if(p->left->type == T_STRUCT)
	{
	    Tree* q;

	    q = p->left->left->right;
	    entry = make_new_struct_entry(p->left);
	    for(q = p->left->left->right; q; q = q->right)
	    {
		check_redefined_struct_member(member, i, q->left);
		member[i] = q->left->val;
		next[i] = struct_tag_table + entry;
		i++;
	    }
	}
	else
	{
	    if(p->left->type == T_VAR)
	    {
		check_redefined_struct_member(member, i, p->left);
		member[i++] = p->left->val;
	    }
	    else
	    {
		check_redefined_struct_member(member, i, p->left->left);
		member[i++] = p->left->left->val;
	    }
	}
    }

    entry = next_entry();
    struct_tag_table[entry].nmember = n;
    struct_tag_table[entry].member_id = member;
    struct_tag_table[entry].next = next;

    return entry;
}

int define_struct(Tree* t)
{
    int entry;
    Tree* p;

    check_struct_tree(t);
    if(t->left->left != NULL)
    {
	Symbol* sym;
	p = t->left->left;

	sym = search_struct_symbol(p, False);

	if(sym->type != None)
	{
	    sprintf(temp_buff, "Redefined struct `%s'",
		    module_name_string(p->module, p->val));
	    yyerror(temp_buff);
	}

	entry = make_new_struct_entry(t);
	sym->type = SYM_STRUCT;
	sym->data.struct_entry = entry;
    }
    else
	entry = make_new_struct_entry(t);

    struct_tree[entry] = t;
    return entry;
}
