/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include "common.h"

#define MAX_NERROR 32		/* maxmum number of errors */

extern int nerror;		/* number of errors */
extern unsigned char *yytext;

#define IS_ASSIGN_EXPR(node) \
    ((node)->type == T_OP && IS_ASSIGNOP((node)->val))

#define NOT_STRING_OP(node)	   \
    (((node)->type == T_OP)     && \
     ((node)->val != ASSIGN)    && \
     ((node)->val != NOT)       && \
     ((node)->val != SIZEOF_OP) && \
     ((node)->val != EQ)        && \
     ((node)->val != NQ)        && \
     ((node)->val != ADD)       && \
     ((node)->val != INT_OP)    && \
     ((node)->val != REAL_OP)   && \
     ((node)->val != COMMANDOP))

#define IS_STRING(node) ((node)->type == T_CSTR)

#define EXIT_IF_TOO_MANY_ERRORS \
if(nerror > MAX_NERROR) \
{ \
    fprintf(stderr, "Sory too many errors\n"); \
    qhexit(1); \
}

int yyerror(const char* errmsg)
{
    if(qhfilename != 0)
	fprintf(stderr, "%s:", STRING(qhfilename));
    fprintf(stderr, "%s, near line %d", errmsg, qhlineno);
    nerror++;
    fprintf(stderr, ", before the token `%s'", yytext);

    fputs(".\n", stderr);

    EXIT_IF_TOO_MANY_ERRORS;

    return 0;			/* dmy */
}

void write_warning_with_line(const char* msg, int line)
{
    fprintf(stderr, "Warning:");
    if(qhfilename != 0)
	fprintf(stderr, "%s:", STRING(qhfilename));
    fprintf(stderr, "%s, near line %d\n", msg, line);
}

void write_warning(const char* msg)
{
    write_warning_with_line(msg, qhlineno);
}

void yyerror2(const char* errmsg, QHTextPos* tp)
{
    if(qhfilename != 0)
	fprintf(stderr, "%s:", STRING(tp->fname));
    fprintf(stderr, "%s, near line %d\n", errmsg, tp->lineno);
    nerror++;

    EXIT_IF_TOO_MANY_ERRORS;
}

void yyerror3(const char* errmsg)
{
    fprintf(stderr, "%s\n", errmsg);
    nerror++;

    EXIT_IF_TOO_MANY_ERRORS;
}

static Bool is_left_value(Tree* t)
{
    if(t->type  == T_VAR ||
       t->type  == T_ARRAY ||
       t->type  == T_ASSOCARRAY ||
       (t->type == T_OP &&
		(t->val == INDIR || t->val == STREF)))
	return True;
    if(t->type == T_COMMA)
	return is_left_value(t->right);
    return False;
}

static Bool check_expr1(Tree* t)
{
    if(t == NULL)
	return True;

    if(IS_ASSIGN_EXPR(t))
    {
	if(!is_left_value(t->left))
	{
	    if(BETWEEN(T_CINT, t->left->type, T_CSTR))
		yyerror2("constant assignment", &t->tp);
	    else
		yyerror2("invalid lvalue in assignment", &t->tp);
	    return False;
	}
    }
    if(NOT_STRING_OP(t))
    {
	if((t->left  != NULL && IS_STRING(t->left)) ||
	   (t->right != NULL && IS_STRING(t->right)))
	{
	    yyerror2("invalid operands", &t->tp);
	    return False;
	}
    }
    if(t->type == T_OP &&
       (IS_INCDECOP(t->val) ||
	t->val == ADDR))
    {
	if(!is_left_value(t->left))
	{
	    yyerror2("invalid lvalue operation", &t->tp);
	    return False;
	}
    }

    return check_expr1(t->left) && check_expr1(t->right);
}

Bool check_expr(Tree* t)
{
    return check_expr1(t);
}

Bool check_redeclaration_farg(intStack* st, int name)
{
    int i;

    for(i = 0; i < st->ptr; i++)
	if(st->elem[i] == name)
	{
	    sprintf(temp_buff, "Redeclaration function arguments `%s'", STRING(name));
	    yyerror(temp_buff);
	    return False;
	}
    return True;
}
