/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include "common.h"

Function* func_table;
MemFunction* memfunctab;
static int sizeof_table;
int ndef_func = 0;
int nmemfunc = 0;

void init_functab(void)
{
    sizeof_table = DEFAULT_MAX_NUMBER_OF_FUNCTIONS;
    ndef_func = 0;
    func_table = (Function *)xmalloc(sizeof(Function) * sizeof_table);

    install_built_in_function("typeof", TYPEOF, 1);
    install_built_in_function("atypeof", ATYPEOF, 1);
    install_built_in_function("system", SYSTEM, 1);
    install_built_in_function("sleep", SLEEP, 1);
    install_built_in_function("dimension", DIMENSION, 1);
    install_built_in_function("arraysize", ARRAYSIZE, 1);
    install_built_in_function("ARGV", ARGV, 1);
    install_built_in_function("exit", EXIT, 1);
    install_built_in_function("srand", SRAND, 1);
    install_built_in_function("time", TIME, 0);
    install_built_in_function("clock", CLOCK, 0);
    install_built_in_function("expand_CELL_space", EXPAND_CELL_SPACE, 1);
    install_built_in_function("clear", CLEAR, 1);
    install_built_in_function("clone", CLONE, 1);
    install_built_in_function("equal", EQUAL, 2);

#ifndef QH_TEXT_ONLY
    /* graphics functions */
    install_built_in_function("gopen", GOPEN, 4);
    install_built_in_function("gclose", GCLOSE, 0);
    install_built_in_function("gwindow", GWINDOW, 4);
    install_built_in_function("gscreen", GSCREEN, 1);
    install_built_in_function("gcopy", GCOPY, 2);
    install_built_in_function("gcopy_area", GCOPY_AREA, 8);
    install_built_in_function("gbacking_store", GBACKING_STORE, 1);
    install_built_in_function("gsync", GSYNC, 1);
    install_built_in_function("gflush", GFLUSH, 1);
    install_built_in_function("gline", GLINE, 4);
    install_built_in_function("gline", GLINE2, 2);
    install_built_in_function("gpset", GPSET, 2);
    install_built_in_function("gbox", GBOX, 4);
    install_built_in_function("gfbox", GFBOX, 4);
    install_built_in_function("gpoly", GPOLY, 2);
    install_built_in_function("gfpoly", GFPOLY, 2);
    install_built_in_function("gcircle", GCIRCLE, 3);
    install_built_in_function("gfcircle", GFCIRCLE, 3);
    install_built_in_function("garc", GARC, 6);
    install_built_in_function("gfarc", GFARC, 6);
    install_built_in_function("gcolor", GCOLOR_FG, 0);
    install_built_in_function("gcolor", GCOLOR_NAME, 1);
    install_built_in_function("gcolor", GCOLOR_RGB, 3);
    install_built_in_function("gpen", GPEN, 1);
    install_built_in_function("gpen", GPEN3, 3);
    install_built_in_function("gbutton", GBUTTON, 0);
    install_built_in_function("gkey", GKEY, 0);
    install_built_in_function("gpause", GPAUSE, 0);
    install_built_in_function("gpause", GPAUSE1, 1);
    install_built_in_function("gpointer", GPOINTER, 1);
    install_built_in_function("gstring", GSTRING, 3);
    install_built_in_function("gfont", GFONT, 1);
    install_built_in_function("gjfont", GJFONT, 1);
    install_built_in_function("gspline", GSPLINE, 2);
    install_built_in_function("gfspline", GFSPLINE, 2);
    install_built_in_function("gclear", GCLEAR, 1);
    install_built_in_function("gtitle_name", GTITLE_NAME, 1);
    install_built_in_function("gicon_name", GICON_NAME, 1);
    install_built_in_function("gpoint", GPOINT, 2);
    install_built_in_function("gbell", GBELL, 0);
    install_built_in_function("gtext_width", GTEXT_WIDTH, 1);
    install_built_in_function("gtext_height", GTEXT_HEIGHT, 0);
    install_built_in_function("glookup_color", GLOOKUP_COLOR, 2);
    install_built_in_function("gsetdash", GSETDASH, 1);
    install_built_in_function("gpixel_per_mmx", GPIXEL_PER_MMX, 0);
    install_built_in_function("gpixel_per_mmy", GPIXEL_PER_MMY, 0);
    install_built_in_function("gdisplay_width", GDISPLAY_WIDTH, 0);
    install_built_in_function("gdisplay_height", GDISPLAY_HEIGHT, 0);
    install_built_in_function("gresize_screen", GRESIZE_SCREEN, 3);
    install_built_in_function("gselect_input", GSELECT_INPUT, 1);
    install_built_in_function("gvisualize_screen", GVISUALIZE_SCREEN, 1);
#endif

    /* string function */
    install_built_in_function("strcpy", STRCPY, 2);
    install_built_in_function("strncpy", STRNCPY, 3);
    install_built_in_function("strcat", STRCAT, 2);
    install_built_in_function("strcmp", STRCMP, 2);
    install_built_in_function("strncmp", STRNCMP, 3);
    install_built_in_function("strlen", STRLEN, 1);
    install_built_in_function("strdup", STRDUP, 1);
    install_built_in_function("chop", CHOP, 1);
    install_built_in_function("atoi", ATOI, 1);
    install_built_in_function("ator", ATOR, 1);
    install_built_in_function("atoc", ATOC, 1);

    /* file operations */
    install_built_in_function("fopen", XFOPEN, 3);
    install_built_in_function("popen", XPOPEN, 2);
    install_built_in_function("fclose", XFCLOSE, 1);
    install_built_in_function("fwrite", XFWRITE2, 2);
    install_built_in_function("fwrite", XFWRITE3, 3);
    install_built_in_function("fread", XFREAD, 3);
    install_built_in_function("fgetc", XFGETC, 1);
    install_built_in_function("fputc", XFPUTC, 2);
    install_built_in_function("ftrunc", XFTRUNC, 2);
    install_built_in_function("fseek", XFSEEK, 3);
    install_built_in_function("fcreate", XFCREATE, 2);
    install_built_in_function("fdelete", XFDELETE, 1);

    /* list operations */
    install_built_in_function("car", CAR, 1);
    install_built_in_function("cdr", CDR, 1);
    install_built_in_function("cons", CONS, 2);
    install_built_in_function("rplaca", RPLACA, 2);
    install_built_in_function("rplacd", RPLACD, 2);
}

void free_funcquad(void)
{
    int i;

    for(i = 0; i < ndef_func; i++)
	reuse_quad_list(&FUNCTAB(i).body);
}

void free_functab(void)
{
    xfree(func_table);
}

int add_new_function(int narg, int name, int module)
{
    if(ndef_func == sizeof_table)
    {
	sizeof_table *= 2;
	func_table = (Function *)xrealloc(func_table, sizeof(Function) * sizeof_table);
    }

    FUNCTAB(ndef_func).name  = name;
    FUNCTAB(ndef_func).module= module;
    FUNCTAB(ndef_func).addr  = 0;
    FUNCTAB(ndef_func).narg  = narg;
    FUNCTAB(ndef_func).state = F_Undefined;
    FUNCTAB(ndef_func).body  = null_quad_list();
    FUNCTAB(ndef_func).load_flag = False;
    FUNCTAB(ndef_func).loadfn = False;
    FUNCTAB(ndef_func).extern_function = False;
    return ndef_func++;
}

Bool define_function(int func)
{
    if(FUNCTAB(func).state == F_Defined)
	return False;
    FUNCTAB(func).state = F_Defined;
    return True;
}

void must_load_function(long name, long module, int narg)
{
    Symbol* sym;

    sym = install_function(name, module, narg);
    FUNCTAB(sym->data.func).loadfn = True;
}

void declare_extern_function(int name)
{
    Symbol* func;
    int     f;

#ifndef USE_DLD_LIBRARY
    if(compile_flag == False)
    {
	/* External function must be compiled */
	yyerror("Can't declare external function");
	return;
    }
#endif

    func = install_function(name, None, 0);

    f = func->data.func;

    if(FUNCTAB(f).extern_function == True)
	return;			/* already decrared */

    FUNCTAB(f).extern_function = True;

    if(FUNCTAB(f).state == F_Defined)
    {
	sprintf(temp_buff, "Already defined function `%s'\n", STRING(name));
	yyerror(temp_buff);
	return;
    }

    FUNCTAB(f).state  = F_Defined;
    FUNCTAB(f).loadfn = True; /* force to load */
    FUNCTAB(f).addr   = -1;
}
