/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include "common.h"


static relgoto_negative[] =
{
    IFNQGOTO,
    IFEQGOTO,
    IFGTGOTO,
    IFGEGOTO,
    IFLTGOTO,
    IFLEGOTO,
};

QuadList gen_if_goto1(Tree* expr, Symbol* label)
{
    QuadList q1;
    QuadList q2;
    QuadList q;
    Symbol* sym1;
    Symbol* sym2;


    if(expr->type == T_OP)
    {
	if(IS_RELOP(expr->val))
	{
	    /* if(left relop right) goto label */
	    sym1 = tree_quadlist(expr->left, &q1);
	    sym2 = tree_quadlist(expr->right, &q2);
	    q = append_quadlist(&q1, &q2);
	    reuse_if_temp_sym(sym1);
	    reuse_if_temp_sym(sym2);
	    q = add_quadlistnode1_last(&q, expr->val - EQ + IFEQGOTO,
				       label, sym1, sym2);
	    q.last->tp = expr->tp;
	    return q;
	}

	if(expr->val == OR)
	{
	    /* if(left || right) goto label */
	    orgoto_quad1(expr->left, expr->right, label, &q);
	    return q;
	}

	if(expr->val == AND)
	{
	    /* if(left && right) goto label */
	    andgoto_quad1(expr->left, expr->right, label, &q);
	    return q;
	}

	if(expr->val == NOT)
	{
	    /* if(! expr) goto label */
	    return gen_ifn_goto1(expr->left, label);
	}
    }
    else if(BETWEEN(T_CINT, expr->type, T_CSTR))
    {
	if(expr->type == T_CINT &&
	   atoi(STRING(expr->val)) == 0)
	    return null_quad_list();

	if(expr->type == T_CREAL &&
	   atof(STRING(expr->val)) == 0.0)
	    return null_quad_list();

	return quad_node_list(GOTONODE(label));
    }


    /* if(expr != 0) goto label */
    sym1 = tree_quadlist(expr, &q);
    reuse_if_temp_sym(sym1);
    q = add_quadlistnode1_last(&q, IFGOTO, label, sym1, NULL);
    q.last->tp = expr->tp;

    return q;
}

QuadList gen_if_goto(Tree* expr, Symbol* label)
{
    QuadList q;

    q = gen_if_goto1(expr, label);
    reuse_tree(expr);

    return q;
}

QuadList gen_ifn_goto1(Tree* expr, Symbol* label)
{
    QuadList q1;
    QuadList q2;
    QuadList q;
    Symbol* sym1;
    Symbol* sym2;

    if(expr->type == T_OP)
    {
	if(IS_RELOP(expr->val))
	{
	    sym1 = tree_quadlist(expr->left, &q1);
	    sym2 = tree_quadlist(expr->right, &q2);
	    q = append_quadlist(&q1, &q2);
	    reuse_if_temp_sym(sym1);
	    reuse_if_temp_sym(sym2);
	    q = add_quadlistnode1_last(&q, relgoto_negative[expr->val - EQ],
				       label, sym1, sym2);
	    q.last->tp = expr->tp;
	    return q;
	}

	if(expr->val == OR)
	{
	    /* if(!(left || right)) goto label */
	    norgoto_quad1(expr->left, expr->right, label, &q);
	    return q;
	}

	if(expr->val == AND)
	{
	    /* if(!(left && right)) goto label */
	    nandgoto_quad1(expr->left, expr->right, label, &q);
	    return q;
	}

	if(expr->val == NOT)
	{
	    /* if(!(! expr)) goto label */
	    return gen_if_goto1(expr->left, label);
	}
    }
    else if(BETWEEN(T_CINT, expr->type, T_CSTR))
    {
	if(expr->type == T_CINT &&
	   atoi(STRING(expr->val)) == 0)
	    return quad_node_list(GOTONODE(label));

	if(expr->type == T_CREAL &&
	   atof(STRING(expr->val)) == 0.0)
	    return quad_node_list(GOTONODE(label));

	return null_quad_list();
    }


    sym1 = tree_quadlist(expr, &q);
    reuse_if_temp_sym(sym1);
    q = add_quadlistnode1_last(&q, IFNGOTO, label, sym1, NULL);
    q.last->tp = expr->tp;

    return q;
}

QuadList gen_ifn_goto(Tree* expr, Symbol* label)
{
    QuadList q;

    q = gen_ifn_goto1(expr, label);
    reuse_tree(expr);

    return q;
}
