/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/X10.h>
#include <sys/types.h>
#include <sys/time.h>
#include "gutil.h"
#if defined(sgi)
#include <strings.h>
#endif

#define MBMASK 0x7f;
#define IS_MB_INDEX(ch) ((unsigned)(ch) >= 0xa0)

#if XlibSpecificationRelease >= 6 || defined(nec_ews)
#define DISP_SOCKET_FD(disp) (((_XPrivDisplay)(disp))->fd);
#else
#define DISP_SOCKET_FD(disp) ((disp)->fd);
#endif

void set_backing_store(Display* disp, Window w, Bool use)
{
    XSetWindowAttributes at;

    if(use)
	at.backing_store = Always;
    else
	at.backing_store = NotUseful;
    XChangeWindowAttributes(disp, w, CWBackingStore, &at);
}

int glookup_char(XKeyEvent* e)
{
    char str[10];
    KeySym key;

    XLookupString(e, str, 10, &key, NULL);
    return key;
}


XFontStruct *gload_query_font(Display* disp, const char* name)
{
    XFontStruct* fs;

    if((fs = XLoadQueryFont(disp, name)) == NULL)
    {
	fprintf(stderr, "Warnning: Can't load font: %s\n", name);
	exit(1);
    }

    return fs;
}


int gmb_text_width(XFontStruct* fs1, XFontStruct* fs2, unsigned const char* string)
{
    int w;

    w = 0;
    while(*string)
    {
	if(IS_MB_INDEX(*string))
	{
	    w += XTextWidth16(fs2, (XChar2b *)string, 1);
	    string += 2;
	}
	else
	{
	    w += XTextWidth(fs1, string, 1);
	    string++;
	}
    }

    return w;
}

void gmb_draw_string(Display* disp, Drawable win, GC gc,
		     XFontStruct* fs1,
		     XFontStruct* fs2,
		     int x, int y,
		     unsigned const char* string)
{
    int dx, dy, dw, dh;
    XChar2b mb;
    int h1, h2;
    Bool mbstat;

    XFontStruct* lfs;
    XFontStruct* fs3;

    if(IS_MB_INDEX(*string))
	lfs = fs2;
    else
	lfs = fs1;

    h1 = FONTHEIGHT(fs1);
    h2 = FONTHEIGHT(fs2);
    if(h1 < h2)
    {
	fs3 = fs2;
	dh = h2;
    }
    else
    {
	fs3 = fs1;
	dh = h1;
    }

    dw = gmb_text_width(fs1, fs2, string);

    dx = x - lfs->min_bounds.lbearing;
    dy = y + fs3->max_bounds.ascent;

    if(IS_MB_INDEX(*string))
    {
	XSetFont(disp, gc, fs2->fid);
	mbstat = True;
    }
    else
    {
	XSetFont(disp, gc, fs1->fid);
	mbstat = False;
    }

    while(*string)
    {
	if(IS_MB_INDEX(*string))
	{
	    mb.byte1 = string[0] & MBMASK;
	    mb.byte2 = string[1] & MBMASK;
	    if(mbstat == False)
	    {
		XSetFont(disp, gc, fs2->fid);
		mbstat = True;
	    }
	    XDrawString16(disp, win, gc, dx, dy, &mb, 1);
	    dx += XTextWidth16(fs2, &mb, 1);
	    string += 2;
	}
	else
	{
	    if(mbstat == True)
	    {
		XSetFont(disp, gc, fs1->fid);
		mbstat = False;
	    }
	    XDrawString(disp, win, gc, dx, dy, string, 1);
	    dx += XTextWidth(fs1, string, 1);
	    string++;
	}
    }
}

void gget_pointer(Display* disp, Window win, int* x, int* y)
{
    Window wdmy;
    int idmy;
    unsigned int udmy;

    if(!XQueryPointer(disp, win, &wdmy, &wdmy, &idmy, &idmy, x, y, &udmy))
    {
	fprintf(stderr, "Can't query mouse pointer.\n");
	exit(1);
    }
}


unsigned long gget_event_mask(Display* disp, Window win)
{
    XWindowAttributes at;

    if(!XGetWindowAttributes(disp, win, &at))
    {
	fprintf(stderr, "Can't get window attributs.(ID: %lu)\n", win);
	exit(1);
    }

    return at.your_event_mask;
}

Bool gsleep_event(Display* disp, XEvent* e, double t)
{
    int width, fd, n;
    fd_set read_fd;
    struct timeval tout;

    XFlush(disp);
    if (XEventsQueued(disp, QueuedAlready) > 0) {
	XNextEvent(disp, e);
	return True;
    }

    fd = DISP_SOCKET_FD(disp);

    tout.tv_sec = (long)t;
    tout.tv_usec = (long)((t - tout.tv_sec) * 1000000);
    FD_ZERO(&read_fd);
    FD_SET(fd, &read_fd);
    width = fd + 1;

    if ((n = select(width, &read_fd, NULL, NULL, &tout)) < 0) {
	perror("select");
	exit(1);
    }

    if (n == 0) {
	bzero((char *)e, sizeof(XEvent));
	return False;
    }

    XNextEvent(disp, e);
    return True;
}

void gmake_bitpattern_dash(unsigned long x, int* s, char dash_list[], int* nelem)
{
    unsigned long mask;
    int i, n, nrot, status;

    /* initialize dash list */
    for(i = 0; i < 32; i++)
	dash_list[i] = 0;

    if(x == 0 || x == 0xffffffff)
    {
	fprintf(stderr,
		"Warning: Invalid arg 1 of gmake_bitpattern_dash().\n");
	*s = *nelem = 0;
	return;
    }

    /* calculate the start of pattern bit. */
    nrot = 0;
    while((x & 0x80000000) != 0x80000000 ||
	  (x & 0x00000001) == 0x00000001)
    {
	x = ((x >> 1) | (x << 31)); /* rotate bit to right */
	nrot++;
    }

    /* divide each bit group */
    status = 1;
    n = 0;
    for(mask = 0x80000000; mask; mask >>= 1)
    {
	if(status == 1)
	{
	    if(mask & x)
		dash_list[n]++;
	    else
	    {
		dash_list[++n]++;
		status = 0;
	    }
	}
	else
	{
	    if(mask & x)
	    {
		dash_list[++n]++;
		status = 1;
	    }
	    else
		dash_list[n]++;
	}
    }
    *nelem = n + 1;
    *s = nrot;
}
