/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"

#define REHASH    11
char* STRING(HASH_SIZE);
long Q_zero_int;
long Q_one_int;

static unsigned long hash(const char *str)
{
    unsigned long h = 0;

    while(*str)
    {
	h = ((h << 3) + (h >> 28)) + *str;
	str++;
    }

    return h % HASH_SIZE;
}

void init_hash_table(void)
{
    long i;

    for(i = 1; i < HASH_SIZE; i++)
	STRING(i) = NULL;
    STRING(None) = xstrdup("");

    Q_zero_int = install_string("0");
    Q_one_int = install_string("1");
}

void delete_hash_table(void)
{
    long i;

    for(i = 0; i < HASH_SIZE; i++)
	if(STRING(i) != NULL)
	    xfree(STRING(i));
}

long install_string(const char* str)
{
    long addr;
    long i;

    addr = hash(str);

    for(i = 0; i < HASH_SIZE; i++)
    {
	if(STRING(addr) == NULL)
	{
	    STRING(addr) = xstrdup(str);
	    return addr;
	}
	if(!strcmp(STRING(addr), str))
	    return addr;
	addr = (addr + REHASH) % HASH_SIZE;
    }

    fprintf(stderr, "Hash table over flow.\n");
    fprintf(stderr, "Please redefine `HASH_SIZE' in `hash.h', and recompile this program.\n");
    qhexit(1);

    return 0;			/* dmy */
}
