/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include "int.defs"
#include "intStack.h"


void makeintStack(intStack* st, int size)
{
    if(size <= 0)
    {
	fprintf(stderr, "intStack: Invalid stack size:%d\n", size);
	exit(1);
    }

    if((st->elem = (int *)xcalloc(size, sizeof(int))) == NULL)
    {
	fprintf(stderr, "intStack: Can't allocate a memory.\n");
	exit(1);
    }
    st->ptr = 0;
    st->size = size;
}

void deleteintStack(intStack* st)
{
    xfree(st->elem);
}

void resizeintStack(intStack* st, int size)
{
    if(size <= 0)
    {
	fprintf(stderr, "intStack: Invalid stack resize:%d\n", size);
	exit(1);
    }

    if((st->elem = (int *)xrealloc(st->elem, sizeof(int) * size)) == NULL)
    {
	fprintf(stderr, "intStack: Can't reallocate a memory.\n");
	exit(1);
    }

    st->size = size;
}

void emptyintStack(intStack* st)
{
    st->ptr = 0;
}

void pushintStack(intStack* st, int value)
{
    if(st->ptr == st->size)
	resizeintStack(st, st->size * 2);
    st->elem[st->ptr++] = value;
}

int topintStack(intStack* st)
{
    if(isemptyintStack(st))
    {
	fprintf(stderr, "intStack: Attempted topintStack() for empty stack.\n");
	exit(1);
    }

    return st->elem[st->ptr - 1];
}

int popintStack(intStack* st)
{
    if(isemptyintStack(st))
    {
	fprintf(stderr, "intStack: Attempted popintStack() for empty stack.\n");
	exit(1);
    }

    return st->elem[--st->ptr];
}
