/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "functab.h"
#include "interp.h"
#include "funcall.h"
#include "keyobj.h"
#include "timer.h"
#if defined(sgi)
#include <strings.h>
#endif

#define DEFAULT_QHPROG_SIZE 4096
/* for compile */
QHProg* qhprog = NULL;
QHProgTraceInfo* qhprog_traceinfo = NULL;
static int qhprog_size = DEFAULT_QHPROG_SIZE;
static int current_qhprog = 0;

int qh_pc;
int qh_sp;
int qh_top;
QHFuncStack fstk;
QHObj null_qhobj;
intStack istack;
QHObj* memstack;
int memstack_size;
int object_size_list[QH_LastType];

void print_memstack();

void init_interp(void)
{
    memstack_size = MEMSTACK_SIZE;
    memstack = (QHObj *)xmalloc(sizeof(QHObj) * memstack_size);
    bzero((char *)memstack, sizeof(QHObj) * memstack_size);

    bzero((char *)&null_qhobj, sizeof(null_qhobj));
    null_qhobj.type = QH_Int;

    object_size_list[QH_Int]     = sizeof(int);
    object_size_list[QH_Real]    = sizeof(double);
    object_size_list[QH_Complex] = sizeof(XComplex);
    object_size_list[QH_String]  = sizeof(char);
    object_size_list[QH_Any]     = sizeof(QHObj);
}

static void print_qhobj(QHObj* x)
{
    int i;

    switch(x->type)
    {
      case QH_Int:
	printf("%d", x->data.i);
	break;
      case QH_Real:
	printf("%.1e", x->data.r);
	break;
      case QH_Complex:
	printf("(%.1e %.1e)", x->data.c.re, x->data.c.im);
	break;
      case QH_String:
	if(x->data.s->str == NULL)
	    printf("\"\"");
	else
	    printf("\"%s\"", x->data.s->str);
	break;
      case QH_Array:
	printf("<");
	switch(x->data.a->type)
	{
	  case QH_Int:
	    printf("int");
	    break;
	  case QH_Real:
	    printf("real");
	    break;
	  case QH_Complex:
	    printf("complex");
	    break;
	}
	printf(">[");
	for(i = 0; i < x->data.a->dim - 1; i++)
	    printf("%d, ", x->data.a->size[i]);
	printf("%d]", x->data.a->size[i]);
	break;
    }
}

static void print_call_prog(QHProg* p, int depth)
{
    int narg, i;
    MemFunction* f;

    if(p->op == CALL)
    {
	f = memfunctab + p->a1;
	narg = f->narg;
	printf("[%d]-> %s/%d(", depth, f->name, narg);
    }
    else /* funcall */
    {
	if(!IS_QHSTRING(memstack + MEMADDR(p->a1)))
	    execute_error("funcall: Wrong type arg 1");
	narg = p->a2;
	printf("[%d]-> funcall(%s/%d", depth, cast_to_string(memstack + MEMADDR(p->a1)), narg);
	if(narg > 0)
	    printf(", ");
    }

    if(narg > 0)
    {

	for(i = narg - 1; i > 0; i--)
	{
	    print_qhobj(memstack + qh_sp + i);
	    printf(", ");
	}
	print_qhobj(memstack + qh_sp);

    }
    printf(")\n");
}

static void print_return_prog(const char* fname, int narg, int res, int depth)
{
    printf("[%d]<- ", depth);
    print_qhobj(memstack + res);
    printf(": %s/%d\n", fname, narg);
}


static double* tcall;
static int *ncall, nop;
static void result_prof(void)
{
    int i;
    double t;

    t = 0;
    for(i = 0; i < nop; i++)
	t += tcall[i];

    puts("msec.\tncall\tusec/n\ttime(%)\top\n");
    for(i = 0; i < nop; i++)
    {
	if(ncall[i])
	{
	    printf("%7.1f %7d %7.1f %7.1f %s\n",
		   tcall[i] * 1000.0,
		   ncall[i],
		   tcall[i] * 1000000.0 / ncall[i],
		   tcall[i] / t * 100,
		   opstr(i + FIRST_INSTRUCTION));
	}
    }
    qhexit(0);
}

static void prof_mode_interp(QHProg* p)
{
    int i, op;
    double t;

    nop = LAST_INSTRUCTION - FIRST_INSTRUCTION + 1;
    ncall = (int *)xmalloc(nop * sizeof(int));
    tcall = (double *)xmalloc(nop * sizeof(double));
    functab[END  - FIRST_INSTRUCTION] = result_prof;
    functab[EXIT - FIRST_INSTRUCTION] = result_prof;

    for(i = 0; i < nop; i++)
    {
	ncall[i] = 0;
	tcall[i] = 0.0;
    }

    while(1)
    {
	op = p[qh_pc].op;
	ncall[op - FIRST_INSTRUCTION]++;
	t = get_laptime();
	QH_CALL(op, (p + qh_pc));
	qh_pc++;
	tcall[op - FIRST_INSTRUCTION] += get_laptime() - t;
    }
}

static void interp_qh_prog(QHProg* p)
{
    qh_proc_t *code_list;
    int i;
    long x;

    code_list = (qh_proc_t *)xmalloc(current_qhprog * sizeof(qh_proc_t));
    for(i = 0; i < current_qhprog; i++)
	code_list[i] = functab[p[i].op - FIRST_INSTRUCTION];

    x = qh_pc;
    while(1)
    {
	(code_list[x])(p + x);
	x = ++qh_pc;
/*
	(code_list[x])(p + x);
	x = ++qh_pc;
	(code_list[x])(p + x);
	x = ++qh_pc;
	(code_list[x])(p + x);
	x = ++qh_pc;
	(code_list[x])(p + x);
	x = ++qh_pc; */
    }
}


void interp(QHProg* p)
{
    makeQHFuncStack(&fstk, 256);
    makeintStack(&istack, 256);
    make_funcall_table();

    /* initialize system clock */
    qh_time(NULL);
    qh_clock(NULL);

    OK_DO_garbage_collecting = True; /* allow garbage collection */

    if(qh_prof_mode)
    {
	prof_mode_interp(p);
	return;
    }

    /* interpreter */
    if(!trace_mode)
    {
#ifndef PROG_DEBUG
	interp_qh_prog(p);
#else
	while(1)
	{
	    print_memstack();

	    printf("## [%d]:%s sp:%d top:%d\n\n",
		   qh_pc,
		   opstr(p[qh_pc].op),
		   qh_sp, qh_top);
	    fflush(stdout);
	    QH_CALL(p[qh_pc].op, (p + qh_pc));
	    qh_pc++;
	}
#endif
    }
    else /* trace mode */
    {
	int depth = 0;
	int f;
	int res = 0;
	QHFunc func;

	while(1)
	{
	    if((p[qh_pc].op == CALL && memfunctab[p[qh_pc].a1].addr > 0) ||
	       p[qh_pc].op == FUNCALL)
	    {
		res = MEMADDR(p[qh_pc].res);
		print_call_prog(p + qh_pc, depth++);
	    }

	    if(p[qh_pc].op == FUNCALL)
	    {
		f = search_funcall_table(cast_to_string(memstack + MEMADDR(p[qh_pc].a1)),
					 p[qh_pc].a2);
		if(memfunctab[f].addr < 0)
		{
		    QH_CALL(p[qh_pc].op, (p + qh_pc));
		    qh_pc++;
		    print_return_prog(memfunctab[f].name, memfunctab[f].narg, res, --depth);
		    continue;
		}
	    }
	    else if(p[qh_pc].op == RETURN)
	    {
		func = topQHFuncStack(&fstk);
		f = func.func;
		QH_CALL(p[qh_pc].op, (p + qh_pc));
		qh_pc++;
		print_return_prog(memfunctab[f].name, memfunctab[f].narg, func.ret, --depth);
		continue;
	    }

	    QH_CALL(p[qh_pc].op, (p + qh_pc));
	    qh_pc++;
	}
    }
    OK_DO_garbage_collecting = False;
}

void execute_error(const char* errmsg)
{
    if(isemptyQHFuncStack(&fstk))
    {
	fprintf(stderr, "%s", errmsg);
    }
    else
    {
	QHFunc f;

	f = topQHFuncStack(&fstk);
	fprintf(stderr, "%s, in function `%s'", errmsg,
		memfunctab[f.func].name);
    }

    write_fname_and_lineno(qhprog_traceinfo + qh_pc);

    fprintf(stderr, ".\n");

    qhexit(1);
}

/* for debug */
void print_memstack1(QHObj* x)
{
    switch(x->type)
    {
      case QH_Int:
	printf("%d ", x->data.i);
	break;
      case QH_Real:
	printf("%.1f ", x->data.r);
	break;
      case QH_Complex:
	printf("(%.1f %.1f) ", x->data.c.re, x->data.c.im);
	break;
      case QH_Array:
	printf("<arr> ");
	break;
      case QH_String:
	printf("\"%s\" ", x->data.s->str);
	break;
      default:
	printf("?? ");
	break;
    }
}

void print_memstack(void)
{
    int i;

    printf("[ ");
    for(i = 0; i < qh_top; i++)
	print_memstack1(memstack + i);
    printf("]\n");
}

void invalid_operation(int op)
{
    sprintf(temp_buff, "Invalid internal operation `%s'", opstr(op));
    execute_error(temp_buff);
}

void next_qhprog(Quad2* q, QHTextPos* tp)
{
    if(qhprog == NULL)
    {
	qhprog = (QHProg *)xmalloc(sizeof(QHProg) * qhprog_size);
    }
    else
    {
	EXPAND_MEMORY_IF_OVERFLOW(QHProg, qhprog, current_qhprog, qhprog_size);
    }

    if(qhprog_traceinfo == NULL)
    {
	qhprog_traceinfo
	    = (QHProgTraceInfo *)xmalloc(sizeof(QHProg) * qhprog_size);
    }
    else
    {
	EXPAND_MEMORY_IF_OVERFLOW(QHProgTraceInfo, qhprog_traceinfo,
				  current_qhprog, qhprog_size);
    }

    qhprog[current_qhprog].op = q->op;
    qhprog[current_qhprog].res = q->res;
    qhprog[current_qhprog].a1 = q->a1;
    qhprog[current_qhprog].a2 = q->a2;

    if(tp != NULL)
    {
	qhprog_traceinfo[current_qhprog].lineno = tp->lineno;
	qhprog_traceinfo[current_qhprog].fname_entry
	    = install_text_filename(tp->fname, STRING(tp->fname));
    }
    else
    {
	qhprog_traceinfo[current_qhprog].lineno      = 0;
	qhprog_traceinfo[current_qhprog].fname_entry = 0;
    }
    current_qhprog++;
}

int get_current_qhprog(void)
{
    return current_qhprog;
}

#define ADDR_SFMT "%-9s"
#define FILENAME_FMT "%-14s"
#define LINENO_FMT "%-5s"

static void print_file_name(char* s)
{
    char* p = s;

    while(*p)
	p++;

    while(s < p && *p != '/')
	p--;
    printf(FILENAME_FMT, p + (*p == '/'));
}

static void print_qhfunc(int f)
{
    printf("F[%d]\t", f);

    if(memfunctab[f].module_name != NULL)
	sprintf(temp_buff, "%s::", memfunctab[f].module_name);
    else
	*temp_buff = '\0';
    strcat(temp_buff, memfunctab[f].name);
    printf("%-15s\t", temp_buff);

    if(memfunctab[f].extern_function)
    {
	printf("extern\t");
	printf("-\t");		/* narg */
	printf("-");		/* offsp */
	return;
    }

    if(memfunctab[f].addr > 0)
	printf("%d\t", memfunctab[f].addr);
    else if(memfunctab[f].addr < 0)
	printf("buit-in\t");
    else
	printf("undef\t");
    printf("%d\t", memfunctab[f].narg);
    printf("%d\n", memfunctab[f].offsp);
}

static void print_progarg(int a)
{
    if(a > 0)
	sprintf(temp_buff, "M[%d]", a);
    else if(a < 0)
	sprintf(temp_buff, "M[sp%d]", a);
    else
	sprintf(temp_buff, "-");
    printf(ADDR_SFMT, temp_buff);
}

static void print_prog(QHProg* p, int i)
{
    if(p->op == CALLEXT)
    {
	printf("[%d]\t%-11s", i, "callext");
	print_progarg(p->res);
	sprintf(temp_buff, "F[%ld]", p->a1);
	printf(ADDR_SFMT, temp_buff);
	printf("%ld\n", p->a2);
	return;
    }

    printf("[%d]\t%-11s", i, opstr(p->op));

    if(IS_GOTOOP(p->op))
    {
	sprintf(temp_buff, "%ld", p->res);
	printf(ADDR_SFMT, temp_buff);
	print_progarg(p->a1);
	print_progarg(p->a2);
    }
    else if(IS_ARRAYOP(p->op))
    {
	print_progarg(p->res);
	print_progarg(p->a1);
	printf("%ld", p->a2);
    }
    else if(p->op == CALL)
    {
	print_progarg(p->res);
	sprintf(temp_buff, "F[%ld]", p->a1);
	printf(ADDR_SFMT, temp_buff);
	printf("%ld\n", p->a2);
    }
    else if(p->op == FUNCALL)
    {
	print_progarg(p->res);
	print_progarg(p->a1);
	printf("%ld", p->a2);
    }
    else if(p->op == NEW)
    {
	print_progarg(p->res);
	switch(p->a1)
	{
	  case QH_Int:	sprintf(temp_buff, "<int>");   break;
	  case QH_Real:	sprintf(temp_buff, "<real>");  break;
	  case QH_Complex:	sprintf(temp_buff, "<cmplx>"); break;
	  case QH_String:	sprintf(temp_buff, "<str>");   break;
	  case QH_Any:	sprintf(temp_buff, "<>");       break;
	  default:          sprintf(temp_buff, "<void>");  break;
	}
	printf(ADDR_SFMT, temp_buff);
	printf("%ld", p->a2);
    }
    else if(p->op == STTNEW)
    {
	print_progarg(p->res);
	sprintf(temp_buff, "%d", p->a1);
	printf(ADDR_SFMT, temp_buff);
	printf(ADDR_SFMT, "-");
    }
    else if(p->op == STSET)
    {
	print_progarg(p->res);
	sprintf(temp_buff, "%d", p->a1);
	printf(ADDR_SFMT, temp_buff);
	print_progarg(p->a2);
    }
    else if(p->op == STREF)
    {
	print_progarg(p->res);
	print_progarg(p->a1);
	sprintf(temp_buff, "%d", p->a2);
	printf(ADDR_SFMT, temp_buff);
    }
    else if(p->op == LIST_OP)
    {
	print_progarg(p->res);
	sprintf(temp_buff, "%d", p->a1);
	printf(ADDR_SFMT, temp_buff);
	printf(ADDR_SFMT, "-");
    }
    else
    {
	print_progarg(p->res);
	print_progarg(p->a1);
	print_progarg(p->a2);
    }
    putchar('\n');
}

void print_qhprog(void)
{
    int i, nprogcode, file_exist_flag;

    putchar('\n');
    puts("**** struct table ****");
    print_struct_table();

    putchar('\n');
    puts("**** functable ****");
    puts("\tname\t\taddr\tnarg\toffsp");
    for(i = 0; i < nmemfunc; i++)
	print_qhfunc(i);
    putchar('\n');

    puts("**** prog-list ****");

    file_exist_flag = 0;
    nprogcode = get_current_qhprog();
    for(i = 0; i < nprogcode; i++)
	if(qhprog_traceinfo[i].fname_entry != 0)
	{
	    file_exist_flag = 1;
	    break;
	}

    for(i = 0; i < nprogcode; i++)
    {
	if(file_exist_flag)
	{
	    if(qhprog_traceinfo[i].fname_entry != 0)
		sprintf(temp_buff, "%-7s ",
		    text_file_entry_to_string(qhprog_traceinfo[i].fname_entry));
	    else
		sprintf(temp_buff, "        ");
	    print_file_name(temp_buff);
	}

	if(qhprog_traceinfo[i].lineno != 0)
	    sprintf(temp_buff, "%d", qhprog_traceinfo[i].lineno);
	else
	    temp_buff[0] = '\0';

	printf(LINENO_FMT, temp_buff);
	
	print_prog(qhprog + i, i);
    }
    putchar('\n');
}
