/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include "common.h"
#include "localvar.h"

/* #define DEBUG */

/* binding list
    -> ((v11, v12, ..., v1n) (v21, v22, ..., v2n') ... ) */

/* for debug */
#ifdef DEBUG
static void print_binding_list(void);
#else
#define nonnil_list()
#endif

static Tree* binding_list = NULL;

static Tree* alloc_binding_node_cell(void)
{
    Tree* node;

    node = alloc_tree();
    node->type = T_PARAM;

    return node;
}

void push_new_binding_list(void)
{
    Tree* node;

    node = alloc_binding_node_cell();
    node->right = binding_list;

    binding_list = node;
}

#ifdef DEBUG
static void nonnil_list(void)
{
    if(binding_list == NULL)
    {
	fprintf(stderr, "Systematic error: local binding stack is empty.\n");
	exit(1);
    }
}
#endif

void pop_new_binding_list(void)
{
    Tree* node;
    Tree* p;

    nonnil_list();

    node = binding_list;
    binding_list = binding_list->right;

    for(p = node->left; p != NULL; p = p->right)
    {
	local_unbind_symbol_name(p->val, p->module);
    }

    node->right = NULL;
    reuse_tree(node);
}

static void check_rebind_vars(Tree* t)
{
    Tree* x;
    Tree* y;

    for(x = t; x != NULL; x = x->right)
	for(y = x->right; y != NULL; y = y->right)
	    if(x->left->val == y->left->val)
	    {
		sprintf(temp_buff, "Redefined local variable `%s'", STRING(x->left->val));
		write_warning(temp_buff);
	    }
}

void bind_local_variable(long name, long module)
{
    Tree* node;

    nonnil_list();

    node = alloc_binding_node_cell();
    node->val = name;
    node->right = binding_list->left;
    binding_list->left = node;

    local_bind_symbol_name(name, module, newtemp_symbol());
}

void bind_variable_in_param_tree(Tree* t)
{
    check_rebind_vars(t);
    for(; t != NULL; t = t->right)
	bind_local_variable(t->left->val, t->left->module);
}

void unbind_variable_in_param_tree(Tree* t)
{
    for(; t != NULL; t = t->right)
    {
	local_unbind_symbol_name(t->left->val, t->left->module);
    }
}


/* for debug */
#ifdef DEBUG
static void print_binding_list(void)
{
    Tree* p1;
    Tree* p2;

    printf("( ");
    for(p1 = binding_list; p1 != NULL; p1 = p1->right)
    {
	p2 = p1->left;
	if(p2 == NULL)
	    printf("nil ");
	else
	{
	    printf("( ");
	    while(p2)
	    {
		printf("%s ", STRING(p2->val));
		p2 = p2->right;
	    }
	    printf(") ");
	}
    }

    printf(")\n");
}
#endif
