/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "common.h"
#include "functab.h"
#include "interp.h"

int qh_extern_argc = 0;

#ifdef USE_DLD_LIBRARY
#include <dld.h>
#endif

void qh_first_instruction(QHProg* p)
{
}

void qh_last_operator(QHProg* p)
{
}

void qh_nop(QHProg* p)
{
}

void qh_assign(QHProg* p)
{
    int a1, a2;

    a1 = MEMADDR(p->res);
    a2 = MEMADDR(p->a1);
    memstack[a1] = memstack[a2];
}

void qh_ipush(QHProg* p)
{
    pushintStack(&istack, cast_to_int(memstack + MEMADDR(p->a1)));
}

void qh_ipush2(QHProg* p)
{
    long addr1, addr2;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    pushintStack(&istack, cast_to_int(memstack + addr1));
    pushintStack(&istack, cast_to_int(memstack + addr2));
}

void qh_ipush3(QHProg* p)
{
    long addr1, addr2, addr3;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    pushintStack(&istack, cast_to_int(memstack + addr1));
    pushintStack(&istack, cast_to_int(memstack + addr2));
    pushintStack(&istack, cast_to_int(memstack + addr3));
}

void qh_param(QHProg* p)
{
    EXPAND_MEMORY_IF_OVERFLOW(QHObj, memstack, qh_top, memstack_size);
    memstack[qh_top++] = memstack[MEMADDR(p->a1)];
}

void qh_param2(QHProg* p)
{
    int addr1, addr2;

    EXPAND_MEMORY_IF_OVERFLOW(QHObj, memstack, qh_top + 1, memstack_size);

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);

    memstack[qh_top++] = memstack[addr1];
    memstack[qh_top++] = memstack[addr2];
}

void qh_param3(QHProg* p)
{
    long addr1, addr2, addr3;

    EXPAND_MEMORY_IF_OVERFLOW(QHObj, memstack, qh_top + 2, memstack_size);

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);

    memstack[qh_top++] = memstack[addr1];
    memstack[qh_top++] = memstack[addr2];
    memstack[qh_top++] = memstack[addr3];
}

void qh_callext(QHProg* p)
{
#ifdef USE_DLD_LIBRARY
    int narg;
    MemFunction* f;
    QHFunc x;

    narg = p->a2;
    f = memfunctab + p->a1;
    x.func = p->a1;
    x.pc = qh_pc;
    x.ret = MEMADDR(p->res);
    x.sp = qh_sp;

    pushQHFuncStack(&fstk, x);

    qh_extern_argc = narg;
    qh_argstart = qh_sp;
    qh_sp += narg;
    qh_top = qh_sp;

    qh_call_extern_function(f, narg, memstack + qh_argstart, memstack + x.ret);
    qh_restore();
#endif
}

void qh_call(QHProg* p)
{
    MemFunction* f;
    QHFunc x;
    f = memfunctab + p->a1;
    x.func = p->a1;
    x.pc = qh_pc;
    x.ret = MEMADDR(p->res);
    x.sp = qh_sp;

    pushQHFuncStack(&fstk, x);

    qh_argstart = qh_sp;
    qh_sp += f->offsp;
    qh_top = qh_sp;
    EXPAND_MEMORY_IF_OVERFLOW(QHObj, memstack, qh_top, memstack_size);

    if(f->addr > 0)
    {
	qh_pc = f->addr - 1;
	if(f->narg < 0)
	    qh_vla_bind_proc(&p->a2);
    }
    else			/* call built in function */
    {
	if(f->extern_function == NULL)
	{
	    QH_CALL(-f->addr, (memstack + x.ret));
	    qh_restore();
	}
    }
}

void qh_ifgoto(QHProg* p)
{
    QHObj* x;

    x = memstack + MEMADDR(p->a1);

    switch(x->type)
    {
      case QH_Int:
	if(x->data.i != 0)
	    qh_pc = p->res - 1;
	break;
      case QH_Real:
	if(x->data.r != 0.0)
	    qh_pc = p->res - 1;
	break;
      case QH_Complex:
	if(x->data.c.re != 0.0 || x->data.c.im != 0.0)
	    qh_pc = p->res - 1;
	break;
      case QH_Keyword:
	if(x->data.i != QH_Key_Null)
	    qh_pc = p->res - 1;
	break;
      default:
	qh_pc = p->res - 1;
	break;
    }
}

void qh_ifngoto(QHProg* p)
{
    QHObj* x;

    x = memstack + MEMADDR(p->a1);

    switch(x->type)
    {
      case QH_Int:
	if(x->data.i == 0)
	    qh_pc = p->res - 1;
	break;
      case QH_Real:
	if(x->data.r == 0.0)
	    qh_pc = p->res - 1;
	break;
      case QH_Complex:
	if(x->data.c.re == 0.0 && x->data.c.im == 0.0)
	    qh_pc = p->res - 1;
	break;
      case QH_Keyword:
	if(x->data.i == QH_Key_Null)
	    qh_pc = p->res - 1;
	break;
    }
}

void qh_ifeqgoto(QHProg* p)
{
    long addr1, addr2;
    QHObj* a1;
    QHObj* a2;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    a1 = memstack + addr1;
    a2 = memstack + addr2;

    switch(SWITCH_DATA_TYPE(a1, a2))
    {
      case QHDATATYPE_II:
	if(a1->data.i == a2->data.i)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_IR:
	if(a1->data.i == a2->data.r)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_IC:
	if(a1->data.i    == a2->data.c.re &&
	   a2->data.c.im == 0)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_RI:
	if(a1->data.r == a2->data.i)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_RR:
	if(a1->data.r == a2->data.r)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_RC:
	if(a1->data.r    == a2->data.c.re &&
	   a2->data.c.im == 0)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_CI:
	if(a1->data.c.re == a2->data.i &&
	   a1->data.c.im == 0)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_CR:
	if(a1->data.c.re == a2->data.r &&
	   a1->data.c.im == 0)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_CC:
	if(a1->data.c.re == a2->data.c.re &&
	   a1->data.c.im == a2->data.c.im)
	    qh_pc = p->res - 1;
	return;
    }

    if(a1->type   == a2->type &&
       a1->data.v == a2->data.v)
	qh_pc = p->res - 1;
}

void qh_ifnqgoto(QHProg* p)
{
    long addr1, addr2;
    QHObj* a1;
    QHObj* a2;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    a1 = memstack + addr1;
    a2 = memstack + addr2;

    switch(SWITCH_DATA_TYPE(a1, a2))
    {
      case QHDATATYPE_II:
	if(a1->data.i != a2->data.i)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_IR:
	if(a1->data.i != a2->data.r)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_IC:
	if(a1->data.i    != a2->data.c.re ||
	   a2->data.c.im != 0)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_RI:
	if(a1->data.r != a2->data.i)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_RR:
	if(a1->data.r != a2->data.r)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_RC:
	if(a1->data.r    != a2->data.c.re ||
	   a2->data.c.im != 0)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_CI:
	if(a1->data.c.re != a2->data.i ||
	   a1->data.c.im != 0)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_CR:
	if(a1->data.c.re != a2->data.r ||
	   a1->data.c.im != 0)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_CC:
	if(a1->data.c.re != a2->data.c.re ||
	   a1->data.c.im != a2->data.c.im)
	    qh_pc = p->res - 1;
	return;
    }

    if(a1->type != a2->type ||
       a1->data.v != a2->data.v)
	qh_pc = p->res - 1;
}

void qh_iflegoto(QHProg* p)
{
    long addr1, addr2;
    QHObj* a1;
    QHObj* a2;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    a1 = memstack + addr1;
    a2 = memstack + addr2;

    switch(SWITCH_DATA_TYPE(a1, a2))
    {
      case QHDATATYPE_II:
	if(a1->data.i <= a2->data.i)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_IR:
	if((double)a1->data.i <= a2->data.r)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_IC:
	if((double)a1->data.i * (double)a1->data.i <= QHCABS2(a2))
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_RI:
	if(a1->data.r <= (double)a2->data.i)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_RR:
	if(a1->data.r <= a2->data.r)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_RC:
	if(a1->data.r * a1->data.r <= QHCABS2(a2))
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_CI:
	if(QHCABS2(a1) <= (double)a2->data.i * (double)a2->data.i)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_CR:
	if(QHCABS2(a1) <= a2->data.r * a2->data.r)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_CC:
	if(QHCABS2(a1) <= QHCABS2(a2))
	    qh_pc = p->res - 1;
	return;
    }
    invalid_operation(p->op);
}

void qh_ifltgoto(QHProg* p)
{
    long addr1, addr2;
    QHObj* a1;
    QHObj* a2;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    a1 = memstack + addr1;
    a2 = memstack + addr2;

    switch(SWITCH_DATA_TYPE(a1, a2))
    {
      case QHDATATYPE_II:
	if(a1->data.i < a2->data.i)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_IR:
	if((double)a1->data.i < a2->data.r)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_IC:
	if((double)a1->data.i * (double)a1->data.i < QHCABS2(a2))
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_RI:
	if(a1->data.r < (double)a2->data.i)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_RR:
	if(a1->data.r < a2->data.r)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_RC:
	if(a1->data.r * a1->data.r < QHCABS2(a2))
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_CI:
	if(QHCABS2(a1) < (double)a2->data.i * (double)a2->data.i)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_CR:
	if(QHCABS2(a1) < a2->data.r * a2->data.r)
	    qh_pc = p->res - 1;
	return;

      case QHDATATYPE_CC:
	if(QHCABS2(a1) < QHCABS2(a2))
	    qh_pc = p->res - 1;
	return;
    }
    invalid_operation(p->op);
}

void qh_ifgegoto(QHProg* p)
{
    long addr1, addr2;
    QHObj* a1;
    QHObj* a2;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    a1 = memstack + addr1;
    a2 = memstack + addr2;

    switch(SWITCH_DATA_TYPE(a1, a2))
    {
      case QHDATATYPE_II:
	if(a1->data.i >= a2->data.i)
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_IR:
	if((double)a1->data.i >= a2->data.r)
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_IC:
	if((double)a1->data.i * (double)a1->data.i >= QHCABS2(a2))
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_RI:
	if(a1->data.r >= (double)a2->data.i)
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_RR:
	if(a1->data.r >= a2->data.r)
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_RC:
	if(a1->data.r * a1->data.r >= QHCABS2(a2))
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_CI:
	if(QHCABS2(a1) >= (double)a2->data.i * (double)a2->data.i)
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_CR:
	if(QHCABS2(a1) >= a2->data.r * a2->data.r)
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_CC:
	if(QHCABS2(a1) >= QHCABS2(a2))
	    qh_pc = p->res - 1;
	break;

      default:
	invalid_operation(p->op);
    }
}

void qh_ifgtgoto(QHProg* p)
{
    long addr1, addr2;
    QHObj* a1;
    QHObj* a2;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    a1 = memstack + addr1;
    a2 = memstack + addr2;

    switch(SWITCH_DATA_TYPE(a1, a2))
    {
      case QHDATATYPE_II:
	if(a1->data.i > a2->data.i)
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_IR:
	if((double)a1->data.i > a2->data.r)
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_IC:
	if((double)a1->data.i * (double)a1->data.i > QHCABS2(a2))
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_RI:
	if(a1->data.r > (double)a2->data.i)
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_RR:
	if(a1->data.r > a2->data.r)
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_RC:
	if(a1->data.r * a1->data.r > QHCABS2(a2))
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_CI:
	if(QHCABS2(a1) > (double)a2->data.i * (double)a2->data.i)
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_CR:
	if(QHCABS2(a1) > a2->data.r * a2->data.r)
	    qh_pc = p->res - 1;
	break;

      case QHDATATYPE_CC:
	if(QHCABS2(a1) > QHCABS2(a2))
	    qh_pc = p->res - 1;
	break;

      default:
	invalid_operation(p->op);
    }
}

void qh_goto(QHProg* p)
{
    qh_pc = p->res - 1;
}

void qh_return(QHProg* p)
{
    int ret;
    QHObj x;

    if(isemptyQHFuncStack(&fstk))
	qhexit(cast_to_int(&memstack[MEMADDR(p->a1)]));

    if(p->a1 != 0)
	x = memstack[MEMADDR(p->a1)];
    else
    {
	x.type = QH_Int;
	x.data.i = 0;
    }
    ret = fstk.elem[fstk.ptr - 1].ret;
    qh_restore();
    memstack[ret] = x;
}

void qh_end(QHProg* p)
{
    qhexit(0);
}

void qh_zero(QHProg* p)
{
    memstack[MEMADDR(p->res)].type = QH_Int;
    memstack[MEMADDR(p->res)].data.i = 0;
}

void qh_one(QHProg* p)
{
    memstack[MEMADDR(p->res)].type = QH_Int;
    memstack[MEMADDR(p->res)].data.i = 1;
}

void qh_complex_op(QHProg* p)
{
    QHObj* x1  = memstack + MEMADDR((p)->a1);
    QHObj* x2  = memstack + MEMADDR((p)->a2);
    QHObj* res = memstack + MEMADDR((p)->res);
    double re, im;

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    re = im = 0.0;

    if(x1->type == QH_Int)
    {
	re = (double)x1->data.i;
    }
    else if(x1->type == QH_Real)
    {
	re = x1->data.r;
    }
    else
    {
	re = x1->data.c.re;
	im = x1->data.c.im;
    }

    if(x2->type == QH_Int)
    {
	im += (double)x2->data.i;
    }
    else if(x2->type == QH_Real)
    {
	im += x2->data.r;
    }
    else
    {
	re -= x2->data.c.im;
	im += x2->data.c.re;
    }
    res->type = QH_Complex;
    res->data.c.re = re;
    res->data.c.im = im;
}

void qh_int_op(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);
    int type = x->type;

    res->type = QH_Int;
    switch(type)
    {
      case QH_Int:
	res->data.i = x->data.i;
	return;

      case QH_Real:
	res->data.i = (int)x->data.r;
	return;

      case QH_Complex:
	res->data.i = (int)x->data.c.re;
	return;

      case QH_String:
	if(x->data.s->str == NULL)
	    res->data.i = 0;
	else
	    res->data.i = atoi(x->data.s->str);
	return;

      case QH_Keyword:
	if(x->data.i != QH_Key_Null)
	    break;
	res->data.i = 0;
	return;
    }
    invalid_operation(p->op);
}

void qh_real_op(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);
    int type = x->type;

    res->type = QH_Real;
    switch(type)
    {
      case QH_Int:
	res->data.r = (double)x->data.i;
	return;

      case QH_Real:
	res->data.r = x->data.r;
	return;

      case QH_Complex:
	res->data.r = x->data.c.re;
	return;

      case QH_String:
	if(x->data.s->str == NULL)
	    res->data.r = 0.0;
	else
	    res->data.i = atof(x->data.s->str);
	return;

      case QH_Keyword:
	if(x->data.i != QH_Key_Null)
	    break;
	res->data.r = 0.0;
	return;
    }

    invalid_operation(p->op);
}

void qh_image_op(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);
    int type = x->type;

    res->type = QH_Real;
    switch(type)
    {
      case QH_Int:
	res->data.r = 0.0;
	return;

      case QH_Real:
	res->data.r = 0.0;
	return;

      case QH_Complex:
	res->data.r = x->data.c.im;
	return;

      case QH_Keyword:
	if(x->data.i != QH_Key_Null)
	    break;
	res->data.r = 0.0;
	return;
    }

    invalid_operation(p->op);
}

void qh_commandop(QHProg* p)
{
    QHObj* x   = memstack + MEMADDR(p->a1);
    QHObj* res = memstack + MEMADDR(p->res);
    char* com;
    FILE* fp;
    QHString* s;

    if(!IS_QHSTRING(x))
	execute_error("Expected string");

    com = x->data.s->str;

    if(com == NULL)
	com = "";

    if((fp = popen(com, "r")) == NULL)
    {
	perror(com);
	execute_error("Command failure");
    }

    while(fgets(temp_buff, BUFSIZ, fp))
	add_string_to_flexible_buff(temp_buff);

    pclose(fp);

    s         = alloc_string();
    s->maxlen = flexible_buff_size;
    s->str    = get_flexible_buff();
    s->header.type = QH_String;
    res->data.s = s;
    res->type = QH_String;
}
