/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "common.h"
#include "functab.h"
#include "interp.h"

/* unary operator */

/********************************* UMINUS ***********************************/
void qh_uminus(QHProg* p)
{
    long addr1, addr3;
    QHObj* x;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Int;
	res->data.i = -x->data.i;
	return;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = -x->data.r;
	return;

      case QH_Complex:
	res->type = QH_Complex;
	res->data.c.re = -x->data.c.re;
	res->data.c.im = -x->data.c.im;
	return;
    }
    invalid_operation(p->op);
}

/********************************* NOT ***********************************/
void qh_not(QHProg* p)
{
    long addr1, addr3;
    QHObj* x;
    QHObj* res;
    int type;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;
    type  = x->type;

    res->type = QH_Int;
    if(type <= QH_Complex)
    {
	switch(type)
	{
	  case QH_Int:
	    res->data.i = !x->data.i;
	    break;

	  case QH_Real:
	    res->data.i = !x->data.r;
	    break;

	  case QH_Complex:
	    res->data.i = !(x->data.c.re || x->data.c.im);
	    break;
	}
    }
    else if(type == QH_Keyword)
    {
	res->data.i = (x->data.i != QH_Key_Null);
    }
    else
    {
	res->data.i = 0;
    }
}


/********************************* NOTNOT ********************************/
void qh_notnot(QHProg* p)
{
    long addr1, addr3;
    QHObj* x;
    QHObj* res;
    int type;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;
    type  = x->type;

    res->type = QH_Int;
    if(type <= QH_Complex)
    {
	switch(type)
	{
	  case QH_Int:
	    res->data.i = !!x->data.i;
	    break;

	  case QH_Real:
	    res->data.i = !!x->data.r;
	    break;

	  case QH_Complex:
	    res->data.i = !!(x->data.c.re || x->data.c.im);
	    break;
	}
    }
    else if(type == QH_Keyword)
	res->data.i = !(x->data.i != QH_Key_Null);
    else
	res->data.i = 1;
}


/********************************* XINC ***********************************/
void qh_xinc(QHProg* p)
{
    long addr1, addr3;
    QHObj* x;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Int;
	res->data.i = x->data.i++;
	return;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = x->data.r++;
	return;

      case QH_Complex:
	res->type = QH_Complex;
	res->data.c.re = x->data.c.re++;
	res->data.c.im = x->data.c.im;
	return;
    }

    invalid_operation(p->op);
}


/********************************* INCX ***********************************/
void qh_incx(QHProg* p)
{
    long addr1, addr3;
    QHObj* x;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Int;
	res->data.i = ++x->data.i;
	return;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = ++x->data.r;
	return;

      case QH_Complex:
	res->type = QH_Complex;
	res->data.c.re = ++x->data.c.re;
	res->data.c.im = x->data.c.im;
	return;
    }
    invalid_operation(p->op);
}

/********************************* XDEC ***********************************/
void qh_xdec(QHProg* p)
{
    long addr1, addr3;
    QHObj* x;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Int;
	res->data.i = x->data.i--;
	return;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = x->data.r--;
	return;

      case QH_Complex:
	res->type = QH_Complex;
	res->data.c.re = x->data.c.re--;
	res->data.c.im = x->data.c.im;
	return;
    }
    invalid_operation(p->op);
}


/********************************* DECX ***********************************/
void qh_decx(QHProg* p)
{
    long addr1, addr3;
    QHObj* x;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;

    switch(x->type)
    {
      case QH_Int:
	res->type = QH_Int;
	res->data.i = --x->data.i;
	return;

      case QH_Real:
	res->type = QH_Real;
	res->data.r = --x->data.r;
	return;

      case QH_Complex:
	res->type = QH_Complex;
	res->data.c.re = --x->data.c.re;
	res->data.c.im = x->data.c.im;
	return;
    }
    invalid_operation(p->op);
}


/********************************* SIZEOF_OP *******************************/
void qh_sizeof_op(QHProg* p)
{
    long addr1, addr3;
    QHObj* x;
    QHObj* res;
    int type;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;
    type  = x->type;

    res->type = QH_Int;
    switch(type)
    {
      case QH_Int:
	res->data.i = sizeof(int);
	break;
      case QH_Real:
	res->data.i = sizeof(double);
	break;
      case QH_Complex:
	res->data.i = sizeof(XComplex);
	break;
      case QH_String:
	res->data.i = x->data.s->maxlen;
	break;
      default:
	res->data.i = 4;
	break;
    }
}

/******************************* STRUCTOF ***********************************/
void qh_structof(QHProg* p)
{
    long addr1, addr3;
    QHObj* x;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;

    if(x->type != QH_Struct)
	res->data.i = -1;
    else
	res->data.i = x->data.st->entry;
    res->type = QH_Int;
}

/********************************* INDIR ***********************************/
void qh_indir(QHProg* p)
{
    QHObj* res = memstack + MEMADDR(p->res);
    int addr = cast_to_int(memstack + MEMADDR(p->a1));

    if(addr <= min_data_addr ||
       addr >= qh_sp)
    {
	sprintf(temp_buff, "address %d couldn't accsess", addr);
	execute_error(temp_buff);
    }

    *res = memstack[addr];
}

/********************************* ADDR ***********************************/
void qh_addr(QHProg* p)
{
    QHObj* res = memstack + MEMADDR(p->res);

    res->type = QH_Int;
    res->data.i = MEMADDR(p->a1);
}

/********************************* LOGNOT ***********************************/
void qh_lognot(QHProg* p)
{
    long addr1, addr3;
    QHObj* x;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;

    if(x->type >= QH_String)
	invalid_operation(p->op);

    res->data.i = ~cast_to_int(x);
    res->type   = QH_Int;
}
