/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "common.h"
#include "functab.h"
#include "interp.h"

/* binary operator */
/********************************* ADD ***********************************/
void qh_add(QHProg* p)
{
    XComplex c;
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    switch(SWITCH_DATA_TYPE(x1, x2))
    {
      case QHDATATYPE_II:
	res->type = QH_Int;
	res->data.i = x1->data.i + x2->data.i;
	break;

      case QHDATATYPE_IR:
	res->type = QH_Real;
	res->data.r = (double)x1->data.i + x2->data.r;
	break;

      case QHDATATYPE_IC:
	res->type = QH_Complex;
	c.re = (double)x1->data.i + x2->data.c.re;
	c.im = x2->data.c.im;
	res->data.c = c;
	break;

      case QHDATATYPE_RI:
	res->type = QH_Real;
	res->data.r = x1->data.r + (double)x2->data.i;
	break;

      case QHDATATYPE_RR:
	res->type = QH_Real;
	res->data.r = x1->data.r + x2->data.r;
	break;

      case QHDATATYPE_RC:
	res->type = QH_Complex;
	c.re = x1->data.r + x2->data.c.re;
	c.im = x2->data.c.im;
	res->data.c = c;
	break;

      case QHDATATYPE_CI:
	res->type = QH_Complex;
	c.re = x1->data.c.re + (double)x2->data.i;
	c.im = x1->data.c.im;
	res->data.c = c;
	break;

      case QHDATATYPE_CR:
	res->type = QH_Complex;
	c.re = x1->data.c.re + x2->data.r;
	c.im = x1->data.c.im;
	res->data.c = c;
	break;

      case QHDATATYPE_CC:
	res->type = QH_Complex;
	c.re = x1->data.c.re + x2->data.c.re;
	c.im = x1->data.c.im + x2->data.c.im;
	res->data.c = c;
	break;

      default:
	invalid_operation(p->op);
    }
}

/********************************* SUB ***********************************/
void qh_sub(QHProg* p)
{
    XComplex c;
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    switch(SWITCH_DATA_TYPE(x1, x2))
    {
      case QHDATATYPE_II:
	res->type = QH_Int;
	res->data.i = x1->data.i - x2->data.i;
	break;

      case QHDATATYPE_IR:
	res->type = QH_Real;
	res->data.r = (double)x1->data.i - x2->data.r;
	break;

      case QHDATATYPE_IC:
	res->type = QH_Complex;
	c.re = (double)x1->data.i - x2->data.c.re;
	c.im = -x2->data.c.im;
	res->data.c = c;
	break;

      case QHDATATYPE_RI:
	res->type = QH_Real;
	res->data.r = x1->data.r - (double)x2->data.i;
	break;

      case QHDATATYPE_RR:
	res->type = QH_Real;
	res->data.r = x1->data.r - x2->data.r;
	break;

      case QHDATATYPE_RC:
	res->type = QH_Complex;
	c.re = x1->data.r - x2->data.c.re;
	c.im = -x2->data.c.im;
	res->data.c = c;
	break;

      case QHDATATYPE_CI:
	res->type = QH_Complex;
	c.re = x1->data.c.re - (double)x2->data.i;
	c.im = x1->data.c.im;
	res->data.c = c;
	break;

      case QHDATATYPE_CR:
	res->type = QH_Complex;
	c.re = x1->data.c.re - x2->data.r;
	c.im = x1->data.c.im;
	res->data.c = c;
	break;

      case QHDATATYPE_CC:
	res->type = QH_Complex;
	c.re = x1->data.c.re - x2->data.c.re;
	c.im = x1->data.c.im - x2->data.c.im;
	res->data.c = c;
	break;
    }
}

/********************************* MUL ***********************************/
void qh_mul(QHProg* p)
{
    XComplex c;
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    switch(SWITCH_DATA_TYPE(x1, x2))
    {
      case QHDATATYPE_II:
	res->type = QH_Int;
	res->data.i = x1->data.i * x2->data.i;
	break;

      case QHDATATYPE_IR:
	res->type = QH_Real;
	res->data.r = (double)x1->data.i * x2->data.r;
	break;

      case QHDATATYPE_IC:
	res->type = QH_Complex;
	c.re = (double)x1->data.i * x2->data.c.re;
	c.im = (double)x1->data.i * x2->data.c.im;
	res->data.c = c;
	break;

      case QHDATATYPE_RI:
	res->type = QH_Real;
	res->data.r = x1->data.r * (double)x2->data.i;
	break;

      case QHDATATYPE_RR:
	res->type = QH_Real;
	res->data.r = x1->data.r * x2->data.r;
	break;

      case QHDATATYPE_RC:
	res->type = QH_Complex;
	c.re = x1->data.r * x2->data.c.re;
	c.im = x1->data.r * x2->data.c.im;
	res->data.c = c;
	break;

      case QHDATATYPE_CI:
	res->type = QH_Complex;
	c.re = x1->data.c.re * (double)x2->data.i;
	c.im = x1->data.c.im * (double)x2->data.i;
	res->data.c = c;
	break;

      case QHDATATYPE_CR:
	res->type = QH_Complex;
	c.re = x1->data.c.re * x2->data.r;
	c.im = x1->data.c.im * x2->data.r;
	res->data.c = c;
	break;

      case QHDATATYPE_CC:
	res->type = QH_Complex;
	c.re = x1->data.c.re * x2->data.c.re - x1->data.c.im * x2->data.c.im;
	c.im = x1->data.c.re * x2->data.c.im + x2->data.c.re * x1->data.c.im;
	res->data.c = c;
	break;
    }
}


/********************************* DIV ***********************************/
void qh_div(QHProg* p)
{
    XComplex c;
    double r;
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    switch(SWITCH_DATA_TYPE(x1, x2))
    {
      case QHDATATYPE_II:
	res->type = QH_Int;
	res->data.i = x1->data.i / x2->data.i;
	break;

      case QHDATATYPE_IR:
	res->type = QH_Real;
	res->data.r = (double)x1->data.i / x2->data.r;
	break;

      case QHDATATYPE_IC:
	r = x2->data.c.re * x2->data.c.re + x2->data.c.im * x2->data.c.im;
	res->type = QH_Complex;
	c.re = (double)x1->data.i * x2->data.c.re / r;
	c.im = -(double)x1->data.i * x2->data.c.im / r;
	res->data.c = c;
	break;

      case QHDATATYPE_RI:
	res->type = QH_Real;
	res->data.r = x1->data.r / (double)x2->data.i;
	break;

      case QHDATATYPE_RR:
	res->type = QH_Real;
	res->data.r = x1->data.r / x2->data.r;
	break;

      case QHDATATYPE_RC:
	r = x2->data.c.re * x2->data.c.re + x2->data.c.im * x2->data.c.im;
	res->type = QH_Complex;
	c.re = x1->data.r * x2->data.c.re / r;
	c.im = -x1->data.r * x2->data.c.im / r;
	res->data.c = c;
	break;

      case QHDATATYPE_CI:
	res->type = QH_Complex;
	c.re = x1->data.c.re / (double)x2->data.i;
	c.im = x1->data.c.im / (double)x2->data.i;
	res->data.c = c;
	break;

      case QHDATATYPE_CR:
	res->type = QH_Complex;
	c.re = x1->data.c.re / x2->data.r;
	c.im = x1->data.c.im / x2->data.r;
	res->data.c = c;
	break;

      case QHDATATYPE_CC:
	r = x2->data.c.re * x2->data.c.re + x2->data.c.im * x2->data.c.im;
	res->type = QH_Complex;
	c.re = (x1->data.c.re * x2->data.c.re
		+ x1->data.c.im * x2->data.c.im) / r;
	c.im = (x2->data.c.re * x1->data.c.im
		- x1->data.c.re * x2->data.c.im) / r;
	res->data.c = c;
	break;
    }
}


/********************************* REM ***********************************/
void qh_rem(QHProg* p)
{
    XComplex c;
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    switch(SWITCH_DATA_TYPE(x1, x2))
    {
      case QHDATATYPE_II:
	res->type = QH_Int;
	res->data.i = x1->data.i % x2->data.i;
	break;

      case QHDATATYPE_IR:
	res->type = QH_Real;
	res->data.r = fmod((double)x1->data.i, x2->data.r);
	break;

      case QHDATATYPE_IC:
	res->type = QH_Complex;
	res->data.c.re = fmod((double)x1->data.i, x2->data.c.re);
	res->data.c.im = 0.0;
	break;

      case QHDATATYPE_RI:
	res->type = QH_Real;
	res->data.r = fmod(x1->data.r, (double)x2->data.i);
	break;

      case QHDATATYPE_RR:
	res->type = QH_Real;
	res->data.r = fmod(x1->data.r, x2->data.r);
	break;

      case QHDATATYPE_RC:
	res->type = QH_Complex;
	res->data.c.re = fmod(x1->data.r, x2->data.c.re);
	res->data.c.im = 0.0;
	break;

      case QHDATATYPE_CI:
	res->type = QH_Complex;
	res->data.c.re = fmod(x1->data.c.re, (double)x2->data.i);
	res->data.c.im = 0.0;
	break;

      case QHDATATYPE_CR:
	res->type = QH_Complex;
	res->data.c.re = fmod(x1->data.c.re, x2->data.r);
	res->data.c.im = 0.0;
	break;

      case QHDATATYPE_CC:
	res->type = QH_Complex;
	c.re = fmod(x1->data.c.re, x2->data.c.re);
	c.im = fmod(x1->data.c.im, x2->data.c.im);
	res->data.c = c;
	break;
    }
}


/********************************* POW ***********************************/
void qh_pow(QHProg* p)
{
    XComplex c;
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    switch(SWITCH_DATA_TYPE(x1, x2))
    {
      case QHDATATYPE_II:
	{
	    int i, r;
	    res->type = QH_Int;
	    if(x2->data.i == 0)
		r = 1;
	    else if(x1->data.i == 0 || x2->data.i < 0)
		r = 0;
	    else
	    {
		r = x1->data.i;
		for(i = 1; i < x2->data.i; i++)
		    r *= x1->data.i;
	    }
	    res->data.i = r;
	}
	break;

      case QHDATATYPE_IR:
	res->type = QH_Real;
	res->data.r = pow((double)x1->data.i, x2->data.r);
	break;

      case QHDATATYPE_IC:
	{
	    double r, th;

	    r  = pow((double)x1->data.i, x2->data.c.re);
	    th = x2->data.c.im;
	    res->type = QH_Complex;
	    c.re = r * cos(th);
	    c.im = r * sin(th);
	    res->data.c = c;
	}
	break;

      case QHDATATYPE_RI:
	res->type = QH_Real;
	if(BETWEEN(-50, x2->data.i, 50))
	{
	    double r;
	    int i;

	    if(x2->data.i == 0)
		r = 1.0;
	    else if(x1->data.r == 0.0)
		r = 0.0;
	    else
	    {
		if(x2->data.i > 0)
		{
		    r = x1->data.r;
		    for(i = 1; i < x2->data.i; i++)
			r *= x1->data.r;
		}
		else
		{
		    r = 1.0 / x1->data.r;
		    for(i = -1; i > x2->data.i; i--)
			r /= x1->data.r;
		}

	    }
	    res->data.r = r;
	}
	else
	    res->data.r = pow(x1->data.r, (double)x2->data.i);
	break;

      case QHDATATYPE_RR:
	res->type = QH_Real;
	res->data.r = pow(x1->data.r, x2->data.r);
	break;

      case QHDATATYPE_RC:
	{
	    double r, th;

	    res->type = QH_Complex;
	    r  = pow(x1->data.r, x2->data.c.re);
	    th = x2->data.c.im;
	    c.re = r * cos(th);
	    c.im = r * sin(th);
	    res->data.c = c;
	}
	break;

      case QHDATATYPE_CI:
	{
	    double a, th, rr;
	    res->type = QH_Complex;
	    if(x2->data.i == 0)
	    {
		res->data.c.re = 1.0;
		res->data.c.im = 0.0;
		break;
	    }
	    if(x1->data.c.re == 0.0 &&
	       x1->data.c.im == 0.0)
	    {
		res->data.c.re = 0.0;
		res->data.c.im = 0.0;
		break;
	    }
	    a = (double)x2->data.i;
	    th = atan2(x1->data.c.im, x1->data.c.re) * a;
	    rr = pow(QHCABS2(x1), a * 0.5);
	    c.re = rr * cos(th);
	    c.im = rr * sin(th);
	    res->data.c = c;
	}
	break;

      case QHDATATYPE_CR:
	{
	    double a, th, rr;
	    res->type = QH_Complex;
	    if(x2->data.r == 0.0)
	    {
		res->data.c.re = 1.0;
		res->data.c.im = 0.0;
		break;
	    }
	    if(x1->data.c.re == 0.0 &&
	       x1->data.c.im == 0.0)
	    {
		res->data.c.re = 0.0;
		res->data.c.im = 0.0;
		break;
	    }
	    a = x2->data.r;
	    th = atan2(x1->data.c.im, x1->data.c.re) * a;
	    rr = pow(QHCABS2(x1), a * 0.5);
	    c.re = rr * cos(th);
	    c.im = rr * sin(th);
	    res->data.c = c;
	}
	break;

      case QHDATATYPE_CC:
	{
	    double xx, yy;
	    double xx1, yy1;
	    double xx2, yy2;
	    double t, r;
	    xx1 = x1->data.c.re;
	    yy1 = x1->data.c.im;
	    xx2 = x2->data.c.re;
	    yy2 = x2->data.c.im;
	    res->type = QH_Complex;
	    if(xx2 == 0.0 && yy2 == 0.0)
	    {
		res->data.c.re = 1.0;
		res->data.c.im = 0.0;
		break;
	    }
	    if(xx1 == 0.0 && yy1 == 0.0)
	    {
		res->data.c.re = 0.0;
		res->data.c.im = 0.0;
		break;
	    }
	    r = xx1 * xx1 + yy1 * yy1;
	    t = atan2(yy1, xx1);
	    xx = xx2 * log(r) * 0.5 - yy2 * t;
	    yy = yy2 * log(r) * 0.5 + xx2 * t;
	    res->data.c.re = exp(xx) * cos(yy);
	    res->data.c.im = exp(xx) * sin(yy);
	}
	break;
    }
}

/********************************* EQ ***********************************/
void qh_eq(QHProg* p)
{
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    res->type = QH_Int;

    if((x1->type | x2->type) >= QH_String)
    {
	if(x1->type == x2->type)
	    res->data.i = x1->data.v == x2->data.v;
	else
	    res->data.i = 0;
	return;
    }

    switch(SWITCH_DATA_TYPE(x1, x2))
    {
      case QHDATATYPE_II:
	res->data.i = x1->data.i == x2->data.i;
	break;

      case QHDATATYPE_IR:
	res->data.i = (double)x1->data.i == x2->data.r;
	break;

      case QHDATATYPE_IC:
	res->data.i = ((double)x1->data.i == x2->data.c.re) &&
	    (x2->data.c.im == 0.0);
	break;

      case QHDATATYPE_RI:
	res->data.i = (x1->data.r == (double)x2->data.i);
	break;

      case QHDATATYPE_RR:
	res->data.i = (x1->data.r == x2->data.r);
	break;

      case QHDATATYPE_RC:
	res->data.i = (x1->data.r == x2->data.c.re) &&
	    (x2->data.c.im == 0.0);
	break;

      case QHDATATYPE_CI:
	res->data.i = (x1->data.c.re == (double)x2->data.i) &&
	    (x1->data.c.im == 0.0);
	break;

      case QHDATATYPE_CR:
	res->data.i = (x1->data.c.re == x2->data.r) &&
	    (x1->data.c.im == 0.0);
	break;

      case QHDATATYPE_CC:
	res->data.i = (x1->data.c.re == x2->data.c.re) &&
	    (x1->data.c.im == x2->data.c.im);
	break;
    }
}

/********************************* NQ ***********************************/
void qh_nq(QHProg* p)
{
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    res->type = QH_Int;

    if((x1->type | x2->type) >= QH_String)
    {
	if(x1->type == x2->type)
	    res->data.i = x1->data.v != x2->data.v;
	else
	    res->data.i = 1;
	return;
    }

    switch(SWITCH_DATA_TYPE(x1, x2))
    {
      case QHDATATYPE_II:
	res->data.i = x1->data.i != x2->data.i;
	break;

      case QHDATATYPE_IR:
	res->data.i = (double)x1->data.i != x2->data.r;
	break;

      case QHDATATYPE_IC:
	res->data.i = ((double)x1->data.i != x2->data.c.re) ||
	    (x2->data.c.im != 0.0);
	break;

      case QHDATATYPE_RI:
	res->data.i = (x1->data.r != (double)x2->data.i);
	break;

      case QHDATATYPE_RR:
	res->data.i = (x1->data.r != x2->data.r);
	break;

      case QHDATATYPE_RC:
	res->data.i = (x1->data.r != x2->data.c.re) ||
	    (x2->data.c.im != 0.0);
	break;

      case QHDATATYPE_CI:
	res->data.i = (x1->data.c.re != (double)x2->data.i) ||
	    (x1->data.c.im != 0.0);
	break;

      case QHDATATYPE_CR:
	res->data.i = (x1->data.c.re != x2->data.r) ||
	    (x1->data.c.im != 0.0);
	break;

      case QHDATATYPE_CC:
	res->data.i = (x1->data.c.re != x2->data.c.re) ||
	    (x1->data.c.im != x2->data.c.im);
	break;
    }
}


/********************************* LE ***********************************/
void qh_le(QHProg* p)
{
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    res->type = QH_Int;

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    switch(SWITCH_DATA_TYPE(x1, x2))
    {
      case QHDATATYPE_II:
	res->data.i = x1->data.i <= x2->data.i;
	break;

      case QHDATATYPE_IR:
	res->data.i = (double)x1->data.i <= x2->data.r;
	break;

      case QHDATATYPE_IC:
	res->data.i = (double)x1->data.i * (double)x1->data.i <= QHCABS2(x2);
	break;

      case QHDATATYPE_RI:
	res->data.i = x1->data.r <= (double)x2->data.i;
	break;

      case QHDATATYPE_RR:
	res->data.i = x1->data.r <= x2->data.r;
	break;

      case QHDATATYPE_RC:
	res->data.i = x1->data.r * x1->data.r <= QHCABS2(x2);
	break;

      case QHDATATYPE_CI:
	res->data.i = QHCABS2(x1) <= (double)x2->data.i * (double)x2->data.i;
	break;

      case QHDATATYPE_CR:
	res->data.i = QHCABS2(x1) <= x2->data.r * x2->data.r;
	break;

      case QHDATATYPE_CC:
	res->data.i = QHCABS2(x1) <= QHCABS2(x2);
	break;
    }
}

/********************************* LT ***********************************/
void qh_lt(QHProg* p)
{
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    res->type = QH_Int;

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    switch(SWITCH_DATA_TYPE(x1, x2))
    {
      case QHDATATYPE_II:
	res->data.i = x1->data.i < x2->data.i;
	break;

      case QHDATATYPE_IR:
	res->data.i = (double)x1->data.i < x2->data.r;
	break;

      case QHDATATYPE_IC:
	res->data.i = (double)x1->data.i * (double)x1->data.i < QHCABS2(x2);
	break;

      case QHDATATYPE_RI:
	res->data.i = x1->data.r < (double)x2->data.i;
	break;

      case QHDATATYPE_RR:
	res->data.i = x1->data.r < x2->data.r;
	break;

      case QHDATATYPE_RC:
	res->data.i = x1->data.r * x1->data.r < QHCABS2(x2);
	break;

      case QHDATATYPE_CI:
	res->data.i = QHCABS2(x1) < (double)x2->data.i * (double)x2->data.i;
	break;

      case QHDATATYPE_CR:
	res->data.i = QHCABS2(x1) < x2->data.r * x2->data.r;
	break;

      case QHDATATYPE_CC:
	res->data.i = QHCABS2(x1) < QHCABS2(x2);
	break;
    }
}


/********************************* GE ***********************************/
void qh_ge(QHProg* p)
{
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    res->type = QH_Int;

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    switch(SWITCH_DATA_TYPE(x1, x2))
    {
      case QHDATATYPE_II:
	res->data.i = x1->data.i >= x2->data.i;
	break;

      case QHDATATYPE_IR:
	res->data.i = (double)x1->data.i >= x2->data.r;
	break;

      case QHDATATYPE_IC:
	res->data.i = (double)x1->data.i * (double)x1->data.i >= QHCABS2(x2);
	break;

      case QHDATATYPE_RI:
	res->data.i = x1->data.r >= (double)x2->data.i;
	break;

      case QHDATATYPE_RR:
	res->data.i = x1->data.r >= x2->data.r;
	break;

      case QHDATATYPE_RC:
	res->data.i = x1->data.r * x1->data.r >= QHCABS2(x2);
	break;

      case QHDATATYPE_CI:
	res->data.i = QHCABS2(x1) >= (double)x2->data.i * (double)x2->data.i;
	break;

      case QHDATATYPE_CR:
	res->data.i = QHCABS2(x1) >= x2->data.r * x2->data.r;
	break;

      case QHDATATYPE_CC:
	res->data.i = QHCABS2(x1) >= QHCABS2(x2);
	break;
    }
}

/********************************* GT ***********************************/
void qh_gt(QHProg* p)
{
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    res->type = QH_Int;

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    switch(SWITCH_DATA_TYPE(x1, x2))
    {
      case QHDATATYPE_II:
	res->data.i = x1->data.i > x2->data.i;
	break;

      case QHDATATYPE_IR:
	res->data.i = (double)x1->data.i > x2->data.r;
	break;

      case QHDATATYPE_IC:
	res->data.i = (double)x1->data.i * (double)x1->data.i > QHCABS2(x2);
	break;

      case QHDATATYPE_RI:
	res->data.i = x1->data.r > (double)x2->data.i;
	break;

      case QHDATATYPE_RR:
	res->data.i = x1->data.r > x2->data.r;
	break;

      case QHDATATYPE_RC:
	res->data.i = x1->data.r * x1->data.r > QHCABS2(x2);
	break;

      case QHDATATYPE_CI:
	res->data.i = QHCABS2(x1) > (double)x2->data.i * (double)x2->data.i;
	break;

      case QHDATATYPE_CR:
	res->data.i = QHCABS2(x1) > x2->data.r * x2->data.r;
	break;

      case QHDATATYPE_CC:
	res->data.i = QHCABS2(x1) > QHCABS2(x2);
	break;
    }
}


/* assign operator */
void qh_aassign(QHProg* p)
{
    long addr1, addr3;
    QHObj* x;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;

    if((x->type | res->type) >= QH_String)
	invalid_operation(p->op);

    switch(SWITCH_DATA_TYPE(res, x))
    {
      case QHDATATYPE_II:
	res->data.i += x->data.i;
	break;

      case QHDATATYPE_IR:
	res->type   = QH_Real;
	res->data.r = (double)res->data.i + x->data.r;
	break;

      case QHDATATYPE_IC:
	res->type = QH_Complex;
	res->data.c.re = (double)res->data.i + x->data.c.re;
	res->data.c.im = x->data.c.im;
	break;

      case QHDATATYPE_RI:
	res->data.r += (double)x->data.i;
	break;

      case QHDATATYPE_RR:
	res->data.r += x->data.r;
	break;

      case QHDATATYPE_RC:
	res->type = QH_Complex;
	res->data.c.re = res->data.r + x->data.c.re;
	res->data.c.im = x->data.c.im;
	break;

      case QHDATATYPE_CI:
	res->data.c.re += (double)x->data.i;
	break;

      case QHDATATYPE_CR:
	res->data.c.re += x->data.r;
	break;

      case QHDATATYPE_CC:
	res->data.c.re += x->data.c.re;
	res->data.c.im += x->data.c.im;
	break;
    }
}

/********************************* SASSIGN ***********************************/
void qh_sassign(QHProg* p)
{
    long addr1, addr3;
    QHObj* x;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;

    if((x->type | res->type) >= QH_String)
	invalid_operation(p->op);

    switch(SWITCH_DATA_TYPE(res, x))
    {
      case QHDATATYPE_II:
	res->data.i -= x->data.i;
	break;

      case QHDATATYPE_IR:
	res->type = QH_Real;
	res->data.r    = (double)res->data.i - x->data.r;
	break;

      case QHDATATYPE_IC:
	res->type = QH_Complex;
	res->data.c.re = (double)res->data.i - x->data.c.re;
	res->data.c.im = -x->data.c.im;
	break;

      case QHDATATYPE_RI:
	res->data.r -= (double)x->data.i;
	break;

      case QHDATATYPE_RR:
	res->data.r -= x->data.r;
	break;

      case QHDATATYPE_RC:
	res->type = QH_Complex;
	res->data.c.re = res->data.r - x->data.c.re;
	res->data.c.im = -x->data.c.im;
	break;

      case QHDATATYPE_CI:
	res->data.c.re -= (double)x->data.i;
	break;

      case QHDATATYPE_CR:
	res->data.c.re -= x->data.r;
	break;

      case QHDATATYPE_CC:
	res->data.c.re -= x->data.c.re;
	res->data.c.im -= x->data.c.im;
	break;
    }
}

/********************************* MASSIGN ***********************************/
void qh_massign(QHProg* p)
{
    XComplex c;
    double r;
    long addr1, addr3;
    QHObj* x;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;

    if((x->type | res->type) >= QH_String)
	invalid_operation(p->op);

    switch(SWITCH_DATA_TYPE(res, x))
    {
      case QHDATATYPE_II:
	res->data.i *= x->data.i;
	break;

      case QHDATATYPE_IR:
	res->type = QH_Real;
	res->data.r    = (double)res->data.i * x->data.r;
	break;

      case QHDATATYPE_IC:
	r = (double)res->data.i;
	res->type = QH_Complex;
	res->data.c.re = r * x->data.c.re;
	res->data.c.im = r * x->data.c.im;
	break;

      case QHDATATYPE_RI:
	res->data.r *= (double)x->data.i;
	break;

      case QHDATATYPE_RR:
	res->data.r *= x->data.r;
	break;

      case QHDATATYPE_RC:
	r = res->data.r;
	res->type = QH_Complex;
	res->data.c.re = r * x->data.c.re;
	res->data.c.im = r * x->data.c.im;
	break;

      case QHDATATYPE_CI:
	r = (double)x->data.i;
	res->type = QH_Complex;
	res->data.c.re *= r;
	res->data.c.im *= r;
	break;

      case QHDATATYPE_CR:
	r = x->data.r;
	res->type = QH_Complex;
	res->data.c.re *= r;
	res->data.c.im *= r;
	break;

      case QHDATATYPE_CC:
	c.re = res->data.c.re * x->data.c.re - res->data.c.im * x->data.c.im;
	c.im = res->data.c.re * x->data.c.im + x->data.c.re * res->data.c.im;
	res->type = QH_Complex;
	res->data.c = c;
	break;
    }
}

/********************************* DASSIGN ***********************************/
void qh_dassign(QHProg* p)
{
    XComplex c;
    double r;
    long addr1, addr3;
    QHObj* x;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;

    if((x->type | res->type) >= QH_String)
	invalid_operation(p->op);

    switch(SWITCH_DATA_TYPE(res, x))
    {
      case QHDATATYPE_II:
	res->data.i /= x->data.i;
	break;

      case QHDATATYPE_IR:
	res->type = QH_Real;
	res->data.r    = (double)res->data.i / x->data.r;
	break;

      case QHDATATYPE_IC:
	r = x->data.c.re * x->data.c.re + x->data.c.im * x->data.c.im;
	c.re = (double) res->data.i * x->data.c.re / r;
	c.im = (double)-res->data.i * x->data.c.im / r;
	res->type = QH_Complex;
	res->data.c = c;
	break;

      case QHDATATYPE_RI:
	res->data.r /= (double)x->data.i;
	break;

      case QHDATATYPE_RR:
	res->data.r /= x->data.r;
	break;

      case QHDATATYPE_RC:
	r = x->data.c.re * x->data.c.re + x->data.c.im * x->data.c.im;
	c.re =  res->data.r * x->data.c.re / r;
	c.im = -res->data.r * x->data.c.im / r;
	res->type = QH_Complex;
	res->data.c = c;
	break;

      case QHDATATYPE_CI:
	r = (double)x->data.i;
	res->data.c.re = res->data.c.re / r;
	res->data.c.im = res->data.c.im / r;
	break;

      case QHDATATYPE_CR:
	r = x->data.r;
	res->data.c.re /= r;
	res->data.c.im /= r;
	break;

      case QHDATATYPE_CC:
	r = x->data.c.re * x->data.c.re + x->data.c.im * x->data.c.im;
	c.re = (res->data.c.re * x->data.c.re
		+ res->data.c.im * x->data.c.im) / r;
	c.im = (x->data.c.re * res->data.c.im
		- res->data.c.re * x->data.c.im) / r;  
	res->data.c = c;
	break;
    }
}


/********************************* RASSIGN ***********************************/
void qh_rassign(QHProg* p)
{
    long addr1, addr3;
    QHObj* x;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr3 = MEMADDR(p->res);
    x     = memstack + addr1;
    res   = memstack + addr3;

    if((x->type | res->type) >= QH_String)
	invalid_operation(p->op);

    switch(SWITCH_DATA_TYPE(res, x))
    {
      case QHDATATYPE_II:
	res->data.i %= x->data.i;
	break;

      case QHDATATYPE_IR:
	res->type = QH_Real;
	res->data.r    = fmod((double)res->data.i, x->data.r);
	break;

      case QHDATATYPE_IC:
	res->type = QH_Complex;
	res->data.c.re = fmod((double)res->data.i, x->data.c.re);
	res->data.c.im = 0.0;
	break;

      case QHDATATYPE_RI:
	res->data.r = fmod(res->data.r, (double)x->data.i);
	break;

      case QHDATATYPE_RR:
	res->data.r = fmod(res->data.r, x->data.r);
	break;

      case QHDATATYPE_RC:
	res->type = QH_Complex;
	res->data.c.re = fmod(res->data.r, x->data.c.re);
	res->data.c.im = 0.0;
	break;

      case QHDATATYPE_CI:
	res->data.c.re = fmod(res->data.c.re, (double)x->data.i);
	res->data.c.im = fmod(res->data.c.im, (double)x->data.i);
	break;

      case QHDATATYPE_CR:
	res->data.c.re = fmod(res->data.c.re, x->data.r);
	res->data.c.im = fmod(res->data.c.im, x->data.r);
	break;

      case QHDATATYPE_CC:
	res->data.c.re = fmod(res->data.c.re, x->data.c.re);
	res->data.c.im = fmod(res->data.c.im, x->data.c.im);
	break;
    }
}


/* *addr = x
   =>
   asgnindir res addr x
 */
/********************************* AINDIR ***********************************/
void qh_aindir(QHProg* p)
{
    int addr = cast_to_int(memstack + MEMADDR(p->a1));

    if(addr <= min_data_addr || addr >= qh_sp)
    {
	sprintf(temp_buff, "address %d couldn't accsess", addr);
	execute_error(temp_buff);
    }

    memstack[MEMADDR(p->res)] = memstack[addr] = memstack[MEMADDR(p->a2)];
}

/****************************** LOGOR ***********************************/
void qh_logor(QHProg* p)
{
    QHObj* x1  = memstack + MEMADDR(p->a1);
    QHObj* x2  = memstack + MEMADDR(p->a2);
    QHObj* res = memstack + MEMADDR(p->res);

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    res->type = QH_Int;
    res->data.i = cast_to_int(x1) | cast_to_int(x2);
}

/****************************** LOGAND ***********************************/
void qh_logand(QHProg* p)
{
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    res->type = QH_Int;
    res->data.i = cast_to_int(x1) & cast_to_int(x2);
}

/****************************** LLSHIFT ***********************************/
void qh_llshift(QHProg* p)
{
    unsigned int val;
    int n;
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    val = cast_to_int(x1);
    n   = cast_to_int(x2);
    res->type = QH_Int;
    res->data.i = val << n;
}

/****************************** LRSHIFT ***********************************/
void qh_lrshift(QHProg* p)
{
    unsigned int val;
    int n;
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    val = cast_to_int(x1);
    n   = cast_to_int(x2);
    res->type = QH_Int;
    res->data.i = val >> n;
}

/****************************** ALSHIFT ***********************************/
void qh_alshift(QHProg* p)
{
    int val, n;
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    val = cast_to_int(x1);
    n   = cast_to_int(x2);
    res->type = QH_Int;
    res->data.i = val << n;
}

/****************************** ARSHIFT ***********************************/
void qh_arshift(QHProg* p)
{
    int val, n;
    long addr1, addr2, addr3;
    QHObj* x1;
    QHObj* x2;
    QHObj* res;

    addr1 = MEMADDR(p->a1);
    addr2 = MEMADDR(p->a2);
    addr3 = MEMADDR(p->res);
    x1    = memstack + addr1;
    x2    = memstack + addr2;
    res   = memstack + addr3;

    if((x1->type | x2->type) >= QH_String)
	invalid_operation(p->op);

    val = cast_to_int(x1);
    n   = cast_to_int(x2);
    res->type = QH_Int;
    res->data.i = val >> n;
}
