/* Copyright (c) 1995, Masanao Izumo */

/* qhc.c: QuickHandy to C
 * This is realy compiler of QuickHandy to executable code.
 * Don't use this function in the current version.
 * In the next version, QuickHandy compiler will be suppoted.
 */

#include <stdio.h>
#include <stdlib.h>
#include "common.h"
#include "functab.h"
#include "interp.h"

static char* volatile_control_code_flag;
static char* need_label_flag;
static void write_memstack(FILE* fp, int i)
{
    switch(memstack[i].type)
    {
      case QH_Int:
	fprintf(fp, "    memstack[%d].type   = QH_Int;\n", i);
	fprintf(fp, "    memstack[%d].data.i = %d;\n", i, memstack[i].data.i);
	break;
      case QH_Real:
	fprintf(fp, "    memstack[%d].type   = QH_Real;\n", i);
	fprintf(fp, "    memstack[%d].data.r = %.19e;\n", i, memstack[i].data.r);
	break;
      case QH_String:
	fprintf(fp, "    memstack[%d].type   = QH_String;\n", i);
	fprintf(fp, "    memstack[%d].data.s = alloc_constant_string(\"%s\");\n", i, string_unexpand(temp_buff, (char *)memstack[i].data.s->str));

	break;
      case QH_Keyword:
	fprintf(fp, "    memstack[%d].type   = QH_Keyword;\n", i);
	fprintf(fp, "    memstack[%d].data.i = 0;\n", i);
	break;
      default:
	fprintf(fp, "    memstack[%d].type   = %d;\n", i, memstack[i].type);
	fprintf(fp, "    memstack[%d].data.i = 0;\n", i);
    }
}

static void write_memfunc(FILE* fp, int i)
{
    fprintf(fp, "    memfunctab[%d].addr  = %d;\n",     i, memfunctab[i].addr);
    fprintf(fp, "    memfunctab[%d].name  = \"%s\";\n", i, memfunctab[i].name);
    if(memfunctab[i].module_name == NULL)
	fprintf(fp, "    memfunctab[%d].module_name  = NULL;\n", i);
    else
	fprintf(fp, "    memfunctab[%d].module_name  = \"%s\";\n", i, memfunctab[i].module_name);
    fprintf(fp, "    memfunctab[%d].narg  = %d;\n",     i, memfunctab[i].narg);
    fprintf(fp, "    memfunctab[%d].offsp = %d;\n",     i, memfunctab[i].offsp);
    if(memfunctab[i].extern_function == NULL)
	fprintf(fp, "    memfunctab[%d].extern_function = NULL;\n", i);
    else
	fprintf(fp, "    memfunctab[%d].extern_function = %s;\n", i, memfunctab[i].name);
}

static void write_extern_declare(FILE* fp)
{
    int i;

    for(i = 0; i < nmemfunc; i++)
	if(memfunctab[i].extern_function)
	    fprintf(fp, "extern void %s(int argc, QHObj* argv, QHObj* ret);\n",
		    memfunctab[i].name);
    fputs("\n", fp);
}

static void write_global_definition(FILE* fp, QHProg* p, int n)
{
    int i;

    fputs("#include \"qhclib.h\"\n", fp);
    fputs("\n", fp);

    write_extern_declare(fp);

    fputs("static QHProg prog[] = {\n", fp);

    for(i = 0; i < n; i++)
	fprintf(fp, "/* [%d]:%-15s */ {%ld, %ld, %ld, %ld},\n", i, opstr(p[i].op), p[i].op, p[i].res, p[i].a1, p[i].a2);

    fputs("};\n\n", fp);

    fputs("static void memset(void)\n{\n", fp);
    for(i = 0; i < qh_sp; i++)
	write_memstack(fp, i);
    fputs("\n", fp);
    if(nmemfunc > 0)
    {
	fprintf(fp, "    nmemfunc   = %d;\n", nmemfunc);
	fprintf(fp, "    memfunctab = (MemFunction *)xmalloc(%d);\n", sizeof(MemFunction) * nmemfunc);
	for(i = 0; i < nmemfunc; i++)
	    write_memfunc(fp, i);
    }
    else
    {
	fprintf(fp, "    nmemfunc   = 0;\n", nmemfunc);
	fprintf(fp, "    memfunctab = NULL;\n", sizeof(MemFunction) * nmemfunc);
    }
    {
	int n, j;

	n = current_struct_entry();

	if(n > 0)
	{
	    fputs("\n", fp);
	    fputs("    /* initialize struct member id */\n", fp);
	    fprintf(fp, "    struct_tag_table = (QHStructTagTable *)xmalloc(%d);\n",
		    n * sizeof(QHStructTagTable));
	    for(i = 0; i < n; i++)
	    {
		fprintf(fp, "    struct_tag_table[%d].nmember = %d;\n",
			i, struct_tag_table[i].nmember);
		fprintf(fp, "    struct_tag_table[%d].member_id = (long *)xmalloc(%d);\n", i, struct_tag_table[i].nmember * sizeof(int));
		fprintf(fp, "    struct_tag_table[%d].next = (QHStructTagTable **)xmalloc(%d);\n", i, struct_tag_table[i].nmember * sizeof(QHStructTagTable *));

		for(j = 0; j < struct_tag_table[i].nmember; j++)
		{
		    fprintf(fp, "    struct_tag_table[%d].member_id[%d] = %d;\n",
			    i, j, struct_tag_table[i].member_id[j]);
		    if(struct_tag_table[i].next[j] == NULL)
			fprintf(fp, "    struct_tag_table[%d].next[%d] = NULL;\n",
				i, j);
		    else
			fprintf(fp, "    struct_tag_table[%d].next[%d] = struct_tag_table + %d;\n",
				i, j, struct_tag_table[i].next[j] - struct_tag_table);
		}
	    }
	}
    }

    fputs("}\n", fp);
    fputs("\n", fp);
}

static void write_main(FILE* fp, QHProg* p, int n)
{
    fputs("void main(int argc, char** argv)\n{\n", fp);
#ifdef USE_DLD_LIBRARY
    fputs("    dld_init(argv[0]);\n", fp);
#endif
    fputs("    OK_DO_garbage_collecting = False;\n", fp);
    fputs("    qhlibint();\n", fp);
    fputs("    memset();\n", fp);
    fputs("    qhprog = prog;\n", fp);
    fprintf(fp, "    qh_top = qh_sp = %d;\n\n", qh_sp);
    fputs("\n", fp);

    /* built in variables */
    fputs("    /* built in variables */\n", fp);
#define SYSTEM_VAR(var) fprintf(fp, "    " #var " = %d;\n", var);
    SYSTEM_VAR(Q_FMT_INT);
    SYSTEM_VAR(Q_FMT_REAL);
    SYSTEM_VAR(Q_FMT_COMPLEX);
    SYSTEM_VAR(Q_FMT_STRING);
    SYSTEM_VAR(Q_NULL);
    SYSTEM_VAR(Q_GEVENT_TYPE);
    SYSTEM_VAR(Q_GEVENT_VALUE);
    SYSTEM_VAR(Q_GPOINTERX);
    SYSTEM_VAR(Q_GPOINTERY);
    SYSTEM_VAR(Q_TIME);
    SYSTEM_VAR(Q_GWINDOW_WIDTH);
    SYSTEM_VAR(Q_GWINDOW_HEIGHT);
    SYSTEM_VAR(Q_GWINDOW_PWIDTH);
    SYSTEM_VAR(Q_GWINDOW_PHEIGHT);
    SYSTEM_VAR(Q_GCURRENT_SCREEN);
    SYSTEM_VAR(Q_GVISUAL_SCREEN);
#undef SYSTEM_VAR
    fputs("    Q_ARGC    = argc - 1;\n", fp);
    fputs("    prog_argc = argc - 1;\n", fp);
    fputs("    prog_argv = argv + 1;\n", fp);
    fprintf(fp, "    min_data_addr = %d;\n", min_data_addr);

    fputs("\n    /* initialize system clock */\n", fp);
    fputs("    qh_time(NULL);\n", fp);
    fputs("    qh_clock(NULL);\n", fp);

    fputs("\n", fp);
    fputs("    make_funcall_table();\n", fp);
    fputs("    qh_pc = 0;\n", fp);
    fputs("    OK_DO_garbage_collecting = True;\n\n", fp);
    fputs("    execute_program();\n", fp);
    fputs("    qhexit(0);\n", fp);
    fputs("}\n", fp);
}

static void memstack_ref(FILE* fp, int i)
{
    if(i >= 0)
	fprintf(fp, "memstack[%d]", i);
    else
	fprintf(fp, "memstack[qh_sp - %d]", -i);
}

static void write_C_code(FILE* fp, QHProg* p, int pc)
{
    switch(p->op)
    {
      case NOP:
	fputs(";\n", fp);
	break;

      case ASSIGN:
	fputs("    ", fp);
	memstack_ref(fp, p->res);
	fputs(" = ", fp);
	memstack_ref(fp, p->a1);
	fputs(";\n", fp);
	break;

      case IPUSH:
	fputs("    pushintStack(&istack, cast_to_int(&", fp);
	memstack_ref(fp, p->a1);
	fputs("));\n", fp);
	break;

      case IPUSH2:
	fputs("    pushintStack(&istack, cast_to_int(&", fp);
	memstack_ref(fp, p->a1);
	fputs("));\n", fp);

	fputs("    pushintStack(&istack, cast_to_int(&", fp);
	memstack_ref(fp, p->a2);
	fputs("));\n", fp);
	break;

      case IPUSH3:
	fputs("    pushintStack(&istack, cast_to_int(&", fp);
	memstack_ref(fp, p->a1);
	fputs("));\n", fp);

	fputs("    pushintStack(&istack, cast_to_int(&", fp);
	memstack_ref(fp, p->a2);
	fputs("));\n", fp);

	fputs("    pushintStack(&istack, cast_to_int(&", fp);
	memstack_ref(fp, p->res);
	fputs("));\n", fp);
	break;

      case PARAM:
	fputs("    EXPAND_MEMORY_IF_OVERFLOW(QHObj, memstack, qh_top, memstack_size);\n", fp);
	fputs("    memstack[qh_top++] = ", fp);
	memstack_ref(fp, p->a1);
	fputs(";\n", fp);
	break;

      case PARAM2:
	fputs("    EXPAND_MEMORY_IF_OVERFLOW(QHObj, memstack, qh_top + 1, memstack_size);\n", fp);

	fputs("    memstack[qh_top++] = ", fp);
	memstack_ref(fp, p->a1);
	fputs(";\n", fp);

	fputs("    memstack[qh_top++] = ", fp);
	memstack_ref(fp, p->a2);
	fputs(";\n", fp);
	break;

      case PARAM3:
	fputs("    EXPAND_MEMORY_IF_OVERFLOW(QHObj, memstack, qh_top + 2, memstack_size);\n", fp);

	fputs("    memstack[qh_top++] = ", fp);
	memstack_ref(fp, p->a1);
	fputs(";\n", fp);

	fputs("    memstack[qh_top++] = ", fp);
	memstack_ref(fp, p->a2);
	fputs(";\n", fp);

	fputs("    memstack[qh_top++] = ", fp);
	memstack_ref(fp, p->res);
	fputs(";\n", fp);
	break;

      case CALL:
	fprintf(fp, "    {\n");
	fprintf(fp, "\tQHFunc x;\n");
	fprintf(fp, "\n");
	fprintf(fp, "\tx.func = %ld;\n", p->a1);
	fprintf(fp, "\tx.pc = %d;\n", pc + 1);
        if(p->res >= 0)
	    fprintf(fp, "\tx.ret = %ld;\n", p->res);
	else
	    fprintf(fp, "\tx.ret = qh_sp - %ld;\n", -p->res);
	fprintf(fp, "\tx.sp = qh_sp;\n");
	fprintf(fp, "\n");
	fprintf(fp, "\tpushQHFuncStack(&fstk, x);\n");
	fprintf(fp, "\n");
	fprintf(fp, "\tqh_argstart = qh_sp;\n");
	fprintf(fp, "\tqh_sp += %d;\n", memfunctab[p->a1].offsp);
	fprintf(fp, "\tqh_top = qh_sp;\n");
	fprintf(fp, "\tEXPAND_MEMORY_IF_OVERFLOW(QHObj, memstack, qh_top, memstack_size);\n", memfunctab[p->a1].offsp);
	fprintf(fp, "\n");
	if(memfunctab[p->a1].addr > 0) /* qh function */
	{
	    if(memfunctab[p->a1].narg < 0) /* v.l.a. */
		fprintf(fp, "\tqh_vla_bind_proc(&prog[%d].a2);\n", pc);
	    fprintf(fp, "\tgoto L%d;\n", memfunctab[p->a1].addr);
	}
	else			/* built in function */
	{
	    fprintf(fp, "\tqh_%s(memstack + x.ret);\t\n", opstr(-memfunctab[p->a1].addr));
	    fprintf(fp, "\tqh_restore();\n");
	}
	fprintf(fp, "    }\n");
	break;

      case CALLEXT:
	fprintf(stderr, "## Warning: Can't compile external function in the current version.\n");

	fprintf(fp, "    {\n");
	fprintf(fp, "\tQHFunc x;\n");
	fprintf(fp, "\n");
	fprintf(fp, "\tx.func = %ld;\n", p->a1);
	fprintf(fp, "\tx.pc = %d;\n", pc + 1);
        if(p->res >= 0)
	    fprintf(fp, "\tx.ret = %ld;\n", p->res);
	else
	    fprintf(fp, "\tx.ret = qh_sp - %ld;\n", -p->res);
	fprintf(fp, "\tx.sp = qh_sp;\n");
	fprintf(fp, "\n");
	fprintf(fp, "\tpushQHFuncStack(&fstk, x);\n");
	fprintf(fp, "\n");
	fprintf(fp, "\tqh_argstart = qh_sp;\n");
	fprintf(fp, "\tqh_sp += %d;\n", p->a2);
	fprintf(fp, "\tqh_top = qh_sp;\n");
	fprintf(fp, "\t%s(%d, memstack + qh_argstart, memstack + x.ret);\t\n",
		memfunctab[p->a1].name, p->a2);
	fprintf(fp, "\tqh_restore();\n");
	fprintf(fp, "    }\n");
	break;

      case FUNCALL:
	fputs("    {\n", fp);
	fputs("\tMemFunction* f;\n", fp);
	fprintf(fp, "\tint narg = %d;\n", p->a2);
	fputs("\tint ret = ", fp);
	if(p->res >= 0)
	    fprintf(fp, "%d;\n", p->res);
	else
	    fprintf(fp, "qh_sp - %d;\n", -p->res);
	fputs("\tint faddr;\n", fp);
	fputs("\tint addr;\n", fp);
	fputs("\n", fp);
	fprintf(fp, "\tqh_pc = %d;\n", pc + 1);
	fprintf(fp, "\taddr = obtain_funcall_address(prog + %d, &faddr);\n", pc);
	fprintf(fp, "\tf = memfunctab + faddr;\n");

	fputs("\n", fp);
	fputs("\tif(addr > 0)\n", fp);
	fputs("\t{\n", fp);
	fprintf(fp, "if(f->narg < 0)\n\t\tqh_vla_bind_proc(&prog[%d].a2);\n", pc);
	fputs("\t    qh_pc = addr;\n", fp);
	fputs("\t    goto pc_jump;\n", fp);
	fputs("\t}\n", fp);
	fputs("\telse\n", fp);
	fputs("\t{\n", fp);
	fputs("\t    if(f->extern_function)\n", fp);
	fputs("\t\tqh_call_extern_function(f, narg, memstack + qh_argstart, memstack + ret);\n", fp);
	fputs("\t    else\n", fp);
	fprintf(fp, "\t\tQH_CALL(-addr, (memstack + ret));\n");
	fputs("\t    qh_restore();\n", fp);
	fputs("\t}\n", fp);
	fputs("    }\n", fp);
	break;

      case IFGOTO:
	fputs("    switch(", fp);
	memstack_ref(fp, p->a1);
	fputs(".type)\n", fp);
	fputs("    {\n", fp);

	fprintf(fp, "      case QH_Int:\n\tif(");
	memstack_ref(fp, p->a1);
	fprintf(fp, ".data.i != 0) goto L%ld;\n\tbreak;\n", p->res);

	fprintf(fp, "      case QH_Real:\n\tif(");
	memstack_ref(fp, p->a1);
	fprintf(fp, ".data.r != 0.0) goto L%ld;\n\tbreak;\n", p->res);

	fprintf(fp, "      case QH_Complex:\n\tif(");
	memstack_ref(fp, p->a1);
	fputs(".data.c.re != 0 || ", fp);
	memstack_ref(fp, p->a1);
	fprintf(fp, ".data.c.im != 0) goto L%ld;\n\tbreak;\n", p->res);

	fprintf(fp, "      case QH_Keyword:\n\tif(");
	memstack_ref(fp, p->a1);
	fprintf(fp, ".data.i != QH_Key_Null) goto L%ld;\n\tbreak;\n", p->res);

	fprintf(fp, "      default:\n\tgoto L%ld;\n\tbreak;\n", p->res);

	fputs("    }\n", fp);
	break;

      case IFNGOTO:
	fputs("    switch(", fp);
	memstack_ref(fp, p->a1);
	fputs(".type)\n", fp);
	fputs("    {\n", fp);

	fprintf(fp, "      case %d:\n\tif(", QH_Int);
	memstack_ref(fp, p->a1);
	fprintf(fp, ".data.i == 0) goto L%ld;\n\tbreak;\n", p->res);

	fprintf(fp, "      case %d:\n\tif(", QH_Real);
	memstack_ref(fp, p->a1);
	fprintf(fp, ".data.r == 0.0) goto L%ld;\n\tbreak;\n", p->res);

	fprintf(fp, "      case %d:\n\tif(", QH_Complex);
	memstack_ref(fp, p->a1);
	fputs(".data.c.re == 0 && ", fp);
	memstack_ref(fp, p->a1);
	fprintf(fp, ".data.c.im == 0) goto L%ld;\n\tbreak;\n", p->res);

	fprintf(fp, "      case QH_Keyword:\n\tif(");
	memstack_ref(fp, p->a1);
	fprintf(fp, ".data.i == QH_Key_Null) goto L%ld;\n\tbreak;\n", p->res);

	fputs("    }\n", fp);
	break;

      case IFEQGOTO:
      case IFNQGOTO:
      case IFLEGOTO:
      case IFLTGOTO:
      case IFGEGOTO:
      case IFGTGOTO:
	fprintf(fp, "    qh_pc = %d;\n", pc);
	fprintf(fp, "    qh_%s(prog + %d);\n", opstr(p->op), pc);
	fprintf(fp, "    if(qh_pc != %d)\n", pc);
	fprintf(fp, "    {\n");
	fprintf(fp, "        qh_pc++;\n");
	fprintf(fp, "        goto pc_jump;\n");
	fprintf(fp, "    }\n");
	break;

      case GOTO:
	fprintf(fp, "    goto L%ld;\n", p->res);
	break;

      case RETURN:

/****************** BUGS( May be occur in the feature ) ******************
 * If call the function from C, the program is not work correctry.
 */
	fprintf(fp, "    {\n");
	fprintf(fp, "\tint ret;\n");
	fprintf(fp, "\tQHObj x;\n");
	fprintf(fp, "\n");
	fprintf(fp, "\tif(isemptyQHFuncStack(&fstk))\n");
	fprintf(fp, "\t{\n");
	fprintf(fp, "\t    qhexit(cast_to_int(&");
	    memstack_ref(fp, p->a1);
	fprintf(fp, "));\n");
	fprintf(fp, "\t}\n");
	fprintf(fp, "\n");
	fprintf(fp, "\tx   = ");
	    memstack_ref(fp, p->a1);
	fprintf(fp, ";\n");
	fprintf(fp, "\tret = fstk.elem[fstk.ptr - 1].ret;\n");
	fprintf(fp, "\tqh_restore();\n");
	fprintf(fp, "\tmemstack[ret] = x;\n");
	fprintf(fp, "\tgoto pc_jump;\n");
	fprintf(fp, "    }\n");
	break;

      case ZERO:
	fputs("    ", fp);
	memstack_ref(fp, p->res);
	fprintf(fp, ".type = %d;\n", QH_Int);
	fputs("    ", fp);
	memstack_ref(fp, p->res);
	fputs(".data.i = 0;\n", fp);
	
      case ONE:
	fputs("    ", fp);
	memstack_ref(fp, p->res);
	fprintf(fp, ".type = %d;\n", QH_Int);
	fputs("    ", fp);
	memstack_ref(fp, p->res);
	fputs(".data.i = 1;\n", fp);

      case END:
	fputs("    qhexit(0);\n", fp);
	break;

      default:
	fprintf(fp, "    qh_%s(prog + %d);\n", opstr(p->op), pc);
	break;
    }
}

static void analisys_baseblock(QHProg* prog, int nprog)
{
    int i;

    volatile_control_code_flag = (char *)xmalloc(sizeof(char) * nprog);
    volatile_control_code_flag[0] = True;
    for(i = 1; i < nprog; i++)
	volatile_control_code_flag[i] = False;

    need_label_flag = (char *)xmalloc(sizeof(char) * nprog);
    need_label_flag[0] = True;
    for(i = 1; i < nprog; i++)
	need_label_flag[i] = False;

    for(i = 0; i < nprog; i++)
    {
	if(IS_GOTOOP(prog[i].op))
	{
	    volatile_control_code_flag[i] = True;
	    need_label_flag[prog[i].res] = True;
	}
	else if(prog[i].op == CALL &&
		memfunctab[prog[i].a1].addr > 0)
	{
	    volatile_control_code_flag[i] = True;
	    need_label_flag[i + 1] = True;
	}
	else if(prog[i].op == FUNCALL)
	{
	    volatile_control_code_flag[i] = True;
	    need_label_flag[i + 1] = True;
	}
	else if(prog[i].op == RETURN)
	{
	    volatile_control_code_flag[i] = True;
	}
    }
    for(i = 0; i < nmemfunc; i++)
	if(memfunctab[i].addr > 0)
	    need_label_flag[memfunctab[i].addr] = True;
}

static void write_pc_jump(FILE* fp, int nprog)
{
    fputs("  pc_jump:\n", fp);
    fputs("    switch(qh_pc)\n", fp);
    fputs("    {\n", fp);
    fputs("      case -1: return;	/* return to C routine */\n", fp);

/*    for(i = 0; i < nprog; i++)
	if(need_label_flag[i])
	    fprintf(fp, "      case %d: goto L%ld;\n", i, (long)i);
    fputs("      default:\n\t", fp);
    fputs("sprintf(temp_buff, \"Systematic error: undefined label: %d\\n\", qh_pc);\n", fp);
    fputs("\texecute_error(temp_buff);\n", fp);
    fputs("    }\n", fp); */
}

static void write_execute_program1(FILE* fp, QHProg* prog, int nprog)
{
    int i;

    fputs("void baseblock_proc0(void)\n", fp);
    fputs("{\n", fp);

    for(i = 1; i < nprog; i++)
    {
	if(volatile_control_code_flag[i]) /* jump code */
	{
	    while(i + 1 < nprog &&
		  volatile_control_code_flag[i + 1])
		i++;

	    if(i + 1 == nprog)
		break;

	    if(!need_label_flag[i+1])
	    {
		fputs("}\n\n", fp);
		fprintf(fp, "static void baseblock_proc%d(void)\n", i+1);
		fputs("{\n", fp);
	    }
	    continue;
	}

	if(need_label_flag[i])	/* start of block */
	{
	    fputs("}\n\n", fp);
	    fprintf(fp, "static void baseblock_proc%d(void)\n", i);
	    fputs("{\n", fp);
	}

	fprintf(fp, "/* %s */\n", opstr(prog[i].op));
	write_C_code(fp, prog + i, i);
    }

    fputs("}\n\n", fp);
}

static void write_execute_program2(FILE* fp, QHProg* prog, int nprog)
{
    int i;

    fputs("void execute_program(void)\n", fp);
    fputs("{\n", fp);
    write_pc_jump(fp, nprog);

    fputs("\n", fp);
    fputs("      case 0:\n", fp);
    fputs("      L0:\n", fp);
    fputs("\tbaseblock_proc0();\n", fp);

    for(i = 1; i < nprog &&
	       !volatile_control_code_flag[i] &&
	       !need_label_flag[i];
	i++);

    while(i < nprog)
    {
	if(!volatile_control_code_flag[i])
	{
	    if(need_label_flag[i])
	    {
		fprintf(fp, "      case %d:\n", i);
		fprintf(fp, "      L%d:\n", i);
	    }
	    fprintf(fp, "\tbaseblock_proc%d();\n", i);
	    i++;
	}

	while(i < nprog &&
	      !volatile_control_code_flag[i] &&
	      !need_label_flag[i])
	    i++;

	while(i < nprog && volatile_control_code_flag[i])
	{
	    if(need_label_flag[i])
	    {
		fprintf(fp, "      case %d:\n", i);
		fprintf(fp, "      L%d:\n", i);
	    }
	    fprintf(fp, "/* %s */\n", opstr(prog[i].op));
	    write_C_code(fp, prog + i, i);
	    i++;
	}
    }

    fputs("        break;\n", fp);
    fputs("      default:\n\t", fp);
    fputs("sprintf(temp_buff, \"Systematic error: undefined label: %d\\n\", qh_pc);\n", fp);
    fputs("\texecute_error(temp_buff);\n", fp);
    fputs("    }\n", fp);
    fputs("}\n\n", fp);
}

void compile_to_C(FILE* fp, QHProg* prog, int nprog)
{
    analisys_baseblock(prog, nprog);

    write_global_definition(fp, prog, nprog);

    write_execute_program1(fp, prog, nprog); /* base block */
    write_execute_program2(fp, prog, nprog);

    write_main(fp, prog, nprog);

    xfree(volatile_control_code_flag);
    xfree(need_label_flag);
}
