/* Copyright (c) 1995, Masanao Izumo */

/* qhclib.c: The library for qh compiler */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
/* #include "qhclib.h" */
#include "common.h"
#include "interp.h"
#include "qhfile.h"
#if defined(sgi)
#include <strings.h>
#endif

char temp_buff[BUFSIZ];

int prog_argc;
char** prog_argv;

int nmemfunc ;
MemFunction* memfunctab ;

QHFuncStack fstk ;
QHObj null_qhobj ;
int qh_pc ;
int qh_sp ;
QHObj* memstack ;
intStack istack ;
int nmemfunc ;
int memstack_size ;
int qh_top ;
int print_garbage_collection = False;
char* display_host_name = NULL;
QHStructTagTable* struct_tag_table;
QHProg* qhprog;

int object_size_list[QH_LastType];

void execute_error(const char* errmsg)
{
    if(isemptyQHFuncStack(&fstk))
    {
	fprintf(stderr, "%s.\n", errmsg);
    }
    else
    {
	QHFunc f;

	f = topQHFuncStack(&fstk);
	fprintf(stderr, "%s, in function `%s'.\n", errmsg,
		memfunctab[f.func].name);
    }

    qhexit(1);
}

void invalid_operation(int op)
{
    sprintf(temp_buff, "Invalid internal operation `%s'", opstr(op));
    execute_error(temp_buff);
}

void free_memory_in_compiling_at_once(void)
{
}

void init_interp(void)
{
    memstack_size = MEMSTACK_SIZE;
    memstack = (QHObj *)xmalloc(sizeof(QHObj) * memstack_size);
    bzero((char *)memstack, sizeof(QHObj) * memstack_size);

    bzero((char *)&null_qhobj, sizeof(null_qhobj));
    null_qhobj.type = QH_Int;

    object_size_list[QH_Int] = sizeof(int);
    object_size_list[QH_Real] = sizeof(double);
    object_size_list[QH_Complex] = sizeof(XComplex);
    object_size_list[QH_Array] = 0;
    object_size_list[QH_String] = sizeof(char);
}

void qhexit(int status)
{
    end_qhfile();
/*    free_cell_object(); */
    exit(status);
}

void qhlibint(void)
{
    init_cell_object();
    init_interp();
    makeQHFuncStack(&fstk, 256);
    makeintStack(&istack, 256);
    init_qhfile();
}

char* string_expand(char *t, const char *s)
{
    char* p = t;
    while(*s)
    {
        if(*s == '\\')
        {
            s++;
            switch(*s)
            {
#define EXPAND(a, b) case a: *t++ = b; break;
                EXPAND('n', '\n');
                EXPAND('t', '\t');
                EXPAND('v', '\v');
                EXPAND('b', '\b');
                EXPAND('r', '\r');
                EXPAND('f', '\f');
                EXPAND('a', '\a');
                EXPAND('0', '\0');
#undef EXPAND
              default:
		*t++ = *s;
            }
	    s++;
	}
	else
	    *t++ = *s++;
    }
    *t = '\0';
    return p;
}
