/* Copyright (c) 1995, Masanao Izumo */

/* qhclib.c: The library for QuickHandy compiler */

#ifndef ___QHCLIB_H_
#define ___QHCLIB_H_

#include <stdio.h>
#include <stdlib.h>
#include "defs.h"
#include "xmalloc.macros"
#include "intStack.h"
#include "QHFunc.defs"
#include "QHFuncStack.h"
#include "qhtypes.h"

#define	FIRST_INSTRUCTION	258
extern int prog_argc;
extern char** prog_argv;

#define QH_FUNCTION(fname) void qhf_ ## fname(int ARGC, QHObj* ARGV, QHObj* RESULT)

extern char temp_buff[];	/* for using temporary string */
void qhexit(int status);

typedef struct _QHProg
{
    long op;
    long res;
    long a1;
    long a2;
} QHProg;

typedef struct _QHProgTraceInfo
{
    unsigned short fname_entry;	/* hash table entry of program source file */
    unsigned short lineno;	/* line number of source file */
} QHProgTraceInfo;

typedef struct _XComplex
{
    double re, im;
} XComplex;


typedef struct _QHObj
{
    int type;                   /* QH_Int,
                                   QH_Real,
                                   QH_Complex,
                                   QH_Array,
                                   QH_AssocArray,
                                   QH_String,
                                   QH_Struct,
                                   QH_List */
    union
    {
        int     i;              /* integer */
        double  r;              /* real */
        XComplex c;             /* complex */
        union _CELL*   cell;    /* cell object */
        struct _Array*  a;      /* array */
        struct _AssocArray* assoc; /* association array */
        struct _QHString* s;    /* string */
        struct _QHStruct* st;   /* struct */
        struct _QHList* list;   /* linked list */
	struct _QHExtData* ext;	/* external definition data */
        void*   v;              /* void pointer (reference to any object */
    } data;                     /* data value */
} QHObj;                        /* object */

struct CELLObjHeader
{
    union _CELL* next;          /* next CELL pointer */
    union _CELL* enext;		/* next CELL pointer */
    short type;                 /* object type: QH_Array or None */
    short mark;                 /* mark bit for garbage collection */
    short shared_data;		/* shared data object if True*/
};

typedef struct _Array
{
    struct CELLObjHeader header; /* header */
    short type;                 /* elem data type:
                                   QH_Int QH_Real QH_Complex
                                   QH_Any */
    short dim;                  /* dimension */
    int nelem;                  /* number of elem */
    int* size;                  /* each size of data */
    void* elem;                 /* array datas */
} Array;

typedef struct _AssocArray
{
    struct CELLObjHeader header; /* header */
    struct _QHObjAssocArr* tab; /* assoc table */
} AssocArray;

typedef struct _QHString
{
    struct CELLObjHeader header; /* header */
    int maxlen;                 /* maximum string buffer length */
    int len;                    /* current string length */
    Bool isconst;               /* is constant string */
    char* str;                  /* string buffer */
    Bool byte_data;		/* True if data object */
} QHString;

typedef struct _QHStruct
{
    struct CELLObjHeader header; /* header */
    short entry;                /* entry to struct table */
    short n;                    /* number of member */
    long* id;                   /* member id list */
    QHObj* member;              /* member */
} QHStruct;

typedef struct _QHList
{
    struct CELLObjHeader header; /* header */
    QHObj car;
    QHObj cdr;
} QHList;

typedef struct _QHExtData
{
    struct CELLObjHeader header; /* header */
    int type;
    void* data;
    void (*marker)(struct _QHExtData *);
    void (*destructor)(struct _QHExtData *);
} QHExtData;

typedef union _CELL
{
    struct CELLObjHeader header; /* header */
    Array a;                    /* array */
    AssocArray assoc;           /* association array */
    QHString s;                 /* string */
    QHStruct st;                /* struct */
    QHList list;                /* linked list */
    QHExtData ext;		/* external defined data */
} CELL;

typedef struct _MemFunction
{
    int   addr;			/* absolute address of function entrance */
    char* name;			/* function name */
    char* module_name;		/* module name */
    short narg;			/* number of arguments */
    short offsp;		/* incremental SP */
    void (*extern_function)();	/* nonnil if extern function */
} MemFunction;

typedef struct _QHStructTagTable
{
    int nmember;
    long* member_id;
    struct _QHStructTagTable** next; /* list of next table entry */
} QHStructTagTable;

extern int qh_top;
extern int qh_sp;
extern int qh_pc;
extern QHObj* memstack;
extern QHObj null_qhobj;
extern MemFunction* memfunctab;
extern int memstack_size;
extern char* display_host_name;
extern QHStructTagTable* struct_tag_table;
extern int qh_extern_argc;

QHString* alloc_constant_string(const char *);
void qh_restore(void);
void extern_unload_function(int argc, QHObj* argv, QHObj* ret);
void qhlibint(void);
void execute_error(const char* errmsg);
extern int OK_DO_garbage_collecting;

CELL* alloc_cell(void);
Array* alloc_array(void);
QHString* alloc_constant_string(const char* str);
QHString* alloc_string(void);


extern QHFuncStack fstk;
extern intStack istack;
extern char* hash_table [];
extern QHProg* qhprog;
extern int nmemfunc;
#include "functab.h"
#include "funcall.h"
#include "builtin.h"


#endif /* ___QHCLIB_H_ */
