/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/param.h>
#if defined(linux) || defined(__alpha__)
#include <unistd.h>
#else
#include <sys/unistd.h>
#include <unistd.h>
#endif
#include "common.h"
#include "interp.h"
#include "qhfile.h"

#define QH_MAXFILENO MIN(16, NOFILE)

#define CHECK_FILENO(entry) if((entry) < 0 || (entry) >= QH_MAXFILENO) bad_file_number(entry);

#define CHECK_OPEN(entry)					\
if(QHFILETYPE(entry) == QHF_Closed)				\
{								\
    sprintf(temp_buff, "file number `%d' is closed", entry);	\
    execute_error(temp_buff);					\
}


enum qhfile_types
{
    QHF_Closed,			/* closed file */
    QHF_Ordinary,		/* ordinary file: use the FILE */
    QHF_FPipe,			/* pipe: use the FILE */
    QHF_FDFile,			/* use the file discripter */
    QHF_Socket			/* socket ( may be used next version ) */
};

typedef struct _QHFile
{
    int type;			/* see qhfile_types */
    union
    {
	int fd;			/* file discripter */
	FILE* fp;		/* file pointer ( may be used next version ) */
    } u;
} QHFile;

#define QHFILETYPE(entry) qhfile[entry].type
#define QHFILEFD(entry) qhfile[entry].u.fd
#define QHFILEFP(entry) qhfile[entry].u.fp

static QHFile qhfile[QH_MAXFILENO];
static int min_fileno;

static void bad_file_number(int entry)
{
    sprintf(temp_buff, "Bad file entry number: %d\n", entry);
    execute_error(temp_buff);
}

void init_qhfile(void)
{
    int i;

    QHFILEFP(0) = stdin;
    QHFILETYPE(0) = QHF_Ordinary;

    QHFILEFP(1) = stdout;
    QHFILETYPE(1) = QHF_Ordinary;

    QHFILEFP(2) = stderr;
    QHFILETYPE(2) = QHF_Ordinary;

    for(i = 3; i < QH_MAXFILENO; i++)
	QHFILETYPE(i) = QHF_Closed;
    min_fileno = 3;
}

static void close_qhfile(int entry)
{
    switch(QHFILETYPE(entry))
    {
      case QHF_Ordinary:
	fclose(QHFILEFP(entry));
	break;
      case QHF_FPipe:
	pclose(QHFILEFP(entry));
	break;
      case QHF_FDFile:
	close(QHFILEFD(entry));
	break;
      case QHF_Socket:
	fprintf(stderr, "## Socket isn't supported.\n");
	break;
    }
}

void end_qhfile(void)
{
    int i;

    for(i = 3; i < QH_MAXFILENO; i++)
	if(QHFILETYPE(i) != QHF_Closed)
	    close_qhfile(i);
}

static int new_fentry(void)
{
    int i;    

    for(i = min_fileno; i < QH_MAXFILENO; i++)
    {
	if(QHFILETYPE(i) == QHF_Closed)
	{
	    min_fileno = i + 1;
	    return i;
	}
    }

    execute_error("Too many file open");

    return 0;			/* dmy */
}

/* fopen("filename", mode, permission); */
void qh_xfopen(QHObj* ret)
{
    QHObj* file_mode = STACK(2);
    char* fname = cast_to_string(STACK(1));
    int parm = cast_to_int(STACK(3));
    int entry = 0;

    if(BETWEEN(QH_Int, file_mode->type, QH_Complex))
    {
	int mode = cast_to_int(file_mode);
	int fd;

	if((O_ACCMODE & mode) == O_WRONLY ||
	   (O_ACCMODE & mode) == O_RDWR)
	    mode |= O_CREAT;

	if((fd = open(fname, mode, parm)) < 0)
	{
	    perror(fname);
	    qhexit(1);
	}

	entry = new_fentry();
	QHFILETYPE(entry) = QHF_FDFile;
	QHFILEFD(entry)   = fd;
    }
    else if(IS_QHSTRING(file_mode))
    {
	FILE* fp;
	char* mode;

	mode = file_mode->data.s->str;
	if((fp = fopen(fname, mode)) == NULL)
	{
	    perror(fname);
	    qhexit(1);
	}

	entry = new_fentry();
	QHFILETYPE(entry) = QHF_Ordinary;
	QHFILEFP(entry)   = fp;
    }
    else
    {
	execute_error("Invalid file mode argument");
    }

    ret->type = QH_Int;
    ret->data.i = entry;
}

void qh_xpopen(QHObj* ret)
{
    char* com = cast_to_string(STACK(1));
    char* mode = cast_to_string(STACK(2));
    int entry;
    FILE* fp;

    if((fp = popen(com, mode)) == NULL)
    {
	perror(com);
	qhexit(1);
    }
    entry = new_fentry();
    QHFILETYPE(entry) = QHF_FPipe;
    QHFILEFP(entry)   = fp;
    ret->type = QH_Int;
    ret->data.i = entry;
}

/* fclose(fd); */
void qh_xfclose(QHObj* ret)
{
    int entry = cast_to_int(STACK(1));
    CHECK_FILENO(entry);

    if(QHFILETYPE(entry) != QHF_Closed)
    {
	close_qhfile(entry);
	QHFILETYPE(entry) = QHF_Closed;
	if(min_fileno > entry)
	    min_fileno = entry;
    }
}

/* fwrite/3(fileno, array, nobj)
 * fwrite/2(fileno, obj);
 */
void qh_xfwrite3(QHObj* ret)
{
    int   entry = cast_to_int(STACK(1));
    QHObj* x    = STACK(2);
    int   len   = cast_to_int(STACK(3));
    void* obj;
    int   n;
    CHECK_FILENO(entry);
    CHECK_OPEN(entry);

    if(x->type != QH_Array &&
       x->type != QH_String)
	execute_error("arg 2 of `fwrite/2,3'  isn't an array object");

    if(x->type == QH_Array)
    {
	obj = x->data.a->elem;
	n   = MIN(x->data.a->nelem,
		  len * sizeof_qhtype(x->data.a->type));
    }
    else
    {
	obj = (void *)x->data.s->str;
	n   = MIN(x->data.s->maxlen, len);
    }

    switch(QHFILETYPE(entry))
    {
      case QHF_Ordinary:
      case QHF_FPipe:
	fwrite((char *)obj, sizeof(char), n, QHFILEFP(entry));
	break;

      case QHF_FDFile:
	if((n = write(QHFILEFD(entry), obj, n)) < 0)
	{
	    perror("write");
	    qhexit(1);
	}
	break;

      case QHF_Socket:
	fprintf(stderr, "## Socket isn't supported yet.\n");
	break;
    }

    ret->type   = QH_Int;
    ret->data.i = n;
}

void qh_xfwrite2(QHObj* ret)
{
    int entry   = cast_to_int(STACK(1));
    QHObj* x = STACK(2);
    void* obj = NULL;
    int size = 0;
    CHECK_FILENO(entry);
    CHECK_OPEN(entry);

    if(x->type == QH_Array)
    {
	obj = x->data.a->elem;
	size = x->data.a->nelem * sizeof_qhtype(x->data.a->type);
    }
    else if(x->type == QH_String)
    {
	obj = (void *)x->data.s->str;
	if(x->data.s->str == NULL)
	    size = 0;
	else
	    size = strlen(x->data.s->str) + 1;
    }
    else
    {
	switch(x->type)
	{
	  case QH_Int:
	    obj = &x->data.i;
	    size = sizeof(int);
	    break;
	  case QH_Real:
	    obj = &x->data.r;
	    size = sizeof(double);
	    break;
	  case QH_Complex:
	    obj = &x->data.c;
	    size = sizeof(XComplex);
	    break;
	  default:
	    execute_error("Invalid argument of fwrite/2");
	    break;
	}
    }

    switch(QHFILETYPE(entry))
    {
      case QHF_Ordinary:
      case QHF_FPipe:
	fwrite((char *)obj, sizeof(char), size, QHFILEFP(entry));
	break;

      case QHF_FDFile:
	if(write(QHFILEFD(entry), obj, size) < 0)
	{
	    perror("write");
	    qhexit(1);
	}
	break;

      case QHF_Socket:
	fprintf(stderr, "## Socket isn't supported yet.\n");
	break;
    }

    ret->type   = QH_Int;
    ret->data.i = size;
}

/* read-bytes = fread(fd, array, nobj);
 */
void qh_xfread(QHObj* ret)
{
    int n = 0;
    int entry = cast_to_int(STACK(1));
    QHObj* x  = STACK(2);
    int nobj = cast_to_int(STACK(3));
    void* obj;
    int   unit;
    int   size;
    CHECK_FILENO(entry);

    if(nobj <= 0)
    {
	ret->type = QH_Int;
	ret->data.i = 0;
	return;
    }

    if(x->type != QH_Array &&
       x->type != QH_String)
	execute_error("arg 2 of `fread/3' isn't an array or string");

    CHECK_OPEN(entry);

    if(x->type == QH_Array)
    {
	int len;

	obj  = x->data.a->elem;
	unit = sizeof_qhtype(x->data.a->type);
	len  = x->data.a->nelem;
	size = unit * MIN(len, nobj);
    }
    else
    {
	touch_string(x->data.s, nobj);
	obj  = (void *)x->data.s->str;
	size = nobj;
    }

    switch(QHFILETYPE(entry))
    {
      case QHF_Ordinary:
      case QHF_FPipe:
	n = fread((char *)obj, sizeof(char), size, QHFILEFP(entry));
	break;

      case QHF_FDFile:
	if((n = read(QHFILEFD(entry), obj, size)) < 0)
	{
	    perror("read");
	    qhexit(1);
	}
	break;

      case QHF_Socket:
	fprintf(stderr, "## Socket isn't supported yet.\n");
	n = 0;
	break;
    }

    ret->type = QH_Int;
    ret->data.i = n;
}

/* ch = fgetc(fd); */
void qh_xfgetc(QHObj* ret)
{
    int entry = cast_to_int(STACK(1));
    char ch;
    int n;
    CHECK_FILENO(entry);
    CHECK_OPEN(entry);

    ret->type = QH_Int;
    switch(QHFILETYPE(entry))
    {
      case QHF_Ordinary:
      case QHF_FPipe:
	ret->data.i = fgetc(QHFILEFP(entry));
	break;

      case QHF_FDFile:
	if((n = read(QHFILEFD(entry), &ch, 1)) < 0)
	{
	    perror("read");
	    qhexit(1);
	}

	if(n != 0)
	    ret->data.i = ch;
	else
	    ret->data.i = EOF;
	break;

      case QHF_Socket:
	fprintf(stderr, "## Socket isn't supported yet.\n");
	break;
    }
}

/* fputc(ch, fd); */
void qh_xfputc(QHObj* ret)
{
    char ch   = cast_to_int(STACK(1));
    int entry = cast_to_int(STACK(2));
    CHECK_FILENO(entry);
    CHECK_OPEN(entry);

    switch(QHFILETYPE(entry))
    {
      case QHF_Ordinary:
      case QHF_FPipe:
	putc(ch, QHFILEFP(entry));
	break;

      case QHF_FDFile:
	if(write(QHFILEFD(entry), &ch, 1) != 1)
	{
	    perror("write");
	    qhexit(1);
	}
	break;

      case QHF_Socket:
	fprintf(stderr, "## Socket isn't supported yet.\n");
	break;
    }
}

/* ftrunc(fd,       bytes);
   ftrunc(filename, bytes);
*/
void qh_xftrunc(QHObj* ret)
{
    QHObj* x = STACK(1);
    int n    = cast_to_int(STACK(2));
    int entry;
    int fd = 0;

    if(IS_QHSTRING(x))
    {
	if(x->data.s->str == NULL)
	    execute_error("String is NULL");
	if(truncate(x->data.s->str, n) < 0)
	{
	    perror((char *)x->data.a->elem);
	    qhexit(1);
	}
    }
    else
    {
	if(x->type >= QH_String)
	    execute_error("Wrong arg 1");
	entry = cast_to_int(x);
	CHECK_FILENO(entry);
	CHECK_OPEN(entry);

	switch(QHFILETYPE(entry))
	{
	  case QHF_Ordinary:
	    fd = fileno(QHFILEFP(entry));
	    break;
	  case QHF_FDFile:
	    fd = QHFILEFD(entry);
	    break;
	  default:
	    execute_error("Invalid file type");
	    break;
	}

	if(ftruncate(fd, n) < 0)
	{
	    perror("ftruncate");
	    exit(1);
	}
    }
}

/* fseek(fd, offset, whence); */
void qh_xfseek(QHObj* ret)
{
    int entry  = cast_to_int(STACK(1));
    int offset = cast_to_int(STACK(2));
    int whence = cast_to_int(STACK(3));
    int pos = 0;
    CHECK_FILENO(entry);
    CHECK_OPEN(entry);

    switch(QHFILETYPE(entry))
    {
      case QHF_Ordinary:
	if((pos = fseek(QHFILEFP(entry), offset, whence)) == -1)
	{
	    perror("fseek");
	    qhexit(1);
	}
	break;
      case QHF_FDFile:
	if((pos = lseek(QHFILEFD(entry), offset, whence)) == -1)
	{
	    perror("lseek");
	    qhexit(1);
	}
	break;
      default:
	execute_error("Invalid file type");
	break;
    }

    ret->type = QH_Int;
    ret->data.i = pos;
}

static int fgetline1(int entry, char* buff, int size)
{
    buff[0] = '\0';
    switch(QHFILETYPE(entry))
    {
      case QHF_Ordinary:
      case QHF_FPipe:
	if(fgets(buff, size, QHFILEFP(entry)) == NULL)
	    return 0;
	return strlen(buff);

      case QHF_FDFile:
      case QHF_Socket:
	{
	    int i, n;
	    char ch;

	    for(i = 0; i < size - 1; i++)
	    {
		if((n = read(QHFILEFD(entry), &ch, 1)) < 0)
		{
		    perror("read");
		    qhexit(1);
		}

		if(n == 0)	/* end of file */
		{
		    buff[i] = '\0';
		    return i;
		}

		buff[i] = ch;
		if(ch == '\n')	/* end of line */
		{
		    buff[i + 1] = '\0';
		    return i + 1;
		}
	    }
	    buff[i] = '\0';	/* read more */
	    return i;
	}
    }
    return 0;		/* dmy */
}

/* getlineop res fiz - */
void qh_getlineop(QHProg* p)
{
    QHObj* res = memstack + MEMADDR(p->res);
    int entry  = cast_to_int(memstack + MEMADDR(p->a1));
    char* s;
    char* temp_ptr;
    int len, size;
    Bool is_data_block = False;

    CHECK_FILENO(entry);
    CHECK_OPEN(entry);

    /* get a line */
    if(data_block_ptr == NULL)
    {
	temp_ptr = temp_buff;
	len = fgetline1(entry, temp_ptr, BUFSIZ);
    }
    else
    {
	temp_ptr = data_block_ptr->data;
    	len = fgetline1(entry, temp_ptr, DATA_BLOCK_SIZE);
	is_data_block = True;
    }

    if(len == 0)		/* EOF */
    {
	res->type = QH_Int;
	res->data.i = 0;
	return;
    }

    if(temp_ptr[len - 1] == '\n')
    {
	size = len;
	if(is_data_block)
	    s = temp_ptr;
	else
	    s = xstrdup(temp_ptr);
    }
    else
    {
	add_string_to_flexible_buff(temp_ptr);
	is_data_block = False;
	do
	{
	    if(fgetline1(entry, temp_buff, BUFSIZ) == 0)
		break;
	    add_string_to_flexible_buff(temp_buff);
	} while(temp_buff[len - 1] != '\n');
	size = flexible_buff_size;
	s = get_flexible_buff();
    }

    res->data.s = alloc_string();
    res->data.s->str = s;
    res->data.s->maxlen = size;
    res->data.s->header.type = QH_String;
    if(is_data_block)
    {
	res->data.s->dbp = data_block_ptr;
	data_block_ptr = data_block_ptr->next;
    }

    res->type = QH_String;
}

void qh_xfcreate(QHObj* ret)
{
    char* path = cast_to_string(STACK(1));
    int mode  = cast_to_int(STACK(2));
    int fd;
    int entry;

    if((fd = creat(path, mode)) < 0)
    {
	perror(path);
	qhexit(1);
    }

    entry = new_fentry();
    QHFILETYPE(entry) = QHF_FDFile;
    QHFILEFD(entry)   = fd;
    ret->type         = QH_Int;
    ret->data.i       = entry;
}

void qh_xfdelete(QHObj* ret)
{
    char* path = cast_to_string(STACK(1));

    if(unlink(path) < 0)
    {
	perror(path);
	qhexit(1);
    }
}
