/* Copyright (c) 1995, Masanao Izumo */

%{
#include <stdlib.h>
#include "common.h"
#include "qhlex.h"
#include "xmalloc.macros"
#ifndef yywrap
#define yywrap() 1
#endif
%}

WHITE			[ \t]
NEWLINE			"\n"
DIGITS			[0-9]+
INUMBER			([0-9]+|0x[0-9a-fA-F]+)[lLuU]?
RNUMBER			({DIGITS}|({DIGITS}\.{DIGITS}))([eE][+-]?{DIGITS})?
IMNUMBER		({DIGITS}(\.{DIGITS})?)([eE][+-]?{DIGITS})?(i|I)
STR			\"([^\\"]|\\.)*\"
CHAR			'\\?.'
IDENT			[a-zA-Z_][a-zA-Z0-9_]*

%%
{WHITE}+	;
{NEWLINE}	{ qhlineno++; }
^#pragma{WHITE}* { return PRAGMA; }
"# "{DIGITS}" ".* { cpp_command(yytext); }
"+="		{ set_current_textpos(&yylval.tp); return AASSIGN; }
"-="		{ set_current_textpos(&yylval.tp); return SASSIGN; }
"*="		{ set_current_textpos(&yylval.tp); return MASSIGN; }
"/="		{ set_current_textpos(&yylval.tp); return DASSIGN; }
"%="		{ set_current_textpos(&yylval.tp); return RASSIGN; }
"||"		{ set_current_textpos(&yylval.tp); return OR;      }
"&&"		{ set_current_textpos(&yylval.tp); return AND;     }
"=="		{ set_current_textpos(&yylval.tp); return EQ;      }
"!="		{ set_current_textpos(&yylval.tp); return NQ;      }
"<="		{ set_current_textpos(&yylval.tp); return LE;      }
">="		{ set_current_textpos(&yylval.tp); return GE;      }
"++"		{ set_current_textpos(&yylval.tp); return INC_OP;  }
"--"		{ set_current_textpos(&yylval.tp); return DEC_OP;  }
"<<"		{ set_current_textpos(&yylval.tp); return LLSHIFT; }
">>"		{ set_current_textpos(&yylval.tp); return LRSHIFT; }
".<"		{ set_current_textpos(&yylval.tp); return ALSHIFT; }
".>"		{ set_current_textpos(&yylval.tp); return ARSHIFT; }
"::"		{ return MODULE_SPEC; }
"..."		{ return DOTDOTDOT; }
QHKEYWORD	{ return QHKEYWORD; }
funcall		{ set_current_textpos(&yylval.tp); return FUNCALL; }
sizeof		{ set_current_textpos(&yylval.tp); return SIZEOF_OP; }
structof	{ set_current_textpos(&yylval.tp); return STRUCTOF;}
loadfn		{ return LOADFN; }
extern		{ return QH_EXTERN; }
function	{ return FUNCTION; }
module		{ return MODULE; }
local		{ return LOCAL; }
goto		{ return GOTO; }
if		{ return IF; }
else		{ return ELSE; }
for		{ return FOR; }
do		{ return DO; }
while		{ return WHILE; }
switch		{ return SWITCH; }
case		{ return CASE; }
default		{ return DEFAULT; }
break		{ return BREAK; }
continue	{ return CONTINUE; }
return		{ set_current_textpos(&yylval.tp); return RETURN; }
write		{ return WRITE; }
writeln		{ return WRITELN; }
swrite		{ return SWRITE; }
new		{ set_current_textpos(&yylval.tp); return NEW; }
global		{ return GLOBAL; }
sin		{ yylval.ival = SIN;	return MATHFUNC1; }
cos		{ yylval.ival = COS;	return MATHFUNC1; }
tan		{ yylval.ival = TAN;	return MATHFUNC1; }
asin		{ yylval.ival = ASIN;	return MATHFUNC1; }
acos		{ yylval.ival = ACOS;	return MATHFUNC1; }
atan		{ yylval.ival = ATAN;	return MATHFUNC1; }
sinh		{ yylval.ival = SINH;	return MATHFUNC1; }
cosh		{ yylval.ival = COSH;	return MATHFUNC1; }
tanh		{ yylval.ival = TANH;	return MATHFUNC1; }
asinh		{ yylval.ival = ASINH;	return MATHFUNC1; }
acosh		{ yylval.ival = ACOSH;	return MATHFUNC1; }
atanh		{ yylval.ival = ATANH;	return MATHFUNC1; }
conjugate	{ yylval.ival = CONJUGATE;	return MATHFUNC1; }
arg		{ yylval.ival = ARG;	return MATHFUNC1; }
abs		{ yylval.ival = ABS;	return MATHFUNC1; }
floor		{ yylval.ival = FLOOR;	return MATHFUNC1; }
ceil		{ yylval.ival = CEIL;	return MATHFUNC1; }
exp		{ yylval.ival = EXP;	return MATHFUNC1; }
log		{ yylval.ival = LOG;	return MATHFUNC1; }
log10		{ yylval.ival = LOG10;	return MATHFUNC1; }
log2		{ yylval.ival = LOG2;	return MATHFUNC1; }
sqrt		{ yylval.ival = SQRT;	return MATHFUNC1; }
int		{ yylval.ival = INT_OP;	return MATHFUNC1; }
real		{ yylval.ival = REAL_OP;	return MATHFUNC1; }
image		{ yylval.ival = IMAGE_OP;	return MATHFUNC1; }
complex		{ yylval.ival = COMPLEX_OP;	return MATHFUNC2; }
string		{ return STRING_OP; }
assoc		{ return ASSOC_SPEC; }
atan2		{ yylval.ival = ATAN2;	return MATHFUNC2; }
rand		{ return RAND; }
struct		{ return STRUCT; }
enum		{ return ENUM; }
volatile	{ return VOLATILE; }
{INUMBER}	{
			set_current_textpos(&yylval.atm.tp);
			yylval.atm.name = int_convert(yytext);
			return INTEGER;
		}
{RNUMBER}	{
			set_current_textpos(&yylval.atm.tp);
			yylval.atm.name = install_string(yytext);
			return REAL;
		}
{IMNUMBER}	{
			set_current_textpos(&yylval.atm.tp);
			yylval.atm.name = install_string(yytext);
			return IMNUMBER;
		}
{CHAR}		{
			set_current_textpos(&yylval.atm.tp);
			yylval.atm.name = char_code_convert(yytext);
			return INTEGER;
		}
{STR}		{
			int i;
			for(i = 0; yytext[i]; i++)
			    if(yytext[i] == '\n')
				qhlineno++;
			set_current_textpos(&yylval.atm.tp);
			yylval.atm.name = qhstringrep(yytext);
			return STR;
		}
{IDENT}		{
			set_current_textpos(&yylval.atm.tp);
			yylval.atm.name = install_string(yytext);
			return IDENTIFIER;
		}
@.*		{ puts(yytext); }
"//".*		{ }
^#!.*		{ }
.		{ set_current_textpos(&yylval.tp); return yytext[0]; }
%%
