/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include "common.h"
#include "cpp.defs"

int qhlineno   = 1;
int qhfilename = 0;
Bool isqhfileopen = False;

extern FILE* cpp_popen(const char* filename, char* path, char** argv);
extern void cpp_pclose(FILE* fp);

#define MAXNAMELEN 256

void set_current_textpos(QHTextPos* tp)
{
    tp->lineno = qhlineno;
    tp->fname  = qhfilename;
}

static void print_cpp_command(void)
{
    int i;

    for(i = 0; cpp_arguments.elem[i]; i++)
	fprintf(stderr, "%s ", cpp_arguments.elem[i]);
}

void qhfileopen(const char* filename)
{
    pushchar_ptrStack(&cpp_arguments, NULL);

    if(verbose_mode)
    {
	fprintf(stderr, "cpp using: ");
	print_cpp_command();
	fprintf(stderr, "\n");
    }

    yyin = cpp_popen(filename, cpp_arguments.elem[0], cpp_arguments.elem);
    isqhfileopen = True;
}

void qhfileclose(void)
{
    if(isqhfileopen)
    {
	cpp_pclose(yyin);
	yyin = stdin;
	isqhfileopen = False;
    }
}

void cpp_command(const char* s)
{
    char buff[MAXNAMELEN];
    int i, j;

    qhlineno = atoi(s + 2) - 1;

    for(i = 2; s[i] != '"'; i++)
	;
    i++;
    for(j = 0; s[i] != '"'; i++, j++)
	buff[j] = s[i];

    buff[j] = '\0';
    qhfilename = install_string(buff);
    qhlineno -= (qhfilename == 0);
}

int qhstringrep(const char* str)
{
    int i;

    str++;			/* skip '"' */
    for(i = 0; *(str + 1) != '\0' && i < BUFSIZ; i++, str++)
	temp_buff[i] = *str;

    temp_buff[i] = '\0';

   return install_string(temp_buff);
}

char* string_expand(char *t, const char *s)
{
    char* p = t;

    while(*s)
    {
        if(*s == '\\')
        {
            s++;
            switch(*s)
            {
#define EXPAND(a, b) case a: *t++ = b; break;
                EXPAND('n', '\n');
                EXPAND('t', '\t');
                EXPAND('v', '\v');
                EXPAND('b', '\b');
                EXPAND('r', '\r');
                EXPAND('f', '\f');
                EXPAND('a', '\a');
                EXPAND('0', '\0');
                EXPAND('\\', '\\');
#undef EXPAND
              default:
		*t++ = *s;
            }
	    if(*s == '\0')
		return p;
	    s++;
	}
	else
	    *t++ = *s++;
    }
    *t = '\0';
    return p;
}

char* string_unexpand(char *t, const char *s)
{
    char* ans = t;

    if(s == NULL)
	return "";

    while(*s)
    {
	switch(*s)
	{
#define UNEXPAND(b, a) case a: *t++ = '\\'; *t++ = b; break;
	    UNEXPAND('n', '\n');
	    UNEXPAND('t', '\t');
	    UNEXPAND('v', '\v');
	    UNEXPAND('b', '\b');
	    UNEXPAND('r', '\r');
	    UNEXPAND('f', '\f');
	    UNEXPAND('a', '\a');
	    UNEXPAND('"', '"');
	    UNEXPAND('\\', '\\');
#undef UNEXPAND
	  default:
	    *t++ = *s;
	}
	s++;
    }
    *t = '\0';

    return ans;
}

int char_code_convert(const char* s)
{
    int ch;

    /* skip "'" */
    s++;

    if(*s != '\\')
	ch = *s;
    else
    {
	switch(*(s + 1))
	{
#define EXPAND(a, b) case a: ch = b; break;
	    EXPAND('n', '\n');
	    EXPAND('t', '\t');
	    EXPAND('v', '\v');
	    EXPAND('b', '\b');
	    EXPAND('r', '\r');
	    EXPAND('f', '\f');
	    EXPAND('a', '\a');
	    EXPAND('0', '\0');
#undef EXPAND
	  default:
	    ch = *(s + 1);
	}
    }

    sprintf(temp_buff, "%d", ch);
    return install_string(temp_buff);
}

int int_convert(const char* s)
{
    int i;

    if(s[0] == '0')
    {
	if(s[1] == 'x')
	    sscanf(s + 2, "%x", &i);
	else if(s[1] == '\0')
	    i = 0;
	else
	    sscanf(s + 1, "%o", &i);
    }
    else
	i = atoi(s);
    sprintf(temp_buff, "%d", i);
    return install_string(temp_buff);
}

char* module_name_string(long module, long name)
{
    static char* name_buff;
    static int   name_buff_len;
    int len1, len2;

    if(module > 0)
	len1 = strlen(STRING(module));
    else
	len1 = 0;

    len2 = strlen(STRING(name));

    if(name_buff == NULL)
    {
	name_buff_len = (len1 + len2 + 3) * 2;
	if(name_buff_len < 256)
	    name_buff_len = 256;
	name_buff = (char *)xmalloc(sizeof(char) * name_buff_len);
    }
    else
    {
	EXPAND_MEMORY_IF_OVERFLOW(char, name_buff,
				  (len1 + len2 + 3), name_buff_len);
    }

    if(module > 0)
	sprintf(name_buff, "%s::%s", STRING(module), STRING(name));
    else if(module == 0)
	sprintf(name_buff, "%s", STRING(name));
    else /* module == -1 */
	sprintf(name_buff, "::%s", STRING(name));

    return name_buff;
}
