#include <stdio.h>
#include "common.h"
#include "qhlist.h"
#include "interp.h"

static QHObj cons_cell(QHObj* a, QHObj* b)
{
    QHObj obj;

    obj.data.list = (QHList *)alloc_cell();
    obj.data.list->car = *a;
    obj.data.list->cdr = *b;
    obj.data.list->header.type = QH_List;
    obj.type = QH_List;

    return obj;
}

void qh_list_op(QHProg* p)
{
    int list_length;
    QHObj* x = memstack + MEMADDR(p->res);

    list_length = p->a1;

    *x = memstack[Q_NULL];
    if(list_length == 0)
 	return;
 
    while(list_length-- > 0)
    {
	*x = cons_cell(memstack + qh_top - 1, x);
	qh_top--;
    }
}

void qh_car(QHObj* ret)
{
    QHObj* x = STACK(1);

    if(x->type != QH_List)
	*ret = memstack[Q_NULL];
    else
	*ret = x->data.list->car;
}

void qh_cdr(QHObj* ret)
{
    QHObj* x = STACK(1);

    if(x->type != QH_List)
	*ret = memstack[Q_NULL];
    else
	*ret = x->data.list->cdr;
}

void qh_cons(QHObj* ret)
{
    *ret = cons_cell(STACK(1), STACK(2));
}

QHList* clone_list(const QHList* list)
{
    QHList* clone;

    clone = (QHList *)alloc_cell();

    qh_copy_clone_object(&clone->car, &list->car);
    qh_copy_clone_object(&clone->cdr, &list->cdr);

    clone->header.type = QH_List;

    return clone;
}

Bool equal_list(const QHList* list1, const QHList* list2)
{
    if(!qh_equal_object(&list1->car, &list2->car))
	return False;
    return qh_equal_object(&list1->cdr, &list2->cdr);
}

void qh_rplaca(QHObj* ret)
{
    QHObj* x = STACK(1);
    QHObj* y = STACK(2);

    if(x->type != QH_List)
	execute_error("The argument of 1 is not a list");

    x->data.list->car = *y;
    *ret = *x;
}

void qh_rplacd(QHObj* ret)
{
    QHObj* x = STACK(1);
    QHObj* y = STACK(2);

    if(x->type != QH_List)
	execute_error("The argument of 1 is not a list");

    x->data.list->cdr = *y;
    *ret = *x;
}
