/* Copyright (c) 1995, Masanao Izumo */

#ifndef ___QHOBJ_H_
#define ___QHOBJ_H_

#include "qhtypes.h"

extern int object_size_list[/* enum qh_objact_types */];

enum qh_op2_data_types
{
    QHDATATYPE_II = QH_Int     | QH_Int << 2,
    QHDATATYPE_IR = QH_Int     | QH_Real << 2,
    QHDATATYPE_IC = QH_Int     | QH_Complex << 2,
    QHDATATYPE_RI = QH_Real    | QH_Int << 2,
    QHDATATYPE_RR = QH_Real    | QH_Real << 2,
    QHDATATYPE_RC = QH_Real    | QH_Complex << 2,
    QHDATATYPE_CI = QH_Complex | QH_Int << 2,
    QHDATATYPE_CR = QH_Complex | QH_Real << 2,
    QHDATATYPE_CC = QH_Complex | QH_Complex << 2
};
#define SWITCH_DATA_TYPE(x1, x2) ((x1)->type | ((x2)->type << 2))

#define QHCABS2(z) ((z)->data.c.re * (z)->data.c.re \
		  + (z)->data.c.im * (z)->data.c.im)

#define DATA_BLOCK_SIZE (256 - 8)
#define DATA_BLOCK_AVAILABLE(size) (data_block_ptr != NULL && (size) < DATA_BLOCK_SIZE)
struct data_block
{
    char data[DATA_BLOCK_SIZE];
    struct data_block* next;
};

extern struct data_block* data_block_ptr;

typedef struct _XComplex
{
    double re, im;
} XComplex;

typedef struct _QHObj
{
    int type;			/* QH_Int,
				   QH_Real,
				   QH_Complex,
				   QH_Array,
				   QH_AssocArray,
				   QH_String,
				   QH_Struct,
				   QH_List */
    union
    {
	int     i;		/* integer */
	double  r;		/* real */
	XComplex c;		/* complex */
	union _CELL*   cell;	/* cell object */
	struct _Array*  a;	/* array */
	struct _AssocArray* assoc; /* association array */
	struct _QHString* s;	/* string */
	struct _QHStruct* st;	/* struct */
	struct _QHList* list;	/* linked list */
	struct _QHExtData* ext;	/* external definition data */
	void*   v;		/* void pointer (reference to any object) */
    } data;			/* data value */
} QHObj;			/* object */

struct CELLObjHeader
{
    union _CELL* next;		/* next CELL pointer */
    union _CELL* enext;		/* next CELL pointer */
    short type;			/* object type: QH_Array or None */
    short mark;			/* mark bit for garbage collection */
    short shared_data;		/* shared data object if True*/
};

typedef struct _Array
{
    struct CELLObjHeader header; /* header */
    short type;			/* elem data type:
				   QH_Int QH_Real QH_Complex
				   QH_Any */
    short dim;			/* dimension */
    int* size;			/* each size of data */
    int nelem;			/* number of elem */
    void* elem;			/* array datas */
    struct data_block* dbp;	/* Not used in current versoin */
} Array;

typedef struct _AssocArray
{
    struct CELLObjHeader header; /* header */
    struct _QHObjAssocArr* tab;	/* assoc table */
} AssocArray;

typedef struct _QHString
{
    struct CELLObjHeader header; /* header */
    int maxlen;			/* maximum string buffer length */
    Bool isconst;		/* is constant string */
    unsigned char* str;		/* string buffer:
				   always str[maxlen] == '\0' */
    Bool fixed;			/* True if fixed length string */
    struct data_block* dbp;	/* unless NULL, str is used in data block */
} QHString;

typedef struct _QHStruct
{
    struct CELLObjHeader header; /* header */
    short entry;		/* entry to struct table */
    short n;			/* number of member */
    long* id;			/* member id list */
    QHObj* member;		/* member */
    struct data_block* dbp;	/* Not used in current versoin */
} QHStruct;

typedef struct _QHList
{
    struct CELLObjHeader header; /* header */
    QHObj car;
    QHObj cdr;
} QHList;

typedef struct _QHExtData
{
    struct CELLObjHeader header; /* header */
    int type;
    void* data;
    void (* marker)(struct _QHExtData *);
    void (* destructor)(struct _QHExtData *);
} QHExtData;

typedef union _CELL
{
    struct CELLObjHeader header; /* header */
    Array a;			/* array */
    AssocArray assoc;		/* association array */
    QHString s;			/* string */
    QHStruct st;		/* struct */
    QHList list;		/* linked list */
    QHExtData ext;		/* external defined data */
} CELL;

#include "cell.h"

#endif /* ___QHOBJ_H_ */







