/* Copyright (c) 1995, Masanao Izumo */

/* qhparse.y */
%{

#define __QHPARSE_H_		/* Not to include "qhparse.h" */

#include <stdio.h>
#include <stdlib.h>
#include "common.h"
#include "stmt.h"
#include "localvar.h"
#include "lex_atomdat.h"
#include "keyobj.h" 
#include "optmz_tree.h"
#include "xmalloc.macros"

int current_scope;		/* SYM_GLOBAL or SYM_LOCAL. */

long current_module = None;	/* -1  : force to global
				 * None: no specification of module
				 * > 0 : module specified
				 */

static intStack farglist;	/* list of function argument's names  */

static intStack param_stack;	/* parameter stack
				 * used for counting the parameter length
				 */

static intStack temp_module;	/* temporary module name
				 * OK even if nested module in the futuer.
				 * if sizeof(int) == 2, then maybe don't work
				 * correctly.
				 */

static QuadList start_up_quadlist;
				/* start up codes */

static QuadList qglob_list;	/* global program codes */

void SAVE_MODULE(long module);	/* push the module name */
void LOAD_MODULE();		/* restore the module name */

#define INC_ISTACK param_stack.elem[param_stack.ptr - 1]++
				/* increment top value */
#define ADD_ISTACK(x) param_stack.elem[param_stack.ptr - 1] += (x)
				/* add x to top value */



#define IS_STMT_FUNC_DEF(t) ((t)->type == T_OP && \
			     (t)->left->type == T_FCALL && \
			     (t)->val == ASSIGN)
				/* Check whether it is statement function. */
%}

%union {
    int       ival;             /* integer value */
    QHTextPos tp;               /* text position */
    Tree*     t;                /* tree value */
    QuadList  q;                /* quad-list value */
    lex_atomdat atm;            /* atom represent */
};

/********** token definitions *****************
 WARNING:
	DON'T CHANGE THE DEFINITION ORDERS. IT IS USED IN INTERPRETING THE
    CODE ITSELF.
 **********************************************/

%token FIRST_INSTRUCTION	/* offset of instruction */
%token NOP			/* no operation */

/* use for `op' in quad-list */
/* binary operators */
%token <tp> ADD SUB MUL DIV REM 
%token <tp> ATAN2 RAND2 COMPLEX_OP
%token <tp> LOGOR LOGAND LLSHIFT LRSHIFT ALSHIFT ARSHIFT
%token <tp> POW EQ NQ LE LT GE GT

/* assign oprators */
%token <tp> AASSIGN SASSIGN MASSIGN DASSIGN RASSIGN ASSIGN

/* assign oprator means:
 * AASSIGN : +=		AASET : []+=
 * SASSIGN : -=		ASSET : []-=
 * MASSIGN : *=		AMSET : []*=
 * DASSIGN : /=		ADSET : []/=
 * RASSIGN : %=		ARSET : []%=
 * ASSIGN  : =		ASET  : []=, ASCSET: {}=
 */

/* unary operators */
%token SIN COS TAN ASIN ACOS ATAN SINH COSH TANH ASINH ACOSH ATANH
%token ABS FLOOR CEIL EXP LOG LOG10 LOG2 SQRT RAND0 RAND1 INT_OP REAL_OP IMAGE_OP
%token CONJUGATE ARG
%token <tp> UMINUS GETLINEOP COMMANDOP
%token <tp> ADDR INDIR LOGNOT NOT NOTNOT ZERO ONE STRUCTOF SIZEOF_OP
%token <tp> XINC INCX XDEC DECX

/* notnot(x) := !!(x)
             := x == 0 ? 0 : 1
  zero(X) := (X, 0)   ( evaluate X, and return zero )
  one(X)  := (X, 1)   ( evaluate X, and return one )

  Warnning: not(not(x)) != x, if x != 0 and x != 1

 optimization:
 not(notnot(x)) := not(x)
 notnot(not(x)) := not(x)

*/

%token LAST_OPERATOR

/* other oprators (use only in quad-list */
%token IPUSH			/* cast the argument to integer, and push to istack */
%token IPUSH2			/* 2 argument version of IPUSH */
%token IPUSH3			/* 3 argument version of IPUSH */
%token PARAM2			/* 2 argument version of PARAMETER */
%token PARAM3			/* 3 argument version of PARAMETER */
%token CALL			/* call function */
%token CALLEXT			/* call extern function */
%token <tp> FUNCALL		/* call function with specified its arguments */
%token ASCREF		/* reference an assoc array value: ascref x a key */
%token ASCSET		/* reference an assoc array value: ascset x a key */
%token AREF1			/* x = a[i]  ====>  aref1 x a i */
%token AREF			/* reference an array value */
%token AASET ASSET AMSET ADSET ARSET ASET /* array assignmant */
%token AXINC AINCX AXDEC ADECX	/* array inc/decrement */
%token AINDIR			/* *p = x */
%token PARAM			/* function argument */
%token IFGOTO IFNGOTO IFEQGOTO IFNQGOTO IFLEGOTO IFLTGOTO IFGEGOTO IFGTGOTO GOTO /* jump */
%token LIST_OP			/* new( ... ) => PARAM + [list res n -] */
%token STREF			/* reference struct value */
%token STSET			/* set struct value */
%token <tp> RETURN		/* return */
%token END			/* same as exit(0) */

/* built-in-functions or procedures */
%token WRITE			/* write arguments */
%token <tp> NEW			/* allocate a new array */
%token <tp> ASCNEW		/* new assoc array */
%token <tp> STRNEW		/* new string object */
%token <tp> STTNEW		/* new struct object */
%token TYPEOF			/* query a data type */
%token ATYPEOF			/* query an array type */
%token DIMENSION		/* query an array dimension */
%token ARRAYSIZE		/* number of array elements */
%token EXIT			/* exit program, and return its arguments to OS */
%token ARGV			/* lookup program argumets */
%token SYSTEM			/* system(" ... ") */
%token SLEEP			/* sleep(secones) */
%token RAND			/* random generate function */
%token SRAND			/* random seed function */
%token TIME			/* get current time */
%token CLOCK			/* get user cpu time */
%token EXPAND_CELL_SPACE	/* allocate more cell memory. */
%token CLEAR			/* zero clear object */
%token CLONE			/* generate a new clone object */
%token EQUAL			/* compare whether equal the two object */

/* string functions */
%token STRCPY STRNCPY STRCAT STRCMP STRNCMP STRLEN STRDUP CHOP
%token BSWRITE SWRITE ESWRITE
%token ATOI ATOR ATOC

/* graphics functions */
%token GOPEN GCLOSE GWINDOW GSCREEN GCOPY GCOPY_AREA GBACKING_STORE GSYNC
%token GCLEAR GTITLE_NAME GICON_NAME GFLUSH GTEXT_WIDTH GTEXT_HEIGHT
%token GPOINT GBELL
%token GLINE GLINE2 GPSET GBOX GFBOX GPOLY GFPOLY GCIRCLE GFCIRCLE GARC GFARC
%token GSTRING GSPLINE GFSPLINE
%token GFONT GJFONT GLOOKUP_COLOR GSETDASH
%token GCOLOR_NAME GCOLOR_RGB GCOLOR_FG GPEN GPEN3
%token GBUTTON GKEY GPAUSE GPAUSE1 GPOINTER
%token GPIXEL_PER_MMX GPIXEL_PER_MMY GDISPLAY_WIDTH GDISPLAY_HEIGHT
%token GRESIZE_SCREEN GSELECT_INPUT GVISUALIZE_SCREEN

/* file operations */
%token XFOPEN XFCLOSE XFWRITE2 XFWRITE3 XFREAD XFGETC XFPUTC XFTRUNC XFSEEK
%token XPOPEN XFCREATE XFDELETE

/* list operations */
%token CAR CDR CONS RPLACA RPLACD

%token LAST_INSTRUCTION

/* only use in lexing or parsing */
%token IDENTIFIER		/* identifier */
%token INTEGER			/* integer */
%token REAL			/* real */
%token IMNUMBER			/* imaginary number */
%token STR			/* string */
%token FUNCTION			/* function fname(...) { ... } */
%token DOTDOTDOT		/* ... */
%token GLOBAL			/* global variable, ... or global::name */
%token MODULE			/* module name { ... } */
%token MODULE_SPEC		/* :: */
%token LOCAL			/* local (v, ...) { ... } */
%token LOADFN			/* loadfn fname/narg; */
%token QH_EXTERN		/* extern fname; */
%token IF ELSE FOR DO WHILE BREAK CONTINUE LABEL
%token <tp> INC_OP DEC_OP	/* ++, -- */
%token STRING_OP		/* string */
%token ANY_SPEC			/* string-specified */
%token ASSOC_SPEC		/* association-array-specified */
%token WRITELN			/* write line */
%token <tp> OR AND		/* ||, && */
%token SWITCH CASE DEFAULT	/* switch-case statement */
%token PRAGMA			/* compiling option */
%token <ival> MATHFUNC1		/* math function operator */
%token <ival> MATHFUNC2		/* math function operator */
%token RAND			/* rand operator */
%token QHKEYWORD		/* keyword */
%token STRUCT			/* struct tag { ... } */
%token ENUM			/* enum [tag] { ... } */
%token VOLATILE			/* volatile function ... */

%nonassoc	LOWER_THEN_ELSE
%nonassoc	ELSE
%left  ','
%right '=' AASSIGN SASSIGN MASSIGN DASSIGN RASSIGN
%nonassoc	LOWER_THEN_COND_OPERATOR
%right '?' ':'
%left OR
%left AND
%left '|'
%left '&'
%left EQ NQ LE '<' GE '>'
%nonassoc GETLINEOP '`'
%left LLSHIFT LRSHIFT ALSHIFT ARSHIFT
%left  '+' '-'
%left  '*' '/' '%'
%right UMINUS '!' '~'
%right '^'
%nonassoc INC_OP DEC_OP
%right ADDR INDIR SIZEOF_OP STRUCTOF
%left <tp> '[' ']' '{' '}'
%right <tp> '.'

%type <atm> IDENTIFIER
%type <atm> INTEGER
%type <atm> REAL
%type <atm> IMNUMBER
%type <atm> STR


%type <tp> ',' '=' '<' '>' '+' '-' '*' '/' '%' '!' '^' '(' '?' '&' '|' '~' '`'
%type <tp> ':'

%type <ival> type_spec
%type <ival> module_spec
%type <ival> defun_header

%type <t> nc_expr_node
%type <t> expr_node
%type <t> expr_node_opt
%type <t> const_expr_node
%type <t> atom_node
%type <t> identifier_node
%type <t> identifier_node_opt
%type <t> function_node
%type <t> parameter_node_r
%type <t> parameter_node
%type <t> new_node
%type <t> rand_node
%type <t> struct_member_list
%type <t> struct_member_define
%type <t> struct_declared_list
%type <t> struct_declared_list_opt
%type <t> defstruct_node

%type <q> expr stmt compound_stmt
%type <q> stmt_list
%type <q> label_stmt
%type <q> goto_stmt
%type <q> return_stmt
%type <q> while_stmt
%type <q> if_stmt
%type <q> for_stmt
%type <q> break_stmt
%type <q> switch_stmt
%type <q> case_label
%type <q> default_label
%type <q> continue_stmt
%type <q> do_while_stmt
%type <q> writeln_stmt
%type <q> write_stmt
%type <q> swrite_stmt
%type <q> local_stmt
%type <q> local_bind1
%type <q> local_bind_list
%type <q> global_stmt_list
%type <q> no_module_global_list
%type <q> no_module_global
%type <q> global_declare
%type <q> global_declare1
%type <q> global_declare_list
%type <q> defstruct
%start prog

%%

prog
    : global_stmt_list
	{
	    qglob_list = append_quadlist(&qglob_list, &$1);
	}
    ;


global_stmt_list
    : /* empty */
	{
	    init_quad_list(&$$);
	}
    | global_stmt_list no_module_global
	{
	    $$ = append_quadlist(&$1, &$2);
	}
    | global_stmt_list MODULE IDENTIFIER '{'
	{
	    SAVE_MODULE($3.name);
	    push_new_binding_list();
	}
      no_module_global_list rb
	{
	    pop_new_binding_list();
	    LOAD_MODULE();	/* must be current_module == None */
	    $$ = append_quadlist(&$1, &$6);
	}
    | global_stmt_list QH_EXTERN IDENTIFIER sc
	{
	    declare_extern_function($3.name);
	    $$ = $1;
	}
    ;

no_module_global
    : stmt	  { $$ = $1; }
    | defun       { init_quad_list(&$$); }
    | loadfn_stmt { init_quad_list(&$$); }
    | global_declare { $$ = $1; }
    ;

global_declare1
    : identifier_node
	{
	    declare_global_variable($1->val, $1->module);
	    reuse_tree($1);
	    init_quad_list(&$$);
	}
    | identifier_node '=' nc_expr_node
	{
	    Tree* t;
	    Symbol* sym;

	    declare_global_variable($1->val, $1->module);

	    t = gen_op_node(ASSIGN, $1, $3, &$2);
	    sym = tree_quadlist(t, &$$);
	    reuse_tree(t);
	    reuse_if_temp_sym(sym);
	}
    ;

global_declare_list
    : global_declare1
	{
	    $$ = $1;
	}
    | global_declare_list ',' global_declare1
	{
	    $$ = append_quadlist(&$1, &$3);
	}
    ;

global_declare
    : GLOBAL global_declare_list sc { $$ = $2;              }
    | GLOBAL sc                     { init_quad_list(&$$); }

no_module_global_list
    : /* empty */ { init_quad_list(&$$); }
    | no_module_global_list no_module_global
	{
	    $$ = append_quadlist(&$1, &$2);
	}
    ;

module_spec
    :            MODULE_SPEC { $$ = -1; }
    | IDENTIFIER MODULE_SPEC { $$ = $1.name; }
    ;

loadfn_head
    : LOADFN
	{
	    if(Wall_flag)
		write_warning("Old format is found");
	}
    ;

loadfn_stmt
    : loadfn_head IDENTIFIER '/' INTEGER sc
	{
	    must_load_function($2.name, current_module, atoi(STRING($4.name)));
	}
    | loadfn_head module_spec IDENTIFIER '/' INTEGER sc
	{
	    if(Wall_flag)
		write_warning("Old format is found");
	    must_load_function($3.name, $2, atoi(STRING($5.name)));
	}
    | loadfn_head IDENTIFIER '/' DOTDOTDOT sc
	{
	    must_load_function($2.name, current_module, -1);
	}
    | loadfn_head module_spec IDENTIFIER '/' DOTDOTDOT sc
	{
	    must_load_function($3.name, $2, -1);
	}
    ;

defun_header
    : FUNCTION
	{
	    pushintStack(&param_stack, 0);
	    emptyintStack(&farglist);
	    current_scope = SYM_LOCAL;
	    $$ = False;
	}
    | VOLATILE FUNCTION
	{
	    pushintStack(&param_stack, 0);
	    emptyintStack(&farglist);
	    current_scope = SYM_LOCAL;
	    $$ = True;
	}
    ;

defun
    : defun_header identifier_node '(' arglist rp compound_stmt
	{
	    int narg = popintStack(&param_stack);

	    if($1)
		must_load_function($2->val, $2->module, narg);
	    compile_function($2->val, $2->module, narg, &$6);
	    clear_local_symbol_table();
	    clear_local_label_symbol_table();
	    clear_local_stmt_func_symbol();
	    clear_local_struct_symbol_table();
	    clear_local_enum_table();
	    current_scope = SYM_GLOBAL;
	    reuse_tree($2);
	}
    | defun_header identifier_node '(' DOTDOTDOT rp
	{
	    combine_vla_special_arguments(current_module);
	}
      compound_stmt
	{
	    popintStack(&param_stack);
	    if($1)
		must_load_function($2->val, $2->module, -1);

	    compile_function($2->val, $2->module, -1, &$7);

	    clear_local_symbol_table();
	    clear_local_label_symbol_table();
	    clear_local_stmt_func_symbol();
	    clear_local_struct_symbol_table();
	    clear_local_enum_table();
	    current_scope = SYM_GLOBAL;
	    reuse_tree($2);
	}
    | defun_header identifier_node '(' arglist rp sc
	{
	    int narg = popintStack(&param_stack);

	    if($1)
		must_load_function($2->val, $2->module, narg);
	    install_function($2->val, $2->module, narg);
	    current_scope = SYM_GLOBAL;
	    reuse_tree($2);
	    clear_local_symbol_table();
	}
    ;

arglist
    : /* empty */
    | IDENTIFIER
	{
	    Symbol* sym;

	    INC_ISTACK;
	    sym = install_local_symbol($1.name, current_module);
	    sym->data.align = topintStack(&param_stack);
	    sym->initialized_flag = True;
	    pushintStack(&farglist, $1.name);
	}
    | IDENTIFIER ',' arglist
	{
	    Symbol* sym;

	    if(check_redeclaration_farg(&farglist, $1.name) == True)
	    {
		INC_ISTACK;
		sym = install_local_symbol($1.name, current_module);
		sym->data.align = topintStack(&param_stack);
		sym->initialized_flag = True;
		pushintStack(&farglist, $1.name);
	    }
	}
    | error { }
    ;

compound_stmt
    : '{'
	{
	    push_new_binding_list();
	}
      stmt_list rb
	{
	    pop_new_binding_list();
	    $$ = $3;
	}
    ;

stmt_list
    : /* empty */    { init_quad_list(&$$); }
    | stmt_list stmt { $$ = append_quadlist(&$1, &$2); }
    ;

stmt
    : compound_stmt { $$ = $1; }
    | sc            { init_quad_list(&$$); }
    | expr sc       { $$ = $1; }
    | label_stmt    { $$ = $1; }
    | goto_stmt     { $$ = $1; }
    | return_stmt   { $$ = $1; }
    | while_stmt    { $$ = $1; }
    | if_stmt       { $$ = $1; }
    | for_stmt      { $$ = $1; }
    | break_stmt    { $$ = $1; }
    | continue_stmt { $$ = $1; }
    | do_while_stmt { $$ = $1; }
    | writeln_stmt  { $$ = $1; }
    | write_stmt    { $$ = $1; }
    | swrite_stmt   { $$ = $1; }
    | local_stmt    { $$ = $1; }
    | pragma_stmt   { init_quad_list(&$$); }
    | defstruct     { $$ = $1; }
    | case_label    { $$ = $1; }
    | default_label { $$ = $1; }
    | switch_stmt   { $$ = $1; }
    | error         { init_quad_list(&$$); }
    | enum_stmt     { init_quad_list(&$$); }
    ;

enum_tag_opt
    : /* empty */
    | IDENTIFIER { }
    ;


enum_list1
    : IDENTIFIER '=' const_expr_node
	{
	    define_enumerate($1.name, $3);
	}
    | IDENTIFIER
	{
	    define_enumerate($1.name, NULL);
	}
    ;

enum_list
    : /* empty */
    | enum_list1
    | enum_list1 ',' enum_list
    ;

enum_stmt
    : ENUM enum_tag_opt
	{
	    start_enumerate();
	}
    '{' enum_list rb sc
    ;

case_label
    : CASE const_expr_node ':'
	{
	    $$ = new_switch_case_label($2);
	    $$.last->tp = $3;
	}
    ;

default_label
    : DEFAULT ':'
	{
	    $$ = new_switch_default_label();
	    $$.last->tp = $2;
	}
    ;

switch_stmt
    : SWITCH '(' expr_node rp { switch_head($3); }
      stmt                    { $$ = gen_switch_stmt(&$6); }
    ;

struct_member_define
    : IDENTIFIER sc
	{
	    $$ = gen_atom_node(T_VAR, $1.name, &$1.tp);
	    INC_ISTACK;
	}
    | IDENTIFIER '=' nc_expr_node sc
	{
	    $$ = gen_node(T_OP, ASSIGN,
			  gen_atom_node(T_VAR, $1.name, &$1.tp),
			  $3, &$2);
	    INC_ISTACK;
	}
    | defstruct_node
	{
	    $$ = $1;
	    ADD_ISTACK($1->left->val);
	}
    | STRUCT identifier_node struct_declared_list sc
	{
	    int ndeclare = 0;
	    Tree* p = $3;

	    if($2 == NULL && p == NULL)
		write_warning("unnamed struct that defines no instances");

	    while(p != NULL)
	    {
		p = p->right;
		ndeclare++;
	    }

	    ADD_ISTACK(ndeclare);
	    $$ = gen_node(T_STRUCT,
			  -1,
			  gen_node(T_PARAM,
				   ndeclare,
				   $2,
				   $3,
				   NULL),
			  NULL,
			  NULL);
	}
    ;

struct_member_list
    : struct_member_define
	{
	    $$ = gen_node(T_PARAM, None, $1, NULL, &$1->tp);
	}
    | struct_member_define struct_member_list
	{
	    $$ = gen_node(T_PARAM, None, $1, $2, NULL);
	}
    | error { $$ = NULL; }
    ;

struct_declared_list
    : IDENTIFIER
	{
	    $$ = gen_node(T_PARAM,
			  None,
			  gen_atom_node(T_VAR, $1.name, &$1.tp),
			  NULL,
			  &$1.tp);
	}
    | IDENTIFIER ',' struct_declared_list
	{
	    $$ = gen_node(T_PARAM,
			  None,
			  gen_atom_node(T_VAR, $1.name, &$1.tp),
			  $3,
			  &$2);
	}
    ;

struct_declared_list_opt
    : /* empty */          { $$ = NULL; }
    | struct_declared_list { $$ = $1;   }
    ;

defstruct_node
    : STRUCT identifier_node_opt '{'
	{
	    pushintStack(&param_stack, 0);
	}
      struct_member_list '}' struct_declared_list_opt sc
	{
	    int ndeclare = 0;
	    Tree* p = $7;

	    if($2 == NULL && p == NULL)
		write_warning("unnamed struct that defines no instances");

	    while(p != NULL)
	    {
		p = p->right;
		ndeclare++;
	    }
	    $$ = gen_node(T_STRUCT,
			  popintStack(&param_stack),
			  gen_node(T_PARAM,
				   ndeclare,
				   $2,
				   $7,
				   NULL),
			  $5,
			  NULL);
	}
    ;

defstruct
    : defstruct_node
	{
	    int new_entry;
	    new_entry = define_struct($1);

	    if($1->left->right != NULL)
		translate_sttnew_list($1->left->right, new_entry, &$$);
/*

	    if($1->left->left == NULL)
		yyerror("Unnamed struct"); */
	    else
		init_quad_list(&$$);
	}
    ;

pragma_stmt
    : PRAGMA IDENTIFIER '(' STR rp
	{
	    pragma_option(STRING($2.name), STRING($4.name));
	}
    ;

local_bind1
    : IDENTIFIER
	{
	    bind_local_variable($1.name, current_module);
	    init_quad_list(&$$);
	}
    | IDENTIFIER '=' nc_expr_node
	{
	    Tree* t;
	    Symbol* sym;

	    bind_local_variable($1.name, current_module);

	    t = gen_op_node(ASSIGN,
			    gen_atom_node(T_VAR, $1.name, &$1.tp),
			    $3,
			    &$2);
	    sym = tree_quadlist(t, &$$);
	    reuse_tree(t);
	    reuse_if_temp_sym(sym);
	}
    ;

local_bind_list
    : /* empty */
	{
	    init_quad_list(&$$);
	}
    | local_bind1
	{
	    $$ = $1;
	}
    | local_bind1 ',' local_bind_list
	{
	    $$ = append_quadlist(&$1, &$3);
	}
    ;

local_stmt
    : LOCAL local_bind_list sc
	{
	    $$ = $2;
	}
    ;

swrite_stmt
    : SWRITE '('
	{
	    pushintStack(&param_stack, 0);
	}
      parameter_node rp sc
	{
	    $$ = gen_swrite_stmt($4, popintStack(&param_stack));
	}
    ;

type_spec
    : /* empty */       { $$ = ANY_SPEC; }
    | MATHFUNC1			/* INT_OP or REAL_OP */
	{
	    switch($1)
	    {
	      case INT_OP:
		$$ = INT_OP;
		break;
	      case REAL_OP:
		$$ = REAL_OP;
		break;
	      default:
		yyerror("Invalid type specified");
		$$ = INT_OP;
		break;
	    }
	}
    | MATHFUNC2			/* CONPLEX_OP only */
	{
	    if($1 != COMPLEX_OP)
		yyerror("Invalid type specified");
	    $$ = COMPLEX_OP;
	}
    ;

write_stmt
    : WRITE
	{
	    pushintStack(&param_stack, 0);
	}
      '(' parameter_node rp sc
	{
	    $$ = gen_write_stmt($4, popintStack(&param_stack), False);
	}
    ;


writeln_stmt
    : WRITELN
	{
	    pushintStack(&param_stack, 0);
	}
      '(' parameter_node rp sc
	{
	    $$ = gen_write_stmt($4, popintStack(&param_stack), True);
	}
    ;

do_while_stmt
    : DO
	{
	    gen_do_while_head_stmt();
	}
      stmt WHILE '(' expr_node rp sc
	{
	    $$ = gen_do_while_tail_stmt($6, &$3);
	}
    ;

break_stmt
    : BREAK sc
	{
	    $$ = gen_break_stmt();
	}
    ;

continue_stmt
    : CONTINUE sc
	{
	    $$ = gen_continue_stmt();
	}
    ;

for_stmt
    : FOR '(' expr_node_opt sc expr_node_opt sc expr_node_opt rp
	{
	    if(W_uninitialized_variable == True)
	    {
		set_initialized_flag_all($3);
		set_initialized_flag_all($5);
	    }
	    for_head_stmt();
	}
      stmt
	{
	    $$ = for_tail_stmt($3, $5, $7, &$10);
	}
    ;

if_stmt
    : IF '(' expr_node rp stmt				%prec LOWER_THEN_ELSE
	{
	    if(W_uninitialized_variable == True)
		set_initialized_flag_all($3);
	    $$ = gen_if_stmt($3, &$5);
	}
    | IF '(' expr_node rp stmt ELSE stmt
	{
	    if(W_uninitialized_variable == True)
		set_initialized_flag_all($3);
	    $$ = gen_if_else_stmt($3, &$5, &$7);
	}
    ;

while_stmt
    : WHILE '(' expr_node rp
	{
	    if(W_uninitialized_variable == True)
		set_initialized_flag_all($3);

	    while_head_stmt();
	}
    stmt
	{
	    $$ = while_tail_stmt($3, &$6);
	}
    ;

return_stmt
    : RETURN sc
	{
	    $$ = quad_node_list(gen_quad_node1(RETURN,
					       NULL,
					       NULL,
					       NULL));
	    $$.last->tp = $1;
	}
    | RETURN expr_node sc
	{
	    $$ = gen_return_stmt($2);
	}
    ;

label_stmt
    : IDENTIFIER ':'
	{
	    $$ = gen_label_stmt($1.name, current_module);
	    $$.last->tp = $1.tp;
	}
    | module_spec IDENTIFIER ':'
	{
	    $$ = gen_label_stmt($2.name, $1);
	    $$.last->tp = $2.tp;
	}
    ;

goto_stmt
    : GOTO IDENTIFIER sc
	{
	    $$ = gen_goto_stmt($2.name, current_module);
	    $$.last->tp = $2.tp;
	}
    | GOTO module_spec IDENTIFIER sc
	{
	    $$ = gen_goto_stmt($3.name, $2);
	    $$.last->tp = $3.tp;
	}

const_expr_node
    : nc_expr_node
	{
	    $$ = calc_constant_expr_tree($1);
	    if($$ == NULL)
		$$ = &null_tree_node;
	}
    ;

expr
    : expr_node
	{
	    Symbol* sym;
	    Tree* t = $1;

	    if(Wall_flag == True &&
	       t->val == EQ  &&
	       t->type == T_OP)
		write_warning("May be type-miss: `=' and `=='");

	    if(IS_STMT_FUNC_DEF(t))
	    {
		if(check_expr(t->right) == True)
		{
		    t->right = optimize_tree(t->right);
		    install_stmt_func_tree(t->left, t->right);
		}
		init_quad_list(&$$);
	    }
	    else
	    {
		sym = tree_quadlist(t, &$$);
		reuse_if_temp_sym(sym);
	    }
	    reuse_tree(t);
	}
    ;

nc_expr_node
    : atom_node     { $$ = $1; }
    | function_node { $$ = $1; }
    | new_node      { $$ = $1; }
    | rand_node	    { $$ = $1; }
    | nc_expr_node '='     nc_expr_node { $$ = gen_op_node( ASSIGN, $1, $3, &$2); }
    | nc_expr_node AASSIGN nc_expr_node { $$ = gen_op_node(AASSIGN, $1, $3, &$2); }
    | nc_expr_node SASSIGN nc_expr_node { $$ = gen_op_node(SASSIGN, $1, $3, &$2); }
    | nc_expr_node MASSIGN nc_expr_node { $$ = gen_op_node(MASSIGN, $1, $3, &$2); }
    | nc_expr_node DASSIGN nc_expr_node { $$ = gen_op_node(DASSIGN, $1, $3, &$2); }
    | nc_expr_node RASSIGN nc_expr_node { $$ = gen_op_node(RASSIGN, $1, $3, &$2); }
    | nc_expr_node '?' nc_expr_node ':' nc_expr_node
	{
	    $$ = gen_cond_op3($1, $3, $5, &$2);
	}
    | nc_expr_node OR  nc_expr_node { $$ = gen_op_node(OR , $1, $3, &$2); }
    | nc_expr_node AND nc_expr_node { $$ = gen_op_node(AND, $1, $3, &$2); }
    | nc_expr_node EQ  nc_expr_node { $$ = gen_op_node(EQ , $1, $3, &$2); }
    | nc_expr_node NQ  nc_expr_node { $$ = gen_op_node(NQ , $1, $3, &$2); }
    | nc_expr_node LE  nc_expr_node { $$ = gen_op_node(LE , $1, $3, &$2); }
    | nc_expr_node '<' nc_expr_node { $$ = gen_op_node(LT , $1, $3, &$2); }
    | nc_expr_node GE  nc_expr_node { $$ = gen_op_node(GE , $1, $3, &$2); }
    | nc_expr_node '>' nc_expr_node { $$ = gen_op_node(GT , $1, $3, &$2); }
    | nc_expr_node '+' nc_expr_node { $$ = gen_op_node(ADD, $1, $3, &$2); }
    | nc_expr_node '-' nc_expr_node { $$ = gen_op_node(SUB, $1, $3, &$2); }
    | nc_expr_node '*' nc_expr_node { $$ = gen_op_node(MUL, $1, $3, &$2); }
    | nc_expr_node '/' nc_expr_node { $$ = gen_op_node(DIV, $1, $3, &$2); }
    | nc_expr_node '%' nc_expr_node { $$ = gen_op_node(REM, $1, $3, &$2); }
    | nc_expr_node '|' nc_expr_node { $$ = gen_op_node(LOGOR, $1, $3, &$2); }
    | nc_expr_node '&' nc_expr_node { $$ = gen_op_node(LOGAND, $1, $3, &$2); }
    | nc_expr_node '.' identifier_node
	{
	    $$ = gen_op_node(STREF, $1, $3, &$2);
	}
    | '<' nc_expr_node '>' %prec GETLINEOP
	{
	    $$ = gen_uop_node(GETLINEOP, $2, &$3);
	}
    | '<' '>' %prec GETLINEOP
	{
	    $$ = gen_uop_node(GETLINEOP,
			      gen_atom_node(T_CINT, Q_zero_int, &$2),
			      &$2);
	}
    | '`' nc_expr_node '`'
	{ $$ = gen_uop_node(COMMANDOP, $2, &$3); }
    | nc_expr_node LLSHIFT nc_expr_node
	{ $$ = gen_op_node(LLSHIFT, $1, $3, &$2); }
    | nc_expr_node LRSHIFT nc_expr_node
	{ $$ = gen_op_node(LRSHIFT, $1, $3, &$2); }
    | nc_expr_node ALSHIFT nc_expr_node
	{ $$ = gen_op_node(ALSHIFT, $1, $3, &$2); }
    | nc_expr_node ARSHIFT nc_expr_node
	{ $$ = gen_op_node(ARSHIFT, $1, $3, &$2); }
    | MATHFUNC1 '(' nc_expr_node ')'
	{ $$ = gen_uop_node($1, $3, &$2);	  }
    | MATHFUNC2 '(' nc_expr_node ',' nc_expr_node rp
	{ $$ = gen_op_node($1, $3, $5, &$2);	  }
    | '-' nc_expr_node %prec UMINUS { $$ = gen_uop_node(UMINUS, $2, &$1); }
    | '~' nc_expr_node { $$ = gen_uop_node(NOT, $2, &$1); }
    | '!' nc_expr_node { $$ = gen_uop_node(NOT, $2, &$1); }
    | nc_expr_node '^' nc_expr_node { $$ = gen_op_node(POW, $1, $3, &$2); }
    | INC_OP nc_expr_node { $$ = gen_op_node(INCX, $2, NULL, &$1); }
    | DEC_OP nc_expr_node { $$ = gen_op_node(DECX, $2, NULL, &$1); }
    | nc_expr_node INC_OP { $$ = gen_op_node(XINC, $1, NULL, &$2);   }
    | nc_expr_node DEC_OP { $$ = gen_op_node(XDEC, $1, NULL, &$2);   }
    | '*' nc_expr_node %prec INDIR
	{ $$ = gen_uop_node(INDIR, $2, &$1); }
    | '&' nc_expr_node %prec ADDR
	{ $$ = gen_uop_node(ADDR, $2, &$1); }
    | nc_expr_node '['
	{
	    pushintStack(&param_stack, 0);
	}
      parameter_node_r rk
	{
	    $$ = gen_node(T_ARRAY, popintStack(&param_stack), $1, $4, &$2);
	}
    | nc_expr_node '{' nc_expr_node rb
	{
	    $$ = gen_node(T_ASSOCARRAY, None, $1, $3, &$2);
	}
    | QHKEYWORD '(' IDENTIFIER ')'
	{
	    $$ = gen_keyword_node($3.name);
	}
    | NEW
	{
	    pushintStack(&param_stack, 0);
	}
      '(' parameter_node rp
	{
	    $$ = gen_node(T_LIST, popintStack(&param_stack), $4, NULL, &$3);
	}
    | SIZEOF_OP nc_expr_node
	{
	    $$ = gen_uop_node(SIZEOF_OP, $2, &$1);
	}
    | SIZEOF_OP '(' type_spec ')'
	{
	    $$ = gen_sizeof_type_spec($3, &$1);
	}
    | SIZEOF_OP '(' STRING_OP ')'
	{
	    $$ = gen_sizeof_type_spec(STRING_OP, &$1);
	}
    | SIZEOF_OP '(' ASSOC_SPEC ')'
	{
	    $$ = gen_sizeof_type_spec(ASSOC_SPEC, &$1);
	}
    | STRUCTOF nc_expr_node
	{
	    $$ = gen_uop_node(STRUCTOF, $2, &$1);
	}
    | STRUCTOF '(' STRUCT identifier_node ')'
	{
	    $$ = gen_structof_node($4, &$1);
	}
    | '(' expr_node rp { $$ = $2; }
    ;

rand_node
    : RAND '(' rp
	{ $$ = gen_uop_node(RAND0, gen_atom_node(T_CINT, Q_zero_int, &$2), &$2); }
    | RAND '(' nc_expr_node ')'
	{ $$ = gen_uop_node(RAND1, $3, &$2); }
    | RAND '(' nc_expr_node ',' nc_expr_node rp
	{ $$ = gen_op_node(RAND2, $3, $5, &$2); }
    ;

new_node
    : NEW type_spec '['
	{
	    pushintStack(&param_stack, 0);
	}
      parameter_node_r rk
	{
	    int dim = popintStack(&param_stack);

	    if(dim == 0)
	    {
		yyerror("array size missing");
		$$ = NULL;
	    }
	    else
	    {
		$5->val = dim;
		$$ = gen_node(T_NEW, $2, $5, NULL, &$1);
	    }
	}
    | NEW STRING_OP
	{
	    $$ = gen_node(T_STRNEW, None, NULL, NULL, &$1);
	}
    | NEW STRING_OP '(' expr_node ')'
	{
	    $$ = gen_node(T_STRNEW, None, $4, NULL, &$1);
	}
    | NEW ASSOC_SPEC
	{
	    $$ = gen_node(T_ASCNEW, None, NULL, NULL, &$1);
	}
    | NEW identifier_node
	{
	    $$ = gen_node(T_STTNEW, None, $2, NULL, &$1);
	}
    ;

expr_node
    : nc_expr_node            { $$ = $1;                     }
    | expr_node ',' expr_node { $$ = gen_comma_node($1, $3, &$2); }
    ;

expr_node_opt
    : /* empty */ { $$ = NULL; }
    | expr_node   { $$ = $1; }
    ;

atom_node
    : identifier_node { $$ = $1; }
    | INTEGER    { $$ = gen_atom_node(T_CINT,  $1.name, &$1.tp); }
    | REAL       { $$ = gen_atom_node(T_CREAL, $1.name, &$1.tp); }
    | STR        { $$ = gen_atom_node(T_CSTR,  $1.name, &$1.tp); }
    | IMNUMBER
	{
	    $$ = gen_op_node(COMPLEX_OP,
			     gen_atom_node(T_CINT, Q_zero_int, &$1.tp),
			     gen_atom_node(T_CREAL, $1.name, &$1.tp),
			     &$1.tp);
	}
    ;

identifier_node
    : IDENTIFIER
	{
	    $$ = gen_atom_node(T_VAR, $1.name, &$1.tp);
	}
    | module_spec IDENTIFIER
	{
	    $$ = gen_atom_node(T_VAR, $2.name, &$2.tp);
	    $$->module = $1;
	    $$->module_specified = True;
	}
    ;

identifier_node_opt
    : /* empty */     { $$ = NULL; }
    | identifier_node { $$ = $1;   }
    ;


function_node
    : identifier_node '('
	{
	    pushintStack(&param_stack, 0);
	}
      parameter_node rp
	{
	    $$ = gen_node(T_FCALL, popintStack(&param_stack), $1, $4, &$2);
	}
    | FUNCALL '(' nc_expr_node rp
	{
	    $$ = gen_node(T_FUNCALL, 0, $3, NULL, &$1);
	}
    | FUNCALL '(' rp
	{
	    yyerror2("Too few arguments `funcall'", &$1);
	    $$ = NULL;
	}
    | FUNCALL '(' nc_expr_node ','
	{
	    pushintStack(&param_stack, 0);
	}
      parameter_node rp
	{
	    $$ = gen_node(T_FUNCALL, popintStack(&param_stack), $3, $6, &$1);
	}
    ;

parameter_node_r
    : { $$ = NULL; }
    | nc_expr_node
	{
	    $$ = gen_node(T_PARAM, None, $1, NULL, &$1->tp);
	    INC_ISTACK;
	}
    | parameter_node_r ',' nc_expr_node 
	{
	    $$ = gen_node(T_PARAM, None, $3, $1, &$2);
	    INC_ISTACK;
	}
    ;

parameter_node
    : { $$ = NULL; }
    | nc_expr_node
	{
	    $$ = gen_node(T_PARAM, None, $1, NULL, &$1->tp);
	    INC_ISTACK;
	}
    | nc_expr_node ',' parameter_node
	{
	    $$ = gen_node(T_PARAM, None, $1, $3, &$2);
	    INC_ISTACK;
	}
    ;

rp
    : ')' { yyerrok; }
    ;
sc
    : ';' { yyerrok; }
    ;

rb
    : '}' { yyerrok; }
    ;

rk
    : ']'  { yyerrok; }
    ;

%%

void init_qhparse(void)
{
    makeintStack(&param_stack, 256);
    makeintStack(&farglist,    16);
    makeintStack(&temp_module, 16);
    current_scope = SYM_GLOBAL;
    start_up_quadlist = start_up_routine();
    init_quad_list(&qglob_list);
    push_new_binding_list();
}

void compile_all(void)
{
    qglob_list = append_quadlist(&start_up_quadlist, &qglob_list);

    qglob_list = compile_global_code(&qglob_list);

    load_qhprog(&qglob_list);

    pop_new_binding_list();
    deleteintStack(&param_stack);
    deleteintStack(&farglist);

    reuse_quad_list(&qglob_list);
}

void SAVE_MODULE(long x)
{
    pushintStack(&temp_module, current_module);
    current_module = x;
}

void LOAD_MODULE()
{
    current_module = popintStack(&temp_module);
}
