/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "common.h"
#include "interp.h"
#include "functab.h"
#include "SWRBuff.defs"
#include "SWRBuffStack.h"
#if defined(sgi)
#include <strings.h>
#endif

#define CHECK_STRING(obj) if(!IS_QHSTRING(obj)) execute_error("String expected");

static SWRBuffStack swrbuff;

void qh_bswrite(QHProg* p)
{
    SWRBuff buff;

    if(swrbuff.elem == NULL)
	makeSWRBuffStack(&swrbuff, 16);	/* first call once */

    buff.swrite_buffer = &memstack[MEMADDR(p->a1)];
    if(!IS_QHSTRING(buff.swrite_buffer))
	execute_error("swrite: String expected");
    if(buff.swrite_buffer->data.s->isconst)
	execute_error("swrite: Attempted modify the constant string");
    buff.swrite_buffer_ptr = 0;
    pushSWRBuffStack(&swrbuff, buff);
}

void qh_swrite(QHProg* p)
{
    SWRBuff buff;
    int len;
    int maxlen;

    buff = popSWRBuffStack(&swrbuff);
    if(!IS_QHSTRING(buff.swrite_buffer))
	execute_error("swrite: String expected");
    if(buff.swrite_buffer->data.s->isconst)
	execute_error("swrite: Attempted modify the constant string");

    maxlen = buff.swrite_buffer->data.s->maxlen;
    len = strlen(obj_string(temp_buff, &memstack[MEMADDR(p->a1)]));

    if(len == 0)
    {
	pushSWRBuffStack(&swrbuff, buff);
	return;
    }

    touch_string(buff.swrite_buffer->data.s, buff.swrite_buffer_ptr + len);
    strncpy(buff.swrite_buffer->data.s->str + buff.swrite_buffer_ptr,
	    temp_buff, len);
    buff.swrite_buffer_ptr += len;

    pushSWRBuffStack(&swrbuff, buff);
}

void qh_eswrite(QHProg* p)
{
    SWRBuff buff;

    buff = popSWRBuffStack(&swrbuff);
    if(!IS_QHSTRING(buff.swrite_buffer))
	execute_error("swrite: String expected");

    buff.swrite_buffer->data.s->str[buff.swrite_buffer_ptr] = '\0';
}

static void qh_internal_strncopy(QHString* s1, QHString* s2, int n)
{
    if(s1->isconst)
	execute_error("Attempted modification of constant string");

    if(n == 0)
    {
	if(s1->str == NULL)
	    return;
	s1->str[0] = '\0';
	return;
    }

    touch_string(s1, n);
    bcopy(s2->str, s1->str, n);
}

void qh_strcpy(QHObj* ret)
{
    int n;
    QHObj* x1 = STACK(1);
    QHObj* x2 = STACK(2);

    CHECK_STRING(x1);
    CHECK_STRING(x2);

    n = (x2->data.s->str ? strlen(x2->data.s->str) : 0);
    qh_internal_strncopy(x1->data.s, x2->data.s, n);
    x1->data.s->str[n] = '\0';
    *ret = *x1;
}

void qh_strncpy(QHObj* ret)
{
    QHObj* x1 = STACK(1);
    QHObj* x2 = STACK(2);
    int n = cast_to_int(STACK(3));

    CHECK_STRING(x1);
    CHECK_STRING(x2);

    touch_string(x2->data.s, n);
    qh_internal_strncopy(x1->data.s, x2->data.s, n);
    *ret = *x1;
}

void qh_strcat(QHObj* ret)
{
    QHObj* x1 = STACK(1);
    QHObj* x2 = STACK(2);
    int n1, n2;

    CHECK_STRING(x1);
    CHECK_STRING(x2);

    if(x2->data.s->str == NULL || x2->data.s->str[0] == '\0')
    {
	*ret = *x1;
	return;
    }

    if(x1->data.s->isconst)
	execute_error("Attempted modification of constant string");

    n2 = strlen(x2->data.s->str);

    if(x1->data.s->str == NULL || x1->data.s->str[0] == '\0')
    {
	qh_internal_strncopy(x1->data.s, x2->data.s, n2);
	*ret = *x1;
	return;
    }

    n1 = strlen(x1->data.s->str);
    touch_string(x1->data.s, n1 + n2);
    bcopy(x2->data.s->str, x1->data.s->str + n1, n2);
    *ret = *x1;
}

void qh_strcmp(QHObj* ret)
{
    QHObj* x1 = STACK(1);
    QHObj* x2 = STACK(2);

    CHECK_STRING(x1);
    CHECK_STRING(x2);

    ret->type = QH_Int;
    ret->data.i = strcmp(x1->data.s->str, x2->data.s->str);
}

void qh_strncmp(QHObj* ret)
{
    QHObj* x1 = STACK(1);
    QHObj* x2 = STACK(2);
    int n = cast_to_int(STACK(3));

    CHECK_STRING(x1);
    CHECK_STRING(x2);

    ret->type = QH_Int;
    ret->data.i = strncmp(x1->data.s->str, x2->data.s->str, n);
}

void qh_strlen(QHObj* ret)
{
    QHObj* x = STACK(1);

    CHECK_STRING(x);

    ret->type = QH_Int;
    ret->data.i = strlen(x->data.s->str);
}

QHString* clone_string(const QHString *x)
{
    QHString* clone;

    clone = alloc_string();

    clone->isconst = False;
    clone->fixed = x->fixed;

    if(x->str != NULL)
    {
	int len = strlen(x->str);

	if(DATA_BLOCK_AVAILABLE(len + 1))
	{
	    clone->str = data_block_ptr->data;
	    clone->dbp = data_block_ptr;
	    data_block_ptr = data_block_ptr->next;
	    bcopy(x->str, clone->str, len);
	    bzero(clone->str + len, DATA_BLOCK_SIZE - len);
	}
	else
	    clone->str = xstrdup(x->str);
	clone->maxlen = len;
    }
    else
    {
	clone->str = NULL;
	clone->maxlen = 0;
    }
    clone->header.type = QH_String;

    return clone;
}

void qh_strdup(QHObj* ret)
{
    QHObj* x = STACK(1);

    CHECK_STRING(x);
    ret->data.s = clone_string(x->data.s);
    ret->type = QH_String;
}

void qh_atoi(QHObj* ret)
{
    char* s;

    s = cast_to_string(STACK(1));

    while(*s != '\0'   &&
	  *s != '-'    &&
	  !isdigit(*s))
	s++;

    ret->type = QH_Int;
    ret->data.i = atoi(s);
}

void qh_ator(QHObj* ret)
{
    char* s;

    s = cast_to_string(STACK(1));

    while(*s != '\0'   &&
	  *s != '-'    &&
	  !isdigit(*s))
	s++;

    ret->type = QH_Real;
    ret->data.r = atof(s);
}

void qh_atoc(QHObj* ret)
{
    char* s;
    double re, im;

    s = cast_to_string(STACK(1));

    while(*s != '\0'   &&
	  *s != '-'    &&
	  !isdigit(*s))
	s++;
    re = atof(s);

    while(*s != '\0'   &&
	  (isdigit(*s) ||
	   *s == '+'   ||
	   *s == '-'   ||
	   *s == 'e'   ||
	   *s == 'E'   ||
	   *s == '.'))
	s++;

    while(*s != '\0'   &&
	  *s != '-'    &&
	  !isdigit(*s))
	s++;
    im = atof(s);


    ret->type = QH_Complex;
    ret->data.c.re = re;
    ret->data.c.im = im;
}

void qh_chop(QHObj* ret)
{
    QHObj* x = STACK(1);
    char* s;

    ret->type = QH_Int;
    ret->data.i = 0;

    if(!IS_QHSTRING(x))
	return;

    if(x->data.s->isconst)
	execute_error("Attempted modification of constant string");

    s = x->data.s->str;
    if(s == NULL || *s == '\0')
	return;

    while(*++s)
	;

    ret->data.i = *(s - 1);
    *(s - 1) = '\0';
}

QHString* alloc_string(void)
{
    QHString* s;

    s = (QHString *)alloc_cell();
    s->maxlen  = 0;
    s->isconst = False;
    s->str     = NULL;
    s->fixed   = False;
    s->dbp     = NULL;

    return s;
}

QHString* alloc_constant_string(const char* str)
{
    QHString* s;

    s = alloc_string();
    string_expand(temp_buff, str);

    s->maxlen      = strlen(temp_buff);
    s->str         = (unsigned char *)xstrdup(temp_buff);
    s->isconst     = True;
    s->fixed       = True;
    s->header.type = QH_String;

    return s;
}

void qh_strnew(QHProg* p)
{
    QHObj* res = memstack + MEMADDR(p->res);
    int fixed_len = cast_to_int(memstack + MEMADDR(p->a1));
    char* s = NULL;
    Bool use_data_block = False;

    if(fixed_len > 0)
    {
	if(DATA_BLOCK_AVAILABLE(fixed_len + 1))
	{
	    s = data_block_ptr->data;
	    use_data_block = True;
	}
	else
	    s = (unsigned char *)xmalloc(fixed_len + 1);
	bzero(s, fixed_len + 1);
    }

    res->data.s = alloc_string();
    res->data.s->header.type = QH_String;
    res->type = QH_String;
    if(fixed_len > 0)
    {
	res->data.s->maxlen = fixed_len;
	res->data.s->str = s;
	res->data.s->fixed = True;
	if(use_data_block == True)
	{
	    res->data.s->dbp = data_block_ptr;
	    data_block_ptr = data_block_ptr->next;
	}
    }
}

Bool equal_string(const QHString* s1, const QHString* s2)
{
    if(s1->str == NULL && s2->str == NULL)
	return True;
    if(s1->str == NULL || s2->str == NULL)
	return False;

    return !strcmp(s1->str, s2->str);
}

Bool touch_string(QHString* s, int align)
{
    if(align < 0)
	execute_error("String is accessed negative region");

    if(s->str == NULL)
    {
	if(DATA_BLOCK_AVAILABLE(align + 1))
	{
	    s->maxlen = DATA_BLOCK_SIZE - 1;
	    s->str = data_block_ptr->data;
	    bzero(s->str, DATA_BLOCK_SIZE);
	    s->dbp = data_block_ptr;
	    data_block_ptr = data_block_ptr->next;
	}
	else
	{
	    s->maxlen = (align + 1) * 2;
	    s->str = (unsigned char *)xmalloc(s->maxlen + 1); /* +1 for '\0' */
	    bzero(s->str, s->maxlen + 1);
	}
	return True;
    }

    if(s->maxlen <= align)
    {
	int new_size;

	if(s->fixed)
	    execute_error("Fixed string is accessed out of the range");

	if(DATA_BLOCK_AVAILABLE(align + 1))
	{
	    /* s->maxlen <= align+1 < DATA_BLOCK_SIZE */
	    new_size = DATA_BLOCK_SIZE - 1;
	    bcopy(s->str, data_block_ptr->data, s->maxlen);
	    bzero(data_block_ptr->data + s->maxlen, DATA_BLOCK_SIZE - s->maxlen);
	    s->str = data_block_ptr->data;
	    s->dbp = data_block_ptr;
	    data_block_ptr = data_block_ptr->next;
	}
	else
	{
	    new_size = (align + 1) * 2;

	    if(s->dbp != NULL)
	    {
		char* new_ptr;

		new_ptr = xmalloc(new_size + 1); /* +1 for '\0' */
		bcopy(s->str, new_ptr, s->maxlen);
		bzero(new_ptr + s->maxlen, new_size + 1 - s->maxlen);

		reuse_data_block(s->dbp);
		s->dbp = NULL;
	    }
	    else
	    {
		s->str = xrealloc(s->str, new_size + 1);
		bzero(s->str + s->maxlen, new_size + 1 - s->maxlen);
	    }
	}
	s->maxlen = new_size;
	return True;
    }
    return False;
}
