#include <stdio.h>
#include "common.h"
#include "interp.h"

void qh_sttnew(QHProg* p)
{
    QHStruct* st;
    QHObj* x = memstack + MEMADDR(p->res);
    int i, n;

    st = (QHStruct *)alloc_cell();
    st->entry = p->a1;
    n = struct_tag_table[p->a1].nmember;
    st->member = (QHObj *)xmalloc(n * sizeof(QHObj));
    st->n = n;
    st->id = struct_tag_table[p->a1].member_id;
    st->dbp = NULL;

    for(i = 0; i < n; i++)
    {
	st->member[i].type = QH_Int;
	st->member[i].data.i = 0;
    }
    x->type = QH_Struct;
    x->data.st = st;
    st->header.type = QH_Struct;
}

static inline int search_struct_member(QHStruct* x, int id)
{
    long* idlist;
    long* orig_idlist;
    long* p;

    orig_idlist = idlist = x->id;
    p = idlist + x->n;

    while(idlist < p - 1)
    {
	if(*idlist == id) return idlist - orig_idlist;
	idlist++;
	if(*idlist == id) return idlist - orig_idlist;
	idlist++;
    }

    if(idlist < p)
	if(*idlist == id) return idlist - orig_idlist;

    execute_error("Structure has no member");
    return 0;
}

void qh_stset(QHProg* p)
{
    int k = 0;
    QHObj* x = memstack + MEMADDR(p->res);
    QHObj* y = memstack + MEMADDR(p->a2);
    QHStruct* st;

    if(x->type != QH_Struct)
	execute_error("Struct expected");

    st = x->data.st;
    k = search_struct_member(st, p->a1);

    if(struct_tag_table[st->entry].next[k] != NULL)
    {
	if(y->type != QH_Struct)
	    execute_error("Struct type mismatch");
	if((int)(struct_tag_table[st->entry].next[k] - struct_tag_table)
	   != y->data.st->entry)
	    execute_error("Struct Type mismatch");
    }

    st->member[k] = *y;
}

void qh_stref(QHProg* p)
{
    int k = 0;
    QHObj* res = memstack + MEMADDR(p->res);
    QHObj* x = memstack + MEMADDR(p->a1);
    QHStruct* st;

    if(x->type != QH_Struct)
	execute_error("Struct expected");

    st = x->data.st;

    k = search_struct_member(x->data.st, p->a2);
    *res = st->member[k];
}

QHStruct* clone_struct(const QHStruct* x)
{
    QHStruct* clone;
    int i, n;

    clone = (QHStruct *)alloc_cell();
    clone->entry = x->entry;
    n = clone->n = x->n;
    clone->member = (QHObj *)xmalloc(n * sizeof(QHObj));
    clone->id = x->id;

    for(i = 0; i < n; i++)
	qh_copy_clone_object(clone->member + i, x->member + i);
    clone->header.type = QH_Struct;

    return clone;
}

Bool equal_struct(const QHStruct* st1, const QHStruct* st2)
{
    int i;

    if(st1->entry != st2->entry)
	return False;

    for(i = 0; i < st1->n; i++)
	if(!qh_equal_object(st1->member + i, st2->member + i))
	    return False;
    return True;
}
