/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <string.h>
#include "common.h"
#include "opstr.h"
#include "interp.h"
#if defined(sgi)
#include <strings.h>
#endif

QuadNode* free_quad = NULL;

QuadNode* alloc_quad_node(void)
{
    QuadNode* p;

    if(free_quad == NULL)
	p = (QuadNode *)xmalloc(sizeof(QuadNode));
    else
    {
	p = free_quad;
	free_quad = free_quad->next;
    }

    bzero((char *)p, sizeof(QuadNode));
    p->next = NULL;
    p->prev = NULL;

    return p;
}

QuadNode* make_quad_node1(Quad1* q)
{
    QuadNode* node;

    node = alloc_quad_node();
    node->u.q1 = *q;
    node->type = Q_DATA1;

    return node;
}

QuadNode* make_quad_node2(Quad2* q)
{
    QuadNode* node;

    node = alloc_quad_node();
    node->u.q2 = *q;
    node->type = Q_DATA2;

    return node;
}

QuadNode* gen_quad_node1(unsigned long op, Symbol* res, Symbol* a1, Symbol* a2)
{
    Quad1 q;

    q.op = op;
    q.res = (res != NULL ? res : null_symbol);
    q.a1  = (a1  != NULL ? a1  : null_symbol);
    q.a2  = (a2  != NULL ? a2  : null_symbol);

    return make_quad_node1(&q);
}

QuadNode* gen_label_node(Symbol* label)
{
    QuadNode* node;

    node = gen_quad_node1(LABEL,
			  NULL,
			  label,
			  NULL);
    label->data.label_ref = node;

    return node;
}


void reuse_quad_node(QuadNode* node)
{
    QuadNode* p;

    if(node == NULL)
	return;

    for(p = node; p->next != NULL; p = p->next)
	;

    p->next = free_quad;

    free_quad = node;
}

void reuse_quad_list(QuadList* q)
{
    reuse_quad_node(q->first);
}

void free_quad_node(void)
{
    QuadNode* p;

    while(free_quad)
    {
	p = free_quad;
	free_quad = free_quad->next;
	xfree(p);
    }
}

void init_quad_list(QuadList* q)
{
    if(q != NULL)
	bzero((char *)q, sizeof(QuadList));
}

QuadList null_quad_list(void)
{
    QuadList q;

    init_quad_list(&q);
    return q;
}

QuadList append_quadlist(QuadList* q1, QuadList* q2)
{
    if(ISEMPTY_QUADLIST(q1))
	return *q2;
    if(ISEMPTY_QUADLIST(q2))
	return *q1;

    q1->last->next  = q2->first;
    q2->first->prev = q1->last;
    q1->last        = q2->last;

    return *q1;
}

QuadList append_quadlist3(QuadList* q1, QuadList* q2, QuadList* q3)
{
    if(ISEMPTY_QUADLIST(q1))
	return append_quadlist(q2, q3);
    if(ISEMPTY_QUADLIST(q2))
	return append_quadlist(q1, q3);
    if(ISEMPTY_QUADLIST(q3))
	return append_quadlist(q1, q2);

    q1->last->next  = q2->first;
    q2->first->prev = q1->last;
    q2->last->next  = q3->first;
    q3->first->prev = q2->last;
    q1->last        = q3->last;

    return *q1;
}

QuadList append_quadlistn(QuadList* q, int n)
{
    int i;
    QuadList x;

    if(n <= 1)
    {
	if(q == NULL || n == 0)
	    return null_quad_list();
	return *q;
    }

    for(x = *q, i = 1; i < n; i++)
	x = append_quadlist(&x, q + i);

    return x;
}

Bool quad_next(QuadList* q)
{
    if(ISLAST_QUADLIST(q))
	return False;

    q->cur = q->cur->next;
    return True;
}

Bool quad_prev(QuadList* q)
{
    if(ISFIRST_QUADLIST(q))
	return False;
    q->cur = q->cur->prev;
    return True;
}

void quad_goto_first(QuadList* q)
{
    q->cur = q->first;
}

void quad_goto_last(QuadList* q)
{
    q->cur = q->last;
}

QuadNode* cur_quad_node(QuadList* q)
{
    return q->cur;
}

QuadList quad_node_list(QuadNode* node)
{
    QuadList q;

    q.first = q.last = q.cur = node;

    return q;
}

#define IF_NULL_GEN_NEWLIST(q, node) if(ISEMPTY_QUADLIST(q)) return quad_node_list(node)
QuadList insert_quad_node_next(QuadList* q, QuadNode* node)
{
    IF_NULL_GEN_NEWLIST(q, node);
    if(ISLAST_QUADLIST(q))
	return add_quad_node_last(q, node);

    node->next = q->cur->next;
    node->prev = q->cur;
    node->prev->next = node;
    node->next->prev = node;

    return *q;
}

QuadList insert_quad_node_prev(QuadList* q, QuadNode* node)
{
    IF_NULL_GEN_NEWLIST(q, node);
    if(ISLAST_QUADLIST(q))
	return add_quad_node_first(q, node);

    node->next = q->cur;
    node->prev = q->cur->prev;
    node->prev->next = node;
    node->next->prev = node;

    return *q;
}

QuadList add_quad_node_first(QuadList* q, QuadNode* node)
{
    IF_NULL_GEN_NEWLIST(q, node);

    node->next = q->first;
    node->next->prev = node;
    q->first = node;

    return *q;
}

QuadList add_quad_node_last(QuadList* q, QuadNode* node)
{
    IF_NULL_GEN_NEWLIST(q, node);

    node->prev = q->last;
    node->prev->next = node;
    q->last = node;

    return *q;
}
#undef IF_NULL_GEN_NEWLIST

QuadList delete_quad_list(QuadList* q)
{
    QuadNode* node;

    if(ISEMPTY_QUADLIST(q))
	return *q;

    if(ISFIRST_QUADLIST(q))
    {
	node = q->first;

	if(ISLAST_QUADLIST(q))
	{
	    reuse_quad_node(node);
	    q->first = q->last = NULL;
	    return *q;
	}

	q->cur = q->first = q->first->next;
	node->next = NULL;
	reuse_quad_node(node);
	q->first->prev = NULL;

	return *q;
    }

    if(ISLAST_QUADLIST(q))
    {
	node = q->last;

	if(ISFIRST_QUADLIST(q))
	{
	    reuse_quad_node(node);
	    q->first = q->last = NULL;
	    return *q;
	}

	q->cur = q->last = q->last->prev;
	reuse_quad_node(node);
	q->last->next = NULL;
	return *q;
    }

    node = q->cur;
    node->prev->next = node->next;
    node->next->prev = node->prev;
    q->cur = node->next;
    node->next = NULL;
    reuse_quad_node(node);

    return *q;
}

QuadList add_quadlistnode1_first(QuadList* q, int op, Symbol* res, Symbol* a1, Symbol* a2)
{
    return add_quad_node_first(q, gen_quad_node1(op, res, a1, a2));
}

QuadList add_quadlistnode1_last(QuadList* q, int op, Symbol* res, Symbol* a1, Symbol* a2)
{
    return add_quad_node_last(q, gen_quad_node1(op, res, a1, a2));
}

static void print_quadarg(int a)
{
    if(a > 0)
	printf("M[%d]\t", a);
    else if(a < 0)
	printf("M[sp%d]\t", a);
    else
	printf("-\t");
}

void print_quadlist2(Quad2* q, int i)
{
    printf("[%d]\t%s\t", i, opstr(q->op));

    if(IS_RELGOTOOP(q->op))
    {
	printf("%ld\t", q->res);
	print_quadarg(q->a1);
	print_quadarg(q->a2);
    }
    else if(IS_GOTOOP(q->op))
    {
	printf("%ld\t", q->res);
	print_quadarg(q->a1);
	print_quadarg(q->a2);
    }
    else if(IS_ARRAYOP(q->op))
    {
	print_quadarg(q->res);
	print_quadarg(q->a1);
	printf("%ld", q->a2);
    }
    else if(q->op == CALL)
    {
	print_quadarg(q->res);
	printf("F[%ld]\t-", q->a1);
    }
    else if(q->op == NEW)
    {
	print_quadarg(q->res);
	switch(q->a1)
	{
	  case QH_Int:		printf("<int>\t");   break;
	  case QH_Real:		printf("<real>\t");  break;
	  case QH_Complex:	printf("<cmplx>\t"); break;
	  case QH_String:	printf("<str>\t");   break;
	  case QH_Any:		printf("-\t");   break;
	  default:              printf("<void>\t");  break;
	}
	printf("%ld", q->a2);
    }
    else
    {
	print_quadarg(q->res);
	print_quadarg(q->a1);
	print_quadarg(q->a2);
    }
    putchar('\n');
}

static void print_quadlist1(Quad1* q)
{
    if(q->op == LABEL)
    {
	printf("%s\n", symbol_name(q->a1));
    }
    else if(IS_RELGOTOOP(q->op))
    {
	printf("%s ", opstr(q->op));
	printf("%s ", symbol_name(q->res));
	printf("%s ", symbol_name(q->a1));
	printf("%s\n", symbol_name(q->a2));
    }
    else if(IS_GOTOOP(q->op))
    {
	printf("%s ", opstr(q->op));
	printf("%s ", symbol_name(q->res));
	printf("%s ", symbol_name(q->a1));
	printf("%s\n", symbol_name(q->a2));
    }
    else
    {
	printf("%s ", opstr(q->op));
	printf("%s<%d> ", symbol_name(q->res), q->res->data.align);
	printf("%s ", symbol_name(q->a1));
	printf("%s\n", symbol_name(q->a2));
    }
}

void print_quadlist(QuadList* q)
{
    QuadNode* node;
    int i;

    i = 1;
    for(node = q->first; node != NULL; node = node->next)
    {
	if(node->tp.fname)
	    printf("%-7s ", STRING(node->tp.fname));
	else
	    printf("        ");

	if(node->tp.lineno)
	    printf("%-7ld ", node->tp.lineno);
	else
	    printf("        ");


	if(node->type == Q_DATA1)
	    print_quadlist1(&node->u.q1);
	else
	    print_quadlist2(&node->u.q2, i);
	i++;
    }
}
