/* Copyright (c) 1995, Masanao Izumo */

#ifndef ___QUADLIST_H_
#define ___QUADLIST_H_

#define ISEMPTY_QUADLIST(q) ((q)->first == NULL)
#define ISFIRST_QUADLIST(q) ((q)->first == (q)->cur)
#define ISLAST_QUADLIST(q)  ((q)->last  == (q)->cur)

enum quad_data_types {
    Q_DATA1,
    Q_DATA2
};

typedef struct _Quad1
{
    long op;
    Symbol* res;
    Symbol* a1;
    Symbol* a2;
} Quad1;

typedef struct _Quad2
{
    long op;
    long res;
    long a1;
    long a2;
} Quad2;

typedef struct _QuadNode
{
    int pos;
    int type;			/* see quad_data_types */
    union
    {
	long op;
	Quad1 q1;
	Quad2 q2;
    } u;
    QHTextPos tp;		/* text position */
    struct _QuadNode* next;
    struct _QuadNode* prev;
} QuadNode;

typedef struct _QuadList
{
    QuadNode* cur;
    QuadNode* first;
    QuadNode* last;
} QuadList;

QuadNode* alloc_quad_node(void);
void reuse_quad_node(QuadNode* node);
void reuse_quad_list(QuadList* q);
void free_quad_node(void);

QuadNode* make_quad_node1(Quad1* q);
QuadNode* make_quad_node2(Quad2* q);
QuadNode* gen_quad_node1(unsigned long op, Symbol* res, Symbol* a1, Symbol* a2);
QuadNode* gen_label_node(Symbol* label);

void init_quad_list(QuadList* q);
/* initialize list */

QuadList null_quad_list(void);
/* null quad list return */

QuadList quad_node_list(QuadNode* node);
/* node to list */

/******** some list oprations ********/
Bool quad_next(QuadList* q);
Bool quad_prev(QuadList* q);
void quad_goto_first(QuadList* q);
void quad_goto_last(QuadList* q);
QuadNode* cur_quad_node(QuadList* q);
QuadList insert_quad_node_next(QuadList* q, QuadNode* node);
QuadList insert_quad_node_prev(QuadList* q, QuadNode* node);
QuadList add_quad_node_first(QuadList* q, QuadNode* node);
QuadList add_quad_node_last(QuadList* q, QuadNode* node);
QuadList delete_quad_list(QuadList* q);	/* delete current node */
QuadList add_quadlistnode1_first(QuadList* q, int op, Symbol* res, Symbol* a1, Symbol* a2);
QuadList add_quadlistnode1_last(QuadList* q, int op, Symbol* res, Symbol* a1, Symbol* a2);

QuadList append_quadlist(QuadList* q1, QuadList* q2);
QuadList append_quadlist3(QuadList* q1, QuadList* q2, QuadList* q3);
QuadList append_quadlistn(QuadList* q, int n);
/* append quad list */

Symbol* tree_quadlist(Tree* t, QuadList* qlist_return);
/* return value is result of tree expression */

void translate_sttnew_list(Tree* param, int entry, QuadList* qlis);

extern QuadNode* free_quad;

/* for debug */
void print_quadlist(QuadList* q);
void print_quadlist2(Quad2* q, int i);

/* if <left> <or/and/nor/nand> <right> goto <label> */
void orgoto_quad1(Tree* left, Tree* right, Symbol* label, QuadList* qlis);
void andgoto_quad1(Tree* left, Tree* right, Symbol* label, QuadList* qlis);
void norgoto_quad1(Tree* left, Tree* right, Symbol* label, QuadList* qlis);
void nandgoto_quad1(Tree* left, Tree* right, Symbol* label, QuadList* qlis);

#endif /* ___QUADLIST_H_ */
