/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "intStack.h"
#include "optmz_tree.h"
#include "SwitchStmt.defs"
#include "SwitchStmtStack.h"

static int label_counter = 0;
static intStack bstack;
static intStack cstack;
static SwitchStmtStack switch_case_stack;
static int  enum_counter = 0;

struct CaseLabelList
{
    Symbol* label_symbol;
    Symbol* sym;
    struct CaseLabelList* next;
};

static struct CaseLabelList* free_case_label_list = NULL;

extern long current_module;

#define CHECK_EQ_AND_ASSIGN(t) \
    if(Wall_flag == True && \
       t != NULL && \
       t->val == ASSIGN  && \
       t->type == T_OP) \
	write_warning_with_line("May be type-miss: `==' and `='", t->tp.lineno);

static struct CaseLabelList* new_case_label_node(void)
{
    struct CaseLabelList* node;

    if(free_case_label_list != NULL)
    {
	node = free_case_label_list;
	free_case_label_list = free_case_label_list->next;
    }
    else
    {
	node = (struct CaseLabelList *)xmalloc(sizeof(struct CaseLabelList));
    }

    node->label_symbol = NULL;
    node->sym = NULL;
    node->next = NULL;

    return node;
}

static void reuse_case_label_node(struct CaseLabelList* node)
{
    struct CaseLabelList* start_node = node;

    if(node == NULL)
	return;
    while(node->next)
	node = node->next;
    node->next = free_case_label_list;
    free_case_label_list = start_node;
}

void init_stmt(void)
{
    makeintStack(&bstack, 16);
    makeintStack(&cstack, 16);
    makeSwitchStmtStack(&switch_case_stack, 256);
}

void free_stmt(void)
{
    struct CaseLabelList* tmp;

    while(free_case_label_list)
    {
	tmp = free_case_label_list;
	free_case_label_list = free_case_label_list->next;
	xfree(tmp);
    }

    deleteintStack(&bstack);
    deleteintStack(&cstack);
    deleteSwitchStmtStack(&switch_case_stack);
}

/* writeln(x, y);
 *
 * write - x    -
 * write - y    -
 * write - "\n" -
 */

QuadList gen_write_stmt(Tree* param, int narg, Bool newline)
{
    Tree* root = param;
    QuadList q, qq;
    int i;
    Symbol* sym;
    static int Q_newline = 0;

    q = null_quad_list();
    if(Q_newline == 0)
	Q_newline = install_string("\\n");

    for(i = 0; i < narg; i++)
    {
	sym = tree_quadlist(param->left, &qq);
	q = append_quadlist(&q, &qq);
	q = add_quadlistnode1_last(&q, WRITE, NULL, sym, NULL);
	q.last->tp = sym->tp;
	reuse_if_temp_sym(sym);
	param = param->right;
    }

    if(newline)
    {
	q = add_quadlistnode1_last(&q,
				   WRITE,
				   NULL,
				   install_constant_symbol(Q_newline, SYM_STR),
				   NULL);
    }

    reuse_tree(root);

    return q;
}

QuadList gen_swrite_stmt(Tree* param, int narg)
{
    Tree* root = param;
    Symbol* bswr;
    Symbol*  swr;
    Symbol* eswr;
    Symbol* sym;
    QuadList q, qq;

    if(narg == 0)
	yyerror("too few arguments `swrite'");

    bswr = install_function(install_string("bswrite"), None, 1);
     swr = install_function(install_string( "swrite"), None, 1);
    eswr = install_function(install_string("eswrite"), None, 1);

    sym = tree_quadlist(param->left, &q);
    q = add_quadlistnode1_last(&q, BSWRITE, NULL, sym, NULL);
    q.last->tp = sym->tp;
    reuse_if_temp_sym(sym);
    param = param->right;

    while(param)
    {
	sym = tree_quadlist(param->left, &qq);
	q = append_quadlist(&q, &qq);
	q = add_quadlistnode1_last(&q, SWRITE, NULL, sym, NULL);
	q.last->tp = sym->tp;
	reuse_if_temp_sym(sym);
	param = param->right;
    }

    q = add_quadlistnode1_last(&q, ESWRITE, NULL, NULL, NULL);
    q.last->tp = sym->tp;

    reuse_tree(root);

    return q;
}
    

void gen_do_while_head_stmt(void)
{
    push_bstack(newlabel());
    push_cstack(newlabel());
}

QuadList gen_do_while_tail_stmt(Tree* expr, QuadList* body)
{
    QuadList q, e;
    Symbol* lb;
    Symbol* lc;
    Symbol* label;

    CHECK_EQ_AND_ASSIGN(expr);

    lb    = install_label(pop_bstack(), current_module);
    lc    = install_label(pop_cstack(), current_module);
    label = install_label(newlabel(),   current_module);
    e = gen_if_goto(expr, label);

    q = add_quad_node_first(body, gen_label_node(label));
    q = add_quad_node_last(&q,    gen_label_node(lc));
    q = append_quadlist(&q, &e);
    q = add_quad_node_last(&q,    gen_label_node(lb));

    return q;
}

QuadList gen_break_stmt(void)
{
    return quad_node_list(GOTONODE(search_label(top_bstack(),
						current_module)));
}

QuadList gen_continue_stmt(void)
{
    return quad_node_list(GOTONODE(search_label(top_cstack(),
						current_module)));
}

void for_head_stmt(void)
{
    push_bstack(newlabel());
    push_cstack(newlabel());
}

QuadList for_tail_stmt(Tree* expr1, Tree* expr2, Tree* expr3, QuadList* body)
{
    QuadList q, e1, e2, e3;
    Symbol* sym;
    Symbol* label;
    Symbol* lb;
    Symbol* lc;
    Symbol* fstart;

    CHECK_EQ_AND_ASSIGN(expr2);

    label  = install_label(newlabel(), current_module);
    fstart = install_label(newlabel(), current_module);
    lb     = install_label(pop_bstack(), current_module);
    lc     = install_label(pop_cstack(), current_module);

    sym = tree_quadlist(expr1, &e1);
    reuse_tree(expr1);
    reuse_if_temp_sym(sym);

    if(expr2 != NULL)
	e2 = gen_if_goto(expr2, label);
    else
	e2 = quad_node_list(GOTONODE(label));

    sym = tree_quadlist(expr3, &e3);
    reuse_tree(expr3);
    reuse_if_temp_sym(sym);

    q = add_quad_node_first(body, gen_label_node(label));
    q = add_quad_node_first(&q, GOTONODE(fstart));
    q = append_quadlist(&e1, &q);
    q = add_quad_node_last(&q, gen_label_node(lc));
    q = append_quadlist(&q, &e3);
    q = add_quad_node_last(&q, gen_label_node(fstart));
    q = append_quadlist(&q, &e2);
    q = add_quad_node_last(&q, gen_label_node(lb));

    return q;
}

QuadList gen_if_stmt(Tree* expr, QuadList* if_body)
{
    Symbol* label;
    QuadList q;

    CHECK_EQ_AND_ASSIGN(expr);

    label = install_label(newlabel(), current_module);

    q = gen_ifn_goto(expr, label);
    q = append_quadlist(&q, if_body);
    q = add_quad_node_last(&q, gen_label_node(label));

    return q;
}

QuadList gen_if_else_stmt(Tree* expr, QuadList* if_body, QuadList* else_body)
{
    QuadList q;
    Symbol* l1;
    Symbol* l2;

    CHECK_EQ_AND_ASSIGN(expr);

    l1 = install_label(newlabel(), current_module);
    l2 = install_label(newlabel(), current_module);

    q = gen_ifn_goto(expr, l1);
    q = append_quadlist(&q, if_body);
    q = add_quad_node_last(&q, GOTONODE(l2));
    q = add_quad_node_last(&q, gen_label_node(l1));
    q = append_quadlist(&q, else_body);
    q = add_quad_node_last(&q, gen_label_node(l2));

    return q;
}

void while_head_stmt(void)
{
    push_bstack(newlabel());
    push_cstack(newlabel());
}

QuadList while_tail_stmt(Tree* expr, QuadList* body)
{
    QuadList q, e;
    Symbol* lb;
    Symbol* lc;
    Symbol* label;

    CHECK_EQ_AND_ASSIGN(expr);

    lb  = install_label(pop_bstack(), current_module);
    lc  = install_label(pop_cstack(), current_module);
    label = install_label(newlabel(), current_module);

    e = gen_if_goto(expr, label);
    q = add_quad_node_first(body, gen_label_node(label));
    q = add_quad_node_first(&q, GOTONODE(lc));
    q = add_quad_node_last(&q, gen_label_node(lc));
    q = append_quadlist(&q, &e);
    q = add_quad_node_last(&q, gen_label_node(lb));

    return q;
}

QuadList gen_return_stmt(Tree* t)
{
    Symbol* sym;
    QuadList q;

    sym = tree_quadlist(t, &q);
    q = add_quadlistnode1_last(&q, RETURN, NULL, sym, NULL);
    q.last->tp = t->tp;

    reuse_tree(t);
    reuse_if_temp_sym(sym);

    return q;
}

void push_bstack(int val)
{
    pushintStack(&bstack, val);
}

void push_cstack(int val)
{
    pushintStack(&cstack, val);
}

int top_bstack(void)
{
    if(isemptyintStack(&bstack))
    {
	yyerror("break statement not within loop.");
	return 0;
    }
    return topintStack(&bstack);
}

int top_cstack(void)
{
    if(isemptyintStack(&cstack))
    {
	yyerror("break statement not within loop.");
	return 0;
    }
    return topintStack(&cstack);
}

int pop_bstack(void)
{
    if(isemptyintStack(&bstack))
    {
	yyerror("Systematic error: bstack is underflow\n");
	qhexit(1);
    }
    return popintStack(&bstack);
}

int pop_cstack(void)
{
    if(isemptyintStack(&cstack))
    {
	yyerror("Systematic error: cstack is underflow\n");
	qhexit(1);
    }
    return popintStack(&cstack);
}

int newlabel(void)
{
    sprintf(temp_buff, "@L%d", label_counter++);
    return install_string(temp_buff);
}

QuadList gen_goto_stmt(long name, long module)
{
    return quad_node_list(GOTONODE(search_label(name, module)));
}

QuadList gen_label_stmt(long name, long module)
{
    QuadNode* node;
    Symbol* sym;

    sym = install_label(name, module);
    node = gen_quad_node1(LABEL,
			  sym,
			  NULL,
			  NULL);

    sym->data.label_ref = node;
    
    return quad_node_list(node);
}


void pragma_option(const char* opt, const char* val)
{
    static Bool init = False;
    static Bool default_Wall_flag;

    if(init == False)
    {
	init = True;
	default_Wall_flag = Wall_flag;
    }

    if(!strcmp(opt, "Wall"))
    {
	if(!strcmp(val, "True"))
	{
	    Wall_flag = True;
	}
	else if(!strcmp(val, "False"))
	{
	    Wall_flag = False;
	}
	else if(!strcmp(val, "default"))
	{
	    Wall_flag = default_Wall_flag;
	}
	else
	{
	    sprintf(temp_buff, "Invalid #pragma option:%s", val);
	    yyerror(temp_buff);
	}
    }
    else if(!strcmp(opt, "msg"))
    {
	printf("pragma message: %s\n", val);
    }

    else
    {
	sprintf(temp_buff, "Invalid #pragma option:%s", opt);
	yyerror(temp_buff);
    }
}

/* switch */
void switch_head(Tree* t)
{
    SwitchStmt sw;

    sw.default_label = NULL;
    sw.expr_sym = tree_quadlist(t, &sw.head_stmt);
    sw.list = NULL;
    pushSwitchStmtStack(&switch_case_stack, sw);
    push_bstack(newlabel());
}

QuadList new_switch_case_label(Tree* t)
{
    struct CaseLabelList* node;
    SwitchStmt sw;
    Symbol* sym = NULL;
    Symbol* label_sym;
    QuadNode* qn;

    sw = popSwitchStmtStack(&switch_case_stack);
    node = new_case_label_node();
    node->next = sw.list;
    sw.list = node;

    switch(t->type)
    {
      case T_CINT:
	sym = install_constant_symbol(t->val, SYM_INT);
	break;
      case T_CREAL:
	sym = install_constant_symbol(t->val, SYM_REAL);
	break;
      case T_CSTR:
	sym = install_constant_symbol(t->val, SYM_STR);
	write_warning("case label is string");
	break;
      default:
	fprintf(stderr, "Systematic error: bad constant object type: %d\n",
		t->type);
	qhexit(1);
    }
    reuse_tree(t);
    label_sym = install_label(newlabel(), current_module);

    sw.list->sym          = sym;
    sw.list->label_symbol = label_sym;
    pushSwitchStmtStack(&switch_case_stack, sw);

    qn = gen_label_node(label_sym);
    label_sym->data.label_ref = qn;
    return quad_node_list(qn);
}

QuadList new_switch_default_label(void)
{
    SwitchStmt sw;
    Symbol* label;
    QuadNode* qn;

    sw = popSwitchStmtStack(&switch_case_stack);
    if(sw.default_label != NULL)
	yyerror("multiple default labels in one switch");

    label = sw.default_label = install_label(newlabel(), current_module);
    qn = gen_quad_node1(LABEL, label, NULL, NULL);
    label->data.label_ref = qn;
    pushSwitchStmtStack(&switch_case_stack, sw);

    return quad_node_list(qn);
}


QuadList gen_switch_stmt(QuadList* q)
{
    Symbol* break_label;
    Symbol* default_label;
    struct CaseLabelList* list;
    SwitchStmt sw;
    QuadList cmp_list;
    Symbol* esym;

    sw = popSwitchStmtStack(&switch_case_stack);
    break_label = install_label(pop_bstack(), current_module);
    list = sw.list;
    esym = sw.expr_sym;
    if(sw.default_label == NULL)
	default_label = break_label;
    else
	default_label = sw.default_label;

    /* make comparing code */
    init_quad_list(&cmp_list);
    while(list)
    {
	cmp_list = add_quadlistnode1_first(&cmp_list,
					   IFEQGOTO,
					   list->label_symbol,
					   esym,
					   list->sym);
	list = list->next;
    }
    cmp_list = add_quadlistnode1_last(&cmp_list, GOTO, default_label,
				      NULL, NULL);

    *q = append_quadlist3(&sw.head_stmt, &cmp_list, q);
    *q = add_quad_node_last(q, gen_label_node(break_label));

    reuse_case_label_node(sw.list);

    return *q;
}

void start_enumerate(void)
{
    enum_counter = 0;
}

void define_enumerate(long name, Tree* t)
{
    if(t != NULL)
    {
	/* t->type shuld be T_CINT */
	if(t->type != T_CINT)
	{
	    sprintf(temp_buff, "enumerator value for `%s' not integer constant"
		    , module_name_string(name, current_module));
	    yyerror(temp_buff);
	    return;
	}
	enum_counter = atoi(STRING(t->val));
    }

    install_new_enumerator(name, current_module, enum_counter++);
}
