/* Copyright (c) 1995, Masanao Izumo */

#ifndef ___SYMTAB_H_
#define ___SYMTAB_H_

#define SYMTABSIZ 257
#define R_REGISTER_REPRESENT "%r"

enum
{
    SYM_GLOBAL = 0,		/* global scope */
    SYM_LOCAL  = 1		/* local scope  */
};

enum symbol_types
{
				/* None */
    SYM_VAR = 1,		/* variable         */
    SYM_INT,			/* constant integer */
    SYM_REAL,			/* constant real    */
    SYM_STR,			/* constant string  */
    SYM_LABEL,			/* label            */
    SYM_FUNC,			/* function         */
    SYM_KEYWORD,		/* only use in translation */
    SYM_STMT_FUNC,		/* statement function definition */
    SYM_STRUCT,			/* struct tag symbol */
    SYM_KEYOBJ,			/* keyword object */
    SYM_ENUM			/* enumerator */
};

enum function_types
{
				/* None */
    FUNC_GLOBAL = 1,		/* ordinary function */
    FUNC_GLOBAL_STFN,		/* global statement function */
    FUNC_LOCAL_STFN		/* local statement function */
};

struct SymbolList
{
    struct _Symbol* sym;
    struct SymbolList* next;
};

typedef struct _Symbol
{
    long name;			/* symbol name              */
    long module;		/* module name              */
    struct SymbolList* local_binded; /* local binded symbol    */
    unsigned char type;		/* see symbol_types         */
    unsigned char ref_cnt;	/* reference counter of binding */

    /* some flags */
    unsigned initialized_flag: 1; /* is initialized */
    unsigned referenced_flag: 1; /* is referenced */
    unsigned declared_flag: 1;	/* is declared */
    unsigned scope: 1;		/* SYM_GLOBAL or SYM_LOCAL */
    unsigned used_variable: 1;	/* is used variable */
    unsigned dont_expand_struct_init: 1; /* don't expand struct initialize */
    unsigned dmy: 2;		/* unused yet               */

    union			/* symbol datas             */
    {
	int align;		/* align of variable        */
	int label;		/* label addr               */
	int func;		/* entry for function table */
	int struct_entry;	/* struct entry number */
	struct _Symbol* enum_val; /* enum value */
	struct _QuadNode* label_ref;	/* reference to QuadNode label */
	struct {
	    struct _Tree* func; /* function */
	    struct _Tree* expr;	/* expression */
	} stmtfn;
    } data;

    QHTextPos tp;		/* text position for symbol */

    struct _Symbol* next;	/* pointer to next entry    */
} Symbol;

extern Symbol* global_symbol[];	      /* global variable */
extern Symbol* local_symbol[];	      /* local variable  */
extern Symbol* local_label_symbol[];  /* local label     */
extern Symbol* global_label_symbol[]; /* global label    */
extern Symbol* constant_symbol[];     /* constant value  */
extern Symbol* function_symbol[];     /* function        */
extern Symbol* keyword_symbol[];      /* keyword         */
extern Symbol* local_stmt_func_symbol[]; /* local statement function defs */
extern Symbol* global_stmt_func_symbol[]; /* global statement function defs */
extern Symbol* local_struct_symbol[]; /* local struct symbols */
extern Symbol* global_struct_symbol[];/* global struct symbols */
extern Symbol* local_enum_symbol[];   /* local enum symbols */
extern Symbol* global_enum_symbol[];  /* global enum symbols */
extern Symbol* null_symbol;	      /* null symbol     */

void init_symbol(void);
/* - Initialize symbol tables.
 *
 * Must call once before using floows functions.
 */

void reuse_symbol(Symbol* sym);
/* - Reuse specified symbol. */

void delete_all_symbol(void);
/* delete all symbol */

Symbol* install_variable(long name, long module);
/* - Install variable
 *
 * If name is defined in local_symbol, returns it's Symbol,
 * otherwise install name in global_symbol, and returns it's Symbol.
 */
Symbol* install_local_symbol(long name, long module);
Symbol* install_global_symbol(long name, long module);



Symbol* install_keyword_symbol(long key);
/* install key number */

Symbol* install_label(long name, long module);
Symbol* search_label(long name, long module);


Symbol* newtemp_symbol(void);
/* allocate new temporary variable space */

Symbol* install_constant_symbol(long name, int symtype);
/* - Install constant */

Symbol* install_function(long name, long module, int narg);
void    install_built_in_function(const char* name, int token, int narg);
/* - Install function */

Symbol* install_integer(int val);
/* - Install integer */

void local_bind_symbol_name(long name, long module, Symbol* tmp);
/* bind name to tmp */

void local_unbind_symbol_name(long name, long module);
/* unbind name */

void install_stmt_func_tree(struct _Tree* func, struct _Tree* expr);
/* install statement function definition */

Symbol* search_stmt_function(struct _Tree* func);
/* search statement function definition
 * func->type must be T_FCALL
 */

void print_all_global_symbol(void);
/* print all global symbols */

void clear_local_symbol_table(void);
void clear_local_label_symbol_table(void);
void clear_global_label_symbol_table(void);
void clear_local_stmt_func_symbol(void);

/* for debugging */
char* symbol_name(Symbol* sym);

void write_unused_local_variable(void);

void set_initialized_flag(long name, long module);
void set_initialized_flag_all(struct _Tree* t);

void declare_global_variable(long name, long module);

/**** struct ****/
Symbol* search_struct_symbol(struct _Tree* tag_node, Bool search_all);
void clear_local_struct_symbol_table(void);
void clear_global_struct_symbol_table(void);

/* enum */
void install_new_enumerator(long name, long module, int enumval);
void clear_local_enum_table(void);
Symbol* search_enumerate(long name, long module);

#endif /* ___SYMTAB_H_ */
