/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "opstr.h"
#include "qhobj.h"
#if defined(sgi)
#include <strings.h>
#endif

extern long current_module;

/* #define CHECK_ALLOCATE_NODE */

Tree* free_tree = NULL;
Tree null_tree_node =
{
    T_CINT,			/* type */
    0,				/* optimize_flag */
    0,				/* ftype */
    0,				/* dmy */
    0,				/* val */
    0,				/* module */
    {
	0,			/* tp.fname */
	0,			/* tp.lineno */
    },
    NULL,			/* left */
    NULL,			/* right */
};

Tree* alloc_tree(void)
{
    Tree* node;

#ifdef CHECK_ALLOCATE_NODE
    static int counter;
#endif

    if(free_tree)
    {
	node = free_tree;
	free_tree = free_tree->right;
    }
    else
	node = (Tree *)xmalloc(sizeof(Tree));

#ifdef CHECK_ALLOCATE_NODE
    printf("%4d TreeNode + %p\n", ++counter, node);
#endif

    bzero((void *)node, sizeof(Tree));

    node->left = node->right = NULL;
    node->module = current_module;

    return node;
}

void reuse_tree(Tree* node)
{
#ifdef CHECK_ALLOCATE_NODE
    static int counter;
#endif

    if(&null_tree_node == node)
	return;
    if(node)
    {
	if(node->type != None)
	{

#ifdef CHECK_ALLOCATE_NODE
	    printf("%4d TreeNode - %p\n", ++counter, node);
#endif
	    node->type = None;
	    reuse_tree(node->left);
	    reuse_tree(node->right);

	    /* concatnate current node to free_tree */
	    node->right = free_tree;
	    free_tree = node;
	}
    }
}

void free_reuse_tree(void)
{
    Tree* p;

    while(free_tree)
    {
	p = free_tree;
	free_tree = free_tree->right;
	xfree(p);
    }
}

Tree* gen_node(int type, int val, Tree* left, Tree* right, QHTextPos* tp)
{
    Tree* node;

    node = alloc_tree();
    node->type = type;
    node->val = val;
    node->left = left;
    node->right = right;
    if(tp != NULL)
	node->tp = *tp;
    else
    {
	node->tp.fname = qhfilename;
	node->tp.lineno = qhlineno;
    }

    return node;
}

Tree* gen_atom_node(int type, int val, QHTextPos* tp)
{
    return gen_node(type, val, NULL, NULL, tp);
}

Tree* gen_comma_node(Tree* left, Tree* right, QHTextPos* tp)
{
    return gen_node(T_COMMA, ',', left, right, tp);
}

Tree* gen_op_node(int val, Tree* left, Tree* right, QHTextPos* tp)
{
    return gen_node(T_OP, val, left, right, tp);
}

Tree* gen_uop_node(int val, Tree* node, QHTextPos* tp)
{
    return gen_node(T_OP, val, node, NULL, tp);
}

static void print_tree1(Tree* t)
{
    if(t == NULL)
	return;
    switch(t->type)
    {
      case T_OP:
	printf("%s(", opstr(t->val));
	print_tree1(t->left);
	printf(", ");
	print_tree1(t->right);
	printf(")");
	break;
      case T_COMMA:
	print_tree1(t->left);
	printf(", ");
	print_tree1(t->right);
	break;
      case T_VAR:
	printf("%s", module_name_string(t->module, t->val));
	break;
      case T_CINT:
      case T_CREAL:
      case T_CSTR:
	printf("%s", STRING(t->val));
	break;
      case T_PARAM:
	while(1)
	{
	    printf("param(");
	    print_tree1(t->left);
	    t = t->right;
	    printf(")");
	    if(t == NULL)
		break;
	    printf(", ");
	}
	break;
      case T_FCALL:
	printf("%s/%ld(", module_name_string(t->left->module, t->left->val), t->val);
	print_tree1(t->right);
	printf(")");
	break;
      case T_ARRAY:
	printf("(");
	print_tree1(t->left);
	printf(")[");
	print_tree1(t->right);
	printf("]");
	break;
      case T_STRUCT:
	printf("struct ");

	if(t->val < 0)
	{
	    printf("%s ", module_name_string(t->left->left->module, t->left->left->val));
	    print_tree1(t->left->right);
	    break;
	}

	if(t->left->left != NULL)
	    printf("%s ", module_name_string(t->left->left->module, t->left->left->val));
	printf("{");
	print_tree1(t->right);
	printf("}");
	if(t->left->val > 0)
	    print_tree1(t->left->right);
	printf(";");
	break;
    }
}

void print_tree(Tree* t)
{
    print_tree1(t);
    putchar('\n');
}

Tree* gen_sizeof_type_spec(int type, QHTextPos* tp)
{
    Tree* t;

    t = alloc_tree();
    t->type = T_CINT;
    t->tp = *tp;

    switch(type)
    {
      case INT_OP:
	sprintf(temp_buff, "%d", sizeof(int));
	break;
      case REAL_OP:
	sprintf(temp_buff, "%d", sizeof(double));
	break;
      case COMPLEX_OP:
	sprintf(temp_buff, "%d", sizeof(XComplex));
	break;
      default:
	strcpy(temp_buff, "4");
	break;
    }
    t->val = install_string(temp_buff);

    return t;
}

Tree* copy_tree(Tree* tree)
{
    Tree* t;

    if(tree == NULL)
	return NULL;

    t = alloc_tree();
    *t = *tree;
    t->left  = copy_tree(tree->left);
    t->right = copy_tree(tree->right);
    return t;
}

/*
         T_COND3
        /       \
cond-expr        None
                /    \
       true-expr      false-expr
*/
Tree* gen_cond_op3(Tree* expr, Tree* true, Tree* false, QHTextPos *tp)
{
    Tree* t = alloc_tree();

    t->tp = *tp;
    t->right = alloc_tree();
    t->right->tp = *tp;
    t->type = T_COND3;

    t->left = expr;
    t->right->left = true;
    t->right->right = false;

    return t;
}

Tree* gen_structof_node(Tree* t, QHTextPos *tp)
{
    Symbol* sym;

    sym = search_struct_symbol(t, True);

    if(sym->type == None)
    {
	sprintf(temp_buff, "Undefined struct `%s'",
		module_name_string(sym->module, sym->name));
	yyerror(temp_buff);
	return &null_tree_node;
    }

    t->type = T_CINT;
    t->tp   = *tp;

    sprintf(temp_buff, "%d", sym->data.struct_entry);
    t->val = install_string(temp_buff);

    return t;
}
