/* Copyright (c) 1995, Masanao Izumo */

#ifndef ___TREE_H_
#define ___TREE_H_

enum tree_types
{
				/* None */
    T_OP = 1,			/* operator */
    T_COMMA,			/* comma */
    T_VAR,			/* variable */
    T_CINT,			/* constant integer */
    T_CREAL,			/* constant real */
    T_CSTR,			/* constant string */
    T_PARAM,			/* function parameter */
    T_FCALL,			/* function call */
    T_ARRAY,			/* array */
    T_ASSOCARRAY,		/* assoc array */
    T_FUNCALL,			/* funcall */
    T_NEW,			/* new */
    T_ASCNEW,			/* assoc new */
    T_STRNEW,			/* string new */
    T_STTNEW,			/* struct new */
    T_COND3,			/* condition operator ( ? : ) */
    T_QHKEYWORD,		/* keyword object: (NULL) */
    T_LIST,			/* list( ... ) */
    T_STRUCT			/* struct tag name; */
};

typedef struct _Tree
{
    char type;			/* see tree_types */
    char optimize_flag;		/* To be eficient optimizing tree */
    char ftype;			/* function type (only use type == T_FUNCALL)*/
    char module_specified;	/* is module specified variable */
    long val;			/* name or value of the node */
    long module;		/* module name */
    QHTextPos tp;		/* text position */
    struct _Tree* left;		/* pointer to left tree node */
    struct _Tree* right;	/* pointer to right tree node */
} Tree;

Tree* alloc_tree(void);
/* Allocate the new tree node.
 * Don't free the node directory, because the node is reused after called
 * reuse_tree();
 */

void reuse_tree(Tree* node);

void free_reuse_tree(void);

Tree* gen_node(int type, int val, Tree* left, Tree* right, QHTextPos *tp);
Tree* gen_atom_node(int type, int val, QHTextPos *tp);
Tree* gen_comma_node(Tree* left, Tree* right, QHTextPos *tp);
Tree* gen_op_node(int val, Tree* left, Tree* right, QHTextPos *tp);
Tree* gen_uop_node(int val, Tree* node, QHTextPos *tp);
Tree* gen_sizeof_type_spec(int type, QHTextPos *tp);
Tree* gen_structof_node(Tree* t, QHTextPos *tp);
Tree* gen_cond_op3(Tree* expr, Tree* true, Tree* false, QHTextPos *tp);

Tree* copy_tree(Tree* tree);

void print_tree(Tree* t);
/* for debugging */

extern Tree* free_tree;		/* list of the reusable tree node */
extern Tree null_tree_node;	/* used if error occur */

#endif /* ___TREE_H_ */
