/* Copyright (c) 1995, Masanao Izumo */

/* Convert Tree to QuadList */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "opstr.h"
#include "newtemp.h"
#include "interp.h"
#include "optmz_tree.h"
#include "localvar.h"

#if defined(sgi)
#include <alloca.h>
#endif /* sgi */

static void param_translate1(Tree* param, int narg, int op, QuadList* qlis);
static Symbol* tree_quadlist1(Tree* t, QuadList* qlist_return);

static Symbol* translate_or_tree(Tree* left, Tree* right, QuadList* q);
static Symbol* translate_and_tree(Tree* left, Tree* right, QuadList* q);
static Symbol* tree_varnode_to_symbol(Tree* node);

extern long current_module;

/*
static void write_bind_list(Symbol* sym)
{
    int i;
    struct SymbolList* p;

    i = 0;
    p = sym->local_binded;
    while(p)
    {
	printf("%s ", STRING(p->sym->name));
	p = p->next;
	if(i > 10)
	{
	    printf("...");
	    break;
	}
	i++;
    }
    puts("");
}*/

Symbol* expand_stmt_func(Symbol* func, int narg, Tree* param, QuadList* qlis)
{
#if defined(__GNUC__)
    QuadList qarg[narg];
    Symbol*  sarg[narg];
    char     need_assign[narg];
#else
    QuadList* qarg = (QuadList *)alloca(narg * sizeof(QuadList));
    Symbol**  sarg = (Symbol **)alloca(narg * sizeof(Symbol *));
    char*     need_assign = (char *)alloca(narg);
#endif
    Tree* func_arg_template;
    Tree* expr;
    Tree* p;
    QuadList q;
    int i;
    Symbol* sym;
    
    func_arg_template = func->data.stmtfn.func->right;
    expr              = func->data.stmtfn.expr;

    /* evaluate the arguments */
    for(i = 0; i < narg; i++)
    {
	sarg[i] = tree_quadlist1(param->left, qarg + i);
	param = param->right;
    }
    *qlis = append_quadlistn(qarg, narg);

    p = func_arg_template;
    for(i = 0; i < narg; i++)
    {
	if(p->left->optimize_flag == True ||
	   IS_TEMP_SYM(sarg[i]))
	{
	    need_assign[i] = False;
	    sym = sarg[i];
	}
	else
	{
	    need_assign[i] = True;
	    sym = newtemp_symbol();
	}

	local_bind_symbol_name(p->left->val, p->left->module, sym);
	p = p->right;
    }

    /* assign param to binded temp variable */
    p = func_arg_template;
    for(i = 0; i < narg; i++)
    {
	if(need_assign[i])
	{
	    sym = install_variable(p->left->val, p->left->module);
	    *qlis = add_quadlistnode1_last(qlis, ASSIGN, sym, sarg[i], NULL);
	}
	p = p->right;
    }

    sym = tree_quadlist1(expr, &q);

    *qlis = append_quadlist(qlis, &q);

    unbind_variable_in_param_tree(func_arg_template);
    return sym;
}

static void param_translate1(Tree* param, int narg, int op, QuadList* qlis)
{
#if defined(__GNUC__)
    QuadList qarg[narg];
    Symbol*  sarg[narg];
#else
    QuadList* qarg = (QuadList *)alloca(narg * sizeof(QuadList));
    Symbol**  sarg = (Symbol **)alloca(narg * sizeof(Symbol *));
#endif
    Tree* p;
    int i;

    if(narg == 0)
    {
	init_quad_list(qlis);
    }
    else
    {
	for(i = 0, p = param; i < narg; i++, p = p->right)
	    sarg[i] = tree_quadlist1(p->left, qarg + i);

	for(i = 0; i < narg; i++)
	    reuse_if_temp_sym(sarg[i]);

	*qlis = append_quadlistn(qarg, narg);
	for(i = 0; i < narg; i++)
	{
	    *qlis = add_quadlistnode1_last(qlis, op, NULL, sarg[i], NULL);
	    qlis->last->tp = sarg[i]->tp;
	}
    }
}

static void ipush_translate(Tree* param, int narg, QuadList* qlis)
{
#if defined(__GNUC__)
    QuadList qarg[narg];
    Symbol*  sarg[narg];
#else
    QuadList* qarg = (QuadList *)alloca(narg * sizeof(QuadList));
    Symbol**  sarg = (Symbol **)alloca(narg * sizeof(Symbol *));
#endif
    Tree* p;
    int i;

    if(narg == 0)
    {
	init_quad_list(qlis);
    }
    else
    {
	for(i = 0, p = param; i < narg; i++, p = p->right)
	    sarg[i] = tree_quadlist1(p->left, qarg + i);
	for(i = 0; i < narg; i++)
	    reuse_if_temp_sym(sarg[i]);
	*qlis = append_quadlistn(qarg, narg);

	for(i = 0; i + 3 <= narg; i += 3)
	    *qlis = add_quadlistnode1_last(qlis, IPUSH3,
					   sarg[i + 2], sarg[i], sarg[i + 1]);
	if(i + 2 == narg)
	    *qlis = add_quadlistnode1_last(qlis, IPUSH2,
					   NULL, sarg[i], sarg[i + 1]);
	else if(i + 1 == narg)
	    *qlis = add_quadlistnode1_last(qlis, IPUSH,
					   NULL, sarg[i], NULL);

    }
}

static Symbol* fcall_translate1(Symbol* func, Tree* param, int narg,
				QuadList* qlis, QHTextPos* tp)
{
#if defined(__GNUC__)
    QuadList qarg[narg];
    Symbol*  sarg[narg];
#else
    QuadList* qarg = (QuadList *)alloca(narg * sizeof(QuadList));
    Symbol**  sarg = (Symbol **)alloca(narg * sizeof(Symbol *));
#endif
    Symbol*  sym;
    Symbol*  narg_sym;
    Bool extf;			/* is external function */
    int i;

    narg_sym = install_keyword_symbol(narg);
    extf = FUNCTAB(func->data.func).extern_function;

    if(narg == 0)
    {
	sym = newtemp_symbol();
	if(!extf)
	    *qlis = quad_node_list(gen_quad_node1(CALL, sym, func, narg_sym));
	else
	    *qlis = quad_node_list(gen_quad_node1(CALLEXT, sym, func,
					install_keyword_symbol(narg)));
    }
    else
    {
	/* translate parameter */
	for(i = 0; i < narg; i++, param = param->right)
	    sarg[i] = tree_quadlist1(param->left, qarg + i);

	*qlis = append_quadlistn(qarg, narg);

	/* PARAM - sarg[i] - */
	for(i = 0; i + 3 <= narg; i += 3)
	{
	    *qlis = add_quadlistnode1_last(qlis, PARAM3,
					   sarg[i + 2], sarg[i], sarg[i + 1]);
	    qlis->last->tp = sarg[i + 2]->tp;
	}

	if(i + 2 == narg)
	{
	    *qlis = add_quadlistnode1_last(qlis, PARAM2,
					   NULL, sarg[i], sarg[i + 1]);
	    qlis->last->tp = sarg[i + 1]->tp;
	}
	else if(i + 1 == narg)
	{
	    *qlis = add_quadlistnode1_last(qlis, PARAM,
					   NULL, sarg[i], NULL);
	    qlis->last->tp = sarg[i]->tp;
	}

	/* CALL sym func - */
	sym = newtemp_symbol();

	if(!extf)
	    *qlis = add_quadlistnode1_last(qlis, CALL, sym, func, narg_sym);
	else
	    *qlis = add_quadlistnode1_last(qlis, CALLEXT, sym, func, install_keyword_symbol(narg));
    }
    qlis->last->tp = *tp;

    /* reuse temporary variable */
    for(i = 0; i < narg; i++)
	reuse_if_temp_sym(sarg[i]);

    return sym;
}

    
/*            T_FCALL<narg>
             /             \
  T_VAR<fname,module>       T_PARAM<None>
*/
static Symbol* fcall_translate(Tree* t, QuadList* qlis)
{
    Tree* param;
    Symbol* func;
    long module;
    long name;
    int narg;

    param  = t->right;
    module = t->left->module;
    name   = t->left->val;
    narg   = t->val;

    if((func = search_stmt_function(t)) == NULL)
	func = install_function(name, module, narg);

    if(func->type == SYM_STMT_FUNC)
    {
	return expand_stmt_func(func, narg, param, qlis);
    }

    return fcall_translate1(func, param, narg, qlis, &t->tp);
}

/*

	            T_ARRAY<dim>
	           /             \
         expression               T_PARAM<None>
*/

static Symbol* array_translate(Tree* t, QuadList* qlis)
{
    int narg = t->val;
    Tree* expr  = t->left;
    Tree* param = t->right;
    QuadList aq;
    Symbol* sarg;
    Symbol* sym;
    Symbol* asym;

    if(narg == 0)
    {
	if(expr->type == T_VAR)
	    sprintf(temp_buff, "Array size missing in `%s'", STRING(expr->val));
	else
	    sprintf(temp_buff, "Array size missing");
	yyerror(temp_buff);
	return null_symbol;
    }

    sym = newtemp_symbol();
    if(narg == 1)
    {
	sarg = tree_quadlist1(param->left, qlis);
	asym = tree_quadlist1(expr, &aq);
	*qlis = append_quadlist(qlis, &aq);
	*qlis = add_quadlistnode1_last(qlis, AREF1, sym, asym, sarg);
	sym->tp = qlis->last->tp = t->tp;
	reuse_if_temp_sym(sarg);
    }
    else
    {
	/* translate parameter */
	ipush_translate(param, narg, qlis);

	asym = tree_quadlist1(expr, &aq);
	*qlis = append_quadlist(&aq, qlis);

	/* AREF sym asym narg */
	*qlis  = add_quadlistnode1_last(qlis, AREF, sym, asym, install_keyword_symbol(narg));
	sym->tp = qlis->last->tp = t->tp;
    }

    /* reuse temporary variable */
    reuse_if_temp_sym(asym);

    return sym;
}
    
static Symbol* funcall_translate(Tree* fname_expr, Tree* param, int narg, QuadList* qlis)
{
    Symbol* res;
    Symbol* fname_sym;
    QuadList name_expr;

    res = newtemp_symbol();

    fname_sym = tree_quadlist1(fname_expr, &name_expr);

    if(narg != 0)
    {
	param_translate1(param, narg, PARAM, qlis);
	*qlis = append_quadlist(&name_expr, qlis);
    }
    else
	*qlis = name_expr;

    *qlis = add_quadlistnode1_last(qlis,
				   FUNCALL,
				   res,
				   fname_sym,
				   install_keyword_symbol(narg));

    reuse_if_temp_sym(fname_sym);

    return res;
}

/*
 *     T_NEW<type>
 *     /        \
 * T_PARAM<dim>  NULL
 */
static Symbol* translate_new(int type, int dim, Tree* param, QuadList* qlis, QHTextPos* tp)
{
    Symbol* sym;

    sym = newtemp_symbol();

    ipush_translate(param, dim, qlis);

    *qlis = add_quadlistnode1_last(qlis,
				   NEW,
				   sym,
				   install_keyword_symbol(type),
				   install_keyword_symbol(dim));
    sym->tp = qlis->last->tp = *tp;
    return sym;
}

Symbol* translate_cond3(Tree* expr, Tree* t, Tree* f, QuadList* qlis, QHTextPos* tp)
{
/*
  $ = E ? T : F
  =>
  ifngoto(E) L1
  $ = T
  goto L2
  L1
  $ = F
  L2
*/
    Symbol* sym;
    Symbol* lab1;
    Symbol* lab2;
    Symbol* symt;
    Symbol* symf;
    QuadList qt;
    QuadList qf;

    lab1 = install_label(newlabel(), current_module);
    lab2 = install_label(newlabel(), current_module);
    sym  = newtemp_symbol();
    symt = tree_quadlist1(t, &qt);
    symf = tree_quadlist1(f, &qf);

/*
  =>
  ifngoto(E) L1
  $ = T
  goto L2
  L1
  $ = F
  L2
*/
    *qlis = gen_ifn_goto1(expr, lab1);
    *qlis = append_quadlist(qlis, &qt);
    *qlis = add_quadlistnode1_last(qlis,
				   ASSIGN,
				   sym,
				   symt,
				   NULL);
    reuse_if_temp_sym(symt);
    *qlis = add_quad_node_last(qlis, GOTONODE(lab2));
    *qlis = add_quad_node_last(qlis, gen_label_node(lab1));
    *qlis = append_quadlist(qlis, &qf);
    *qlis = add_quadlistnode1_last(qlis,
				   ASSIGN,
				   sym,
				   symf,
				   NULL);
    reuse_if_temp_sym(symf);
    *qlis = add_quad_node_last(qlis, gen_label_node(lab2));

    return sym;
}


/*
 *      T_ASCNEW<None>
 *     /        \
 * NULL          NULL
 */
static Symbol* translate_ascnew(QHTextPos* tp, QuadList* qlis)
{
    Symbol* sym;

    sym = newtemp_symbol();
    *qlis = quad_node_list(gen_quad_node1(ASCNEW,
					  sym,
					  NULL,
					  NULL));
    sym->tp = qlis->last->tp = *tp;
    return sym;
}

/*
 *      T_STRNEW<None>
 *     /        \
 * NULL,expr    NULL
 */
static Symbol* translate_strnew(QHTextPos* tp, Tree* expr, QuadList* qlis)
{
    Symbol* sym;
    Symbol* expr_sym;
    QuadList eq;

    sym = newtemp_symbol();

    if(expr != NULL)
	expr_sym = tree_quadlist1(expr, &eq);
    else
	expr_sym = install_integer(0);

    *qlis = quad_node_list(gen_quad_node1(STRNEW,
					  sym,
					  expr_sym,
					  NULL));
    sym->tp = qlis->last->tp = *tp;
    if(expr != NULL)
	reuse_if_temp_sym(expr_sym);
    return sym;
}

static void new_struct_proc_local(Symbol* sym, int entry, Tree* t,
				  QuadList* qlis)
{
    Symbol* sym2;
    Tree* p;
    QuadList q;
    long* member;
    int i;

    if(t->val < 0)
	t = struct_tree[entry];

    *qlis = quad_node_list(gen_quad_node1(STTNEW,
					  sym,
					  install_keyword_symbol(entry),
					  NULL));

    member = struct_tag_table[entry].member_id;
    i = 0;
    for(p = t->right; p; p = p->right)
    {
	if(p->left->type == T_VAR)
	    i++;
	else if(p->left->type == T_OP)
	{
	    sym2 = tree_quadlist1(p->left->right, &q);
	    *qlis = append_quadlist(qlis, &q);
	    *qlis = add_quadlistnode1_last(qlis,
					   STSET,
					   sym,
					   install_keyword_symbol(member[i]),
					   sym2);
	    reuse_if_temp_sym(sym2);
	    i++;
	}
	else if(p->left->type == T_STRUCT)
	{
	    Tree* qt;
	    int local_entry;

	    local_entry = (int)(struct_tag_table[entry].next[i]
				- struct_tag_table);

	    for(qt = p->left->left->right; qt; qt = qt->right)
	    {
		sym2 = newtemp_symbol();

		new_struct_proc_local(sym2,
				      local_entry,
				      p->left,
				      &q);
		*qlis = append_quadlist(qlis, &q);
		*qlis = add_quadlistnode1_last(qlis,
					       STSET,
					       sym,
					       install_keyword_symbol(member[i]),
					       sym2);
		reuse_if_temp_sym(sym2);
		i++;
	    }
	}
    }
}

static void new_struct_proc(Symbol* sym, Tree* tag, QuadList* qlis)
{
    Symbol* tag_sym;
    int entry;

    init_quad_list(qlis);
    tag_sym = search_struct_symbol(tag, True);
    if(tag_sym->type == None)
    {
	sprintf(temp_buff, "Undefined struct `%s'",
		module_name_string(tag->module, tag->val));
	yyerror(temp_buff);
	return;
    }

    entry = tag_sym->data.struct_entry;
    new_struct_proc_local(sym, entry, struct_tree[entry], qlis);
}


/*
 *      T_STTNEW<None>
 *     /        \
 * NAME          NULL
 */
static Symbol* translate_sttnew(Tree* tag, QHTextPos* tp, QuadList* qlis)
{
    Symbol* sym;
    
    sym = newtemp_symbol();

    new_struct_proc(sym, tag, qlis);

    return sym;
}

void translate_sttnew_list(Tree* param, int entry, QuadList* qlis)
{
    QuadList q;
    Symbol* sym;

    init_quad_list(qlis);
    while(param)
    {
	sym = tree_varnode_to_symbol(param->left);
	new_struct_proc_local(sym, entry, struct_tree[entry], &q);
	*qlis = append_quadlist(qlis, &q);
	param = param->right;
    }
}


static Symbol* translate_list_op(Tree* param, int narg,
				 QHTextPos* tp, QuadList* qlis)
{
#if defined(__GNUC__)
    QuadList qarg[narg];
    Symbol*  sarg[narg];
#else
    QuadList* qarg = (QuadList *)alloca(narg * sizeof(QuadList));
    Symbol**  sarg = (Symbol **)alloca(narg * sizeof(Symbol *));
#endif
    Symbol*  sym;
    int i;

    if(narg == 0)
    {
	sym = newtemp_symbol();
	*qlis = quad_node_list(gen_quad_node1(LIST_OP, sym,
					      install_keyword_symbol(0),
					      NULL));
    }
    else
    {
	for(i = 0; i < narg; i++, param = param->right)
	    sarg[i] = tree_quadlist1(param->left, qarg + i);
	*qlis = append_quadlistn(qarg, narg);
	/* PARAM - sarg[i] - */
	for(i = 0; i + 3 <= narg; i += 3)
	{
	    *qlis = add_quadlistnode1_last(qlis, PARAM3,
					   sarg[i + 2], sarg[i], sarg[i + 1]);
	    qlis->last->tp = sarg[i + 2]->tp;
	}

	if(i + 2 == narg)
	{
	    *qlis = add_quadlistnode1_last(qlis, PARAM2,
					   NULL, sarg[i], sarg[i + 1]);
	    qlis->last->tp = sarg[i + 1]->tp;
	}
	else if(i + 1 == narg)
	{
	    *qlis = add_quadlistnode1_last(qlis, PARAM,
					   NULL, sarg[i], NULL);
	    qlis->last->tp = sarg[i]->tp;
	}

	sym = newtemp_symbol();
	*qlis = add_quadlistnode1_last(qlis, LIST_OP, sym,
				       install_keyword_symbol(narg),
				       NULL);
    }
    qlis->last->tp = *tp;

    /* reuse temporary variable */
    for(i = 0; i < narg; i++)
	reuse_if_temp_sym(sarg[i]);

    return sym;
}

static Symbol* translate_assoc_ref(Tree* left, Tree* right,
				   QHTextPos* tp, QuadList* qlis)
{
    QuadList lq;
    QuadList rq;
    Symbol* ls;
    Symbol* rs;
    Symbol* sym;

    ls = tree_quadlist1(left, &lq);
    rs = tree_quadlist1(right, &rq);
    *qlis = append_quadlist(&lq, &rq);

    sym = newtemp_symbol();
    *qlis = add_quadlistnode1_last(qlis, ASCREF, sym, ls, rs);
    reuse_if_temp_sym(ls);
    reuse_if_temp_sym(rs);
    sym->tp = qlis->last->tp = *tp;

    return sym;
}

static Symbol* translate_assoc_set(Tree* asct, Tree* keyt, Tree* t,
				   QHTextPos* tp, QuadList* qlis)
{
    QuadList aq;
    QuadList kq;
    QuadList tq;
    Symbol*  as;
    Symbol*  ks;
    Symbol*  ts;

    as = tree_quadlist1(asct, &aq);
    ks = tree_quadlist1(keyt, &kq);
    ts = tree_quadlist1(t,    &tq);

    *qlis = append_quadlist3(&aq, &kq, &tq);
    *qlis = add_quadlistnode1_last(qlis, ASCSET, as, ks, ts);

    reuse_if_temp_sym(as);
    reuse_if_temp_sym(ks);

    ts->tp = qlis->last->tp = *tp;

    return ts;
}


/*
 * a{key} op= x
 * =>
 * $1 = a
 * $2 = key
 * $3 = $1{$2}
 * $3 op= x
 * $1{$2} = $3
 * <= s3
 */
Symbol* translate_assoc_opset(int op, Tree* asct, Tree* keyt, Tree* t,
			      QHTextPos* tp, QuadList* qlis)
{
    Symbol *s1;
    Symbol *s2;
    Symbol *s3;
    Symbol *x;
    QuadList q1, q2, q3;

    s1 = tree_quadlist1(asct, &q1);
    s2 = tree_quadlist1(keyt, &q2);
    x  = tree_quadlist1(t,    &q3);
    s3 = newtemp_symbol();

    *qlis = append_quadlist3(&q1, &q2, &q3);
    *qlis = add_quadlistnode1_last(qlis, ASCREF, s3, s1, s2);
    qlis->last->tp = *tp;
    *qlis = add_quadlistnode1_last(qlis, op, s3, x, NULL);
    qlis->last->tp = *tp;
    *qlis = add_quadlistnode1_last(qlis, ASCSET, s1, s2, s3);
    s3->tp = qlis->last->tp = *tp;

    reuse_if_temp_sym(s1);
    reuse_if_temp_sym(s2);
    reuse_if_temp_sym(x);

    return s3;
}

/* (x).y += (z)
 *
 * $0 = x
 * $1 = $0.y
 * $2 = z
 * $1 += $2
 * $0.y = $1
 */

Symbol* translate_struct_opset(int op, Tree* tag, int id, Tree* t,
			       QHTextPos* tp, QuadList* qlis)
{
    Symbol* t0;
    Symbol* t1;
    Symbol* t2;
    QuadList q;
    Symbol* y;

    y = install_keyword_symbol(id);
    t0 = tree_quadlist1(tag, qlis);
    t1 = newtemp_symbol();
    *qlis = add_quadlistnode1_last(qlis, STREF, t1, t0, y);
    qlis->last->tp = *tp;
    t2 = tree_quadlist1(t, &q);
    *qlis = append_quadlist(qlis, &q);
    *qlis = add_quadlistnode1_last(qlis, op, t1, t2, NULL);
    qlis->last->tp = *tp;
    *qlis = add_quadlistnode1_last(qlis, STSET, t0, y, t1);
    qlis->last->tp = *tp;
    reuse_if_temp_sym(t0);
    reuse_if_temp_sym(t2);
    return t1;
}


/* (*x)++
   =>
 * s1 = *x
 * s2 = s1++
 * *x = s1
 * <= s2
 */
Symbol* translate_indir_incdec(int op, Tree* t, QHTextPos* tp, QuadList* qlis)
{
    Symbol* s1;
    Symbol* s2;
    Symbol* x;

    s1 = newtemp_symbol();
    s2 = newtemp_symbol();

    x = tree_quadlist1(t, qlis);
    *qlis = add_quadlistnode1_last(qlis, INDIR, s1, x, NULL);
    qlis->last->tp = *tp;
    *qlis = add_quadlistnode1_last(qlis, op, s2, s1, NULL);
    *qlis = add_quadlistnode1_last(qlis, AINDIR, s1, x, s1);

    reuse_if_temp_sym(s1);
    reuse_if_temp_sym(x);

    return s2;
}


/* a{key}++
 * =>
 * s1 = a
 * s2 = key
 * s3 = s1{s2}
 * s4 = s3++
 * s1{s2} = s3
 * <= s4
 */
Symbol* translate_assoc_incdec(int op, Tree* asct, Tree* keyt,
			       QHTextPos* tp, QuadList* qlis)
{
    Symbol *s1;
    Symbol *s2;
    Symbol *s3;
    Symbol *s4;
    QuadList q1, q2;

    s1 = tree_quadlist1(asct, &q1);
    s2 = tree_quadlist1(keyt, &q2);
    s3 = newtemp_symbol();
    s4 = newtemp_symbol();
    *qlis = append_quadlist(&q1, &q2);
    *qlis = add_quadlistnode1_last(qlis, ASCREF, s3, s1, s2);
    qlis->last->tp = *tp;
    *qlis = add_quadlistnode1_last(qlis, op, s4, s3, NULL);
    qlis->last->tp = *tp;
    *qlis = add_quadlistnode1_last(qlis, ASCSET, s1, s2, s3);
    s4->tp = qlis->last->tp = *tp;

    reuse_if_temp_sym(s1);
    reuse_if_temp_sym(s2);
    reuse_if_temp_sym(s3);

    return s4;
}

static Symbol* tree_varnode_to_symbol(Tree* t)
{
    Symbol* sym;

    if(t->module_specified == False)
    {
	if((sym = search_enumerate(t->val, t->module)) == NULL)
	    sym = install_variable(t->val, t->module);
    }
    else
    {
	if((sym = search_enumerate(t->val, None)) == NULL)
	    sym = install_global_symbol(t->val, t->module);
    }

    sym->tp = t->tp;
    sym->used_variable = True;

    if(W_uninitialized_variable == True)
    {
	if(sym->type == SYM_VAR &&
	   sym->initialized_flag == False &&
	   STRING(sym->name)[0] != temp_prefix)
	{
	    sym->initialized_flag = True;
	    sprintf(temp_buff, "`%s' may be used uninitialized",
		    module_name_string(sym->module, sym->name));
	    write_warning(temp_buff);
	}
    }

    if(W_undeclared_global_variable == True)
    {
	if(sym->scope == SYM_GLOBAL &&
	   STRING(sym->name)[0] != temp_prefix &&
	   sym->declared_flag == False)
	{
	    sym->declared_flag = True;
	    sprintf(temp_buff, "`%s' undeclared global variable", module_name_string(sym->module, sym->name));
	    write_warning(temp_buff);
	}
    }

    return sym;
}

static Symbol* tree_quadlist1(Tree* t, QuadList* qlis)
{
    Quad1 q;
    Symbol* sym;		/* return symbol */
    QuadList lq;
    QuadList rq;
    QuadNode* node;

    sym = null_symbol;
    init_quad_list(qlis);

    if(t == NULL)
	return null_symbol;

    switch(t->type)
    {
      case T_OP:
	if(IS_ASSIGNOP(t->val))
	{
	    if(t->val == ASSIGN && t->left->type == T_VAR)
		set_initialized_flag(t->left->val, t->left->module);

	    if(t->left->type == T_ARRAY) /* a[] = x, a[] op= x */
	    {
		Tree* a = t->left;
		QuadList qq;
		Symbol* ss;

		ipush_translate(a->right, a->val, &lq);

		ss  = tree_quadlist1(a->left,  &qq);
		sym = tree_quadlist1(t->right,  &rq);

		node = gen_quad_node1(AASET + (t->val - AASSIGN), /* aset */
				      ss,
				      sym,
				      install_keyword_symbol(a->val));
		sym->tp = node->tp = t->tp;

		*qlis = append_quadlist3(&qq, &lq, &rq);
		*qlis = add_quad_node_last(qlis, node);
		reuse_if_temp_sym(ss);
	    }
	    else if(t->left->type == T_OP &&
		    t->left->val  == INDIR) /* *x = y, *x op= y */
	    {
		if(t->val == ASSIGN) /* *x = y */
		{
		    q.op = AINDIR;
		    sym = q.res = newtemp_symbol();
		    q.a1 = tree_quadlist1(t->left->left,  &lq);
		    q.a2 = tree_quadlist1(t->right,  &rq);
		    *qlis = append_quadlist(&lq, &rq);
		    *qlis = add_quadlistnode1_last(qlis,
						   q.op,
						   q.res,
						   q.a1,
						   q.a2);
		    reuse_if_temp_sym(q.a1);
		    reuse_if_temp_sym(q.a2);
		}
		else		/* *x op= y */
		{
		    Symbol* sl;
		    Symbol* sr;
		    QuadList ql;
		    QuadList qr;
		    Symbol* st;

		    /* *left op= right
		       =>
		       $ = *left
		       $ += right
		       *left = $
		     */
		    sym = newtemp_symbol();
		    st = newtemp_symbol();
		    sl = tree_quadlist1(t->left->left, &ql);
		    sr = tree_quadlist1(t->right,       &qr);
		    *qlis = append_quadlist(&ql, &qr);
		    *qlis = add_quadlistnode1_last(qlis, INDIR, st, sl, NULL);
		    qlis->last->tp = t->tp;
		    *qlis = add_quadlistnode1_last(qlis, t->val, st, sr, NULL);
		    qlis->last->tp = t->tp;
		    *qlis = add_quadlistnode1_last(qlis, AINDIR, sym, sl, st);
		    qlis->last->tp = t->tp;
		    reuse_if_temp_sym(sr);
		    reuse_if_temp_sym(sl);
		    reuse_temp(st->name);
		}
	    }
	    else if(t->left->type == T_ASSOCARRAY) /* a{} = x, a{} op= x */
	    {
		if(t->val == ASSIGN) /* a{} = x */
		    sym = translate_assoc_set(t->left->left,
					      t->left->right,
					      t->right,
					      &t->tp,
					      qlis);
		else
		    sym = translate_assoc_opset(t->val,
						t->left->left,
						t->left->right,
						t->right,
						&t->tp,
						qlis);
	    }
	    else if(t->left->type == T_OP &&
		    t->left->val == STREF) /* tag.y = x, tag.y op= x */
	    {
		if(t->val == ASSIGN)
		{
		    q.op = STSET;
		    q.res = tree_quadlist1(t->left->left, &lq);
		    q.a1 = install_keyword_symbol(t->left->right->val);
		    sym = q.a2 = tree_quadlist1(t->right, &rq);
		    reuse_if_temp_sym(q.res);
		    node = make_quad_node1(&q);
		    *qlis = append_quadlist(&lq, &rq);
		    *qlis = add_quad_node_last(qlis, node);
		    qlis->last->tp = t->tp;
		}
		else
		{
		    sym = translate_struct_opset(t->val,
						 t->left->left,
						 t->left->right->val,
						 t->right,
						 &t->tp,
						 qlis);
		}
	    }
	    else		/* assign */
	    {
		q.op        = t->val;
		sym = q.res = tree_quadlist1(t->left,   &lq);
		q.a1        = tree_quadlist1(t->right,  &rq);
		q.a2        = null_symbol;
		reuse_if_temp_sym(q.a1);
		if(sym->type == SYM_INT) /* from enum */
		    yyerror("constant assignment");

		if(t->val == ASSIGN &&
		   t->right->type == T_OP &&
		   CAN_CHANGE_RESULT_ADDR_OP(t->right->val) &&
		   t->left->type == T_VAR  &&
		   IS_TEMP_SYM(q.a1))
		{
		    *qlis = rq;
		    qlis->last->u.q1.res = q.res;
		}
		else
		{
		    node = make_quad_node1(&q);
		    *qlis = append_quadlist(&lq, &rq);
		    *qlis = add_quad_node_last(qlis, node);
		    qlis->last->tp = t->tp;
		}
		sym->referenced_flag = True;
	    }
	} /* !IS_ASSIGNOP(t->val) */
	else if(IS_INCDECOP(t->val) && t->left->type == T_ARRAY)
	{   /* array inc/decrement */
	    Tree* a = t->left;

	    sym = newtemp_symbol();
	    ipush_translate(a->right, a->val, qlis);

	    *qlis = add_quadlistnode1_last(qlis,
					   t->val - XINC + AXINC,
					   sym,
					   install_variable(a->left->val,
							    a->left->module),
					   install_keyword_symbol(a->val));
	    sym->tp = qlis->last->tp = t->tp;
	}
	else if(IS_INCDECOP(t->val) && t->left->type == T_OP &&
		t->left->val  == INDIR)
	{
	    sym = translate_indir_incdec(t->val,
					 t->left->left,
					 &t->tp,
					 qlis);
	}
	else if(IS_INCDECOP(t->val) && t->left->type == T_ASSOCARRAY)
	{
	    sym = translate_assoc_incdec(t->val,
					 t->left->left,
					 t->left->right,
					 &t->tp,
					 qlis);
	}
	else if(t->val == OR)
	{
	    sym = translate_or_tree(t->left, t->right, qlis);
	}
	else if(t->val == AND)
	{
	    sym = translate_and_tree(t->left, t->right, qlis);
	}
	else if(t->val == STREF)
	{
	    Symbol* s1;
	    /* $0 = tag.mid => [stref res tag mid] */
	    sym = newtemp_symbol();
	    s1 = tree_quadlist1(t->left, qlis);
	    *qlis = add_quadlistnode1_last(
		qlis, STREF, sym, s1,
		install_keyword_symbol(t->right->val));
	    qlis->last->tp = t->tp;
	    reuse_if_temp_sym(s1);
	}
	else if(BETWEEN(ADD, t->val, LAST_OPERATOR))
	{
	    q.a1 = tree_quadlist1(t->left,  &lq);
	    q.a2 = tree_quadlist1(t->right,  &rq);
	    q.op  = t->val;
	    sym = q.res = newtemp_symbol();

	    reuse_if_temp_sym(q.a1);
	    reuse_if_temp_sym(q.a2);

	    node = make_quad_node1(&q);
	    *qlis = append_quadlist(&lq, &rq);
	    *qlis = add_quad_node_last(qlis, node);
	    sym->tp = qlis->last->tp = t->tp;
	}
	else
	{
	    fprintf(stderr, "Systematic error. undefined oprator:%ld\n", t->val);
	    exit(1);
	}
	break;

      case T_COMMA:
	sym = tree_quadlist1(t->left,  &lq);
	reuse_if_temp_sym(sym);
	sym = tree_quadlist1(t->right,  &rq);
	*qlis = append_quadlist(&lq, &rq);
	break;
      case T_VAR:
	init_quad_list(qlis);
	return tree_varnode_to_symbol(t);
	break;
      case T_QHKEYWORD:
	init_quad_list(qlis);
	sym = install_constant_symbol(t->val, SYM_KEYOBJ);
	sym->tp = t->tp;
	return sym;
	break;
      case T_CINT:
	init_quad_list(qlis);
	sym = install_constant_symbol(t->val, SYM_INT);
	sym->tp = t->tp;
	return sym;
	break;
      case T_CREAL:
	init_quad_list(qlis);
	sym = install_constant_symbol(t->val, SYM_REAL);
	sym->tp = t->tp;
	return sym;
	break;
      case T_CSTR:
	init_quad_list(qlis);
	sym = install_constant_symbol(t->val, SYM_STR);
	sym->tp = t->tp;
	return sym;
	break;
      case T_PARAM:
	/* error */
	fprintf(stderr, "Systematic error: T_PARAM is using in tree_quadlist1()\n");
	qhexit(1);
	break;
      case T_FCALL:
	/*            T_FCALL<narg>
	             /             \
          T_VAR<fname,module>       T_PARAM<None>
	*/
	return fcall_translate(t, qlis);
	break;
      case T_ARRAY:
	/*            T_ARRAY<dim>
	             /             \
           expression               T_PARAM<None>
	*/
	return array_translate(t, qlis);
	break;
      case T_FUNCALL:
	return funcall_translate(t->left, t->right, t->val, qlis);
	break;
      case T_NEW:
	/*
	 *     T_NEW<type>
	 *     /        \
	 * T_PARAM<dim>   NULL
	 */
	return translate_new(t->val, t->left->val, t->left, qlis, &t->tp);
	break;
      case T_COND3:
	/*
         *         T_COND3
         *        /       \
         *cond-expr        None
         *                /    \
         *       true-expr      false-expr
	 */
	return translate_cond3(t->left, t->right->left, t->right->right, qlis, &t->tp);
	break;

      case T_ASSOCARRAY:
	/*
	 *      T_ASSOCARRAY<None>
         *     /                  \
         * expr                    key
         */
	return translate_assoc_ref(t->left, t->right, &t->tp, qlis);
	break;

      case T_ASCNEW:
	/*
	 *      T_ASCNEW<None>
         *     /        \
         * NULL          NULL
         */
	return translate_ascnew(&t->tp, qlis);
	break;

      case T_STRNEW:
	/*
	 *      T_STRNEW<None>
         *     /        \
         * NULL,expr    NULL
         */
	return translate_strnew(&t->tp, t->left, qlis);
	break;
      case T_STTNEW:
	/*
	 *      T_STTNEW<None>
         *     /        \
         * NAME          NULL
         */
	return translate_sttnew(t->left, &t->tp, qlis);
	break;
      case T_LIST:
	/*
	 *      T_LIST<n>
         *     /      \
         * PARAM        NULL
         */
	return translate_list_op(t->left, t->val, &t->tp, qlis);
	break;
      default:
	fprintf(stderr, "Systematic error: Unknown tree data type: %d, in tree_quadlist1()\n", t->type);
	qhexit(1);
    }

    return sym;
}

Symbol* tree_quadlist(Tree* t, QuadList* qlis)
{
    if(t == NULL || t->type == None)
    {
	init_quad_list(qlis);
	return null_symbol;
    }

    if(check_expr(t) == True)
    {
	t = optimize_tree(t);
	return tree_quadlist1(t, qlis);
    }
    else
    {
	init_quad_list(qlis);
	return null_symbol;
    }
}


void orgoto_quad1(Tree* left, Tree* right, Symbol* label, QuadList* qlis)
{
    QuadList q;

    /*
      if(left || right) goto lable
      =>
      ifgoto(left)  label
      ifgoto(right) label
     */

    init_quad_list(&q);
    init_quad_list(qlis);

    /* left */
    if(left->type == T_OP && left->val == OR)
    {
	orgoto_quad1(left->left, left->right, label, &q);
    }
    else if(left->type == T_OP && left->val == AND)
    {
	andgoto_quad1(left->left, left->right, label, &q);
    }
    else
    {
	q = gen_if_goto1(left, label);
    }
    *qlis = append_quadlist(qlis, &q);


    /* right */
    if(right->type == T_OP && right->val == OR)
    {
	orgoto_quad1(right->left, right->right, label, &q);
    }
    else if(right->type == T_OP && right->val == AND)
    {
	andgoto_quad1(right->left, right->right, label, &q);
    }
    else
    {
	q = gen_if_goto1(right, label);
    }
    *qlis = append_quadlist(qlis, &q);
}

void andgoto_quad1(Tree* left, Tree* right, Symbol* label, QuadList* qlis)
{
    Symbol* sym;
    QuadList q;

    /*
      if(left && right) goto label
      =>
      ifngoto(left) sym
      ifgoto(right) label
      sym
     */

    init_quad_list(&q);
    init_quad_list(qlis);
    sym = install_label(newlabel(), current_module);

    /* left */
    if(left->type == T_OP && left->val == AND)
    {
	nandgoto_quad1(left->left, left->right, sym, &q);
    }
    else if(left->type == T_OP && left->val == OR)
    {
	norgoto_quad1(left->left, left->right, sym, &q);
    }
    else
    {
	q = gen_ifn_goto1(left, sym);
    }
    *qlis = append_quadlist(qlis, &q);


    /* right */
    if(right->type == T_OP && right->val == AND)
    {
	andgoto_quad1(right->left, right->right, label, &q);
    }
    else if(right->type == T_OP && right->val == OR)
    {
	orgoto_quad1(right->left, right->right, label, &q);
    }
    else
    {
	q = gen_if_goto1(right, label);
    }
    *qlis = append_quadlist(qlis, &q);

    *qlis = add_quad_node_last(qlis, gen_label_node(sym));
}

void norgoto_quad1(Tree* left, Tree* right, Symbol* label, QuadList* qlis)
{
    Symbol* sym;
    QuadList q;

    /*
      if(!(left || right)) goto label
      =>
      ifgoto(left) sym
      ifngoto(right) label
      sym
     */

    init_quad_list(qlis);
    init_quad_list(&q);
    sym = install_label(newlabel(), current_module);

    /* left */
    if(left->type == T_OP && left->val == OR)
    {
	orgoto_quad1(left->left, left->right, sym, &q);
    }
    else if(left->type == T_OP && left->val == AND)
    {
	andgoto_quad1(left->left, left->right, sym, &q);
    }
    else
    {
	q = gen_if_goto1(left, sym);
    }
    *qlis = append_quadlist(qlis, &q);


    /* right */
    if(right->type == T_OP && right->val == OR)
    {
	norgoto_quad1(right->left, right->right, label, &q);
    }
    else if(right->type == T_OP && right->val == AND)
    {
	nandgoto_quad1(right->left, right->right, label, &q);
    }
    else
    {
	q = gen_ifn_goto1(right, label);
    }
    *qlis = append_quadlist(qlis, &q);

    *qlis = add_quad_node_last(qlis, gen_label_node(sym));
}

void nandgoto_quad1(Tree* left, Tree* right, Symbol* label, QuadList* qlis)
{
    QuadList q;

    /*
      if(!(left && right)) goto label
      =>
      ifngoto(left) label
      ifngoto(right) label
      */

    init_quad_list(&q);
    init_quad_list(qlis);

    /* left */
    if(left->type == T_OP && left->val == AND)
    {
	nandgoto_quad1(left->left, left->right, label, &q);
    }
    else if(left->type == T_OP && left->val == OR)
    {
	norgoto_quad1(left->left, left->right, label, &q);
    }
    else
    {
	q = gen_ifn_goto1(left, label);
    }
    *qlis = append_quadlist(qlis, &q);


    /* right */
    if(right->type == T_OP && right->val == AND)
    {
	nandgoto_quad1(right->left, right->right, label, &q);
    }
    else if(right->type == T_OP && right->val == OR)
    {
	norgoto_quad1(right->left, right->right, label, &q);
    }
    else
    {
	q = gen_ifn_goto1(right, label);
    }
    *qlis = append_quadlist(qlis, &q);
}

/*
  orgoto_quad1(A, B, label) :=   if(A || B) goto lable
  =>
  ifgoto(A) label
  ifgoto(B) label

  andgoto_quad1(A, B, label) :=  if(A && B) goto label
  =>
  ifngoto(A) L
  ifgoto(B) label
  L

  norgoto_quad1(A, B, label) :=  if(!(A || B)) goto label
  =>
  ifgoto(A) L
  ifngoto(B) label
  L

  nandgoto_quad1(A, B, label) :=  if(!(A && B)) goto label
  =>
  ifngoto(A) label
  ifngoto(B) label

     ||     ||     ||     ||
    /  \   /  \   /  \   /  \  
   ||   A A   || &&   A A   &&

     ||     ||     ||
    /  \   /  \   /  \
   ||  && ||  && &&  &&
   
     &&     &&     &&     &&
    /  \   /  \   /  \   /  \
   ||   A A   || &&   A A   &&

     &&     &&     &&
    /  \   /  \   /  \
   ||  && ||  && &&  &&
*/

static Symbol* translate_or_tree(Tree* left, Tree* right, QuadList* q)
{
    Symbol* lab1;
    Symbol* lab2;
    Symbol* sym;

    /*
      X = (left || right)
      =>
      ifgoto(left) L1
      ifgoto(right) L1
      X = 0
      goto L2
      L1
      X = 1
      L2
     */

    lab1 = install_label(newlabel(), current_module);
    lab2 = install_label(newlabel(), current_module);
    sym = newtemp_symbol();

    orgoto_quad1(left, right, lab1, q);
    *q = add_quadlistnode1_last(q, ASSIGN, sym, install_integer(0), NULL);
    *q = add_quad_node_last(q, GOTONODE(lab2));
    *q = add_quad_node_last(q, gen_label_node(lab1));
    *q = add_quadlistnode1_last(q, ASSIGN, sym, install_integer(1), NULL);
    *q = add_quad_node_last(q, gen_label_node(lab2));

    return sym;
}

static Symbol* translate_and_tree(Tree* left, Tree* right, QuadList* q)
{
    Symbol* lab1;
    Symbol* lab2;
    Symbol* sym;

    /*
      X = (left && right)
      =>
      ifngoto(left) L1
      ifngoto(right) L1
      X = 1
      goto L2
      L1
      X = 0
      L2
     */

    lab1 = install_label(newlabel(), current_module);
    lab2 = install_label(newlabel(), current_module);
    sym = newtemp_symbol();

    nandgoto_quad1(left, right, lab1, q);
    *q = add_quadlistnode1_last(q, ASSIGN, sym, install_integer(1), NULL);
    *q = add_quad_node_last(q, GOTONODE(lab2));
    *q = add_quad_node_last(q, gen_label_node(lab1));
    *q = add_quadlistnode1_last(q, ASSIGN, sym, install_integer(0), NULL);
    *q = add_quad_node_last(q, gen_label_node(lab2));

    return sym;
}
