/* Copyright (c) 1995, Masanao Izumo */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "interp.h"
#if defined(sgi)
#include <strings.h>
#endif

char* flexible_buff = NULL;
int   flexible_buff_size = 0;
int   flexible_buff_ptr = 0;

#undef malloc
#undef calloc
#undef realloc
#undef strdup
#undef free

void* xmalloc(int size)
{
    void* p;

    if((p = malloc(size)) == NULL)
    {
	if(size > 0)
	{
	    if(OK_DO_garbage_collecting == False)
	    {
		fprintf(stderr, "Can't allocate memory.\n");
		qhexit(1);
	    }
	    else
	    {
		garbage_collect();
		p = malloc(size); /* try once more */
		if(p == NULL)
		{
		    fprintf(stderr, "Can't allocate memory.\n");
		    qhexit(1);
		}
	    }		
	}
	else
	{
	    fprintf(stderr, "Systematic error: Invalid xmalloc argument: %d\n", size);
	    qhexit(1);
	}
    }

    return p;
}

void* xcalloc(int nelem, int elsize)
{
    void* p;

    if((p = calloc(nelem, elsize)) == NULL)
    {
	if(nelem > 0)
	{
	    if(OK_DO_garbage_collecting == False)
	    {
		fprintf(stderr, "Can't allocate memory.\n");
		qhexit(1);
	    }
	    else
	    {
		garbage_collect();
		p = calloc(nelem, elsize); /* try once more */
		if(p == NULL)
		{
		    fprintf(stderr, "Can't allocate memory.\n");
		    qhexit(1);
		}
	    }		
	}
	else
	{
	    fprintf(stderr, "Systematic error: Invalid xmalloc argument: %d\n", nelem);
	    qhexit(1);
	}
    }

    return p;
}

void* xrealloc(void* p, int size)
{
    if((p = realloc(p, size)) == NULL)
    {
	if(size > 0)
	{
	    if(OK_DO_garbage_collecting == False)
	    {
		fprintf(stderr, "Can't reallocate memory.\n");
		qhexit(1);
	    }
	    else
	    {
		garbage_collect();
		p = realloc(p, size); /* try once more */
		if(p == NULL)
		{
		    fprintf(stderr, "Can't reallocate memory.\n");
		    qhexit(1);
		}
	    }
	}
	else
	{
	    fprintf(stderr, "Systematic error: Invalid xrealloc argument: %d\n", size);
	    qhexit(1);
	}
    }

    return p;
}

void* xcrealloc(void* p, int old_size, int new_size)
{
    if(old_size >= new_size)
	return p;

    p = xrealloc(p, new_size);
    bzero((char *)p + old_size, new_size - old_size);

    return p;
}

char* xstrdup(const char* s)
{
    char* p;

    p = strdup(s);

    if(p == NULL)
    {
	if(OK_DO_garbage_collecting == False)
	{
	    fprintf(stderr, "Can't allocate memory.\n");
	    qhexit(1);
	}
	else
	{
	    garbage_collect();
	    p = strdup(s);	/* try once more */
	    if(p == NULL)
	    {
		fprintf(stderr, "Can't allocate memory.\n");
		qhexit(1);
	    }
	}
    }

    return p;
}

void xfree(void* p)
{
    if(p != NULL)
	free(p);
}

void clear_flexible_buff(void)
{
    xfree(flexible_buff);
    flexible_buff = NULL;
    flexible_buff_size = 0;
    flexible_buff_ptr = 0;
}

void need_flexible_buff(int size)
{
    if(flexible_buff_size == 0)
    {
	size = MAX(size, BUFSIZ);
	flexible_buff = (char *)xmalloc(size);
	flexible_buff_size = size;
    }
    else if(size > flexible_buff_size)
    {
	while(size > flexible_buff_size)
	    flexible_buff_size *= 2;
	flexible_buff = (char *)xrealloc(flexible_buff, flexible_buff_size);
    }
}

void add_string_to_flexible_buff(const char* s)
{
    int i, len;

    len = strlen(s);

    need_flexible_buff(len + flexible_buff_ptr + 1);
    
    for(i = 0; i < len; i++)
	flexible_buff[flexible_buff_ptr + i] = *s++;
    flexible_buff[flexible_buff_ptr + len] = '\0';
    flexible_buff_ptr += len;
}

char* get_flexible_buff(void)
{
    char* p;

    p = flexible_buff;
    flexible_buff = NULL;
    clear_flexible_buff();

    return p;
}
